@@ -1,25 +1,21 @@
 /* checksig.c: verify the signature of an RPM */
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <unistd.h>
-#include <fcntl.h>
-#include <string.h>
+#include "system.h"
+
+#include "build/rpmbuild.h"
 
 #include "checksig.h"
-#include "intl.h"
-#include "rpmlib.h"
 #include "rpmlead.h"
 #include "signature.h"
-#include "messages.h"
-#include "lib/misc.h"
 
-int doReSign(int add, char *passPhrase, char **argv)
+int doReSign(int add, char *passPhrase, const char **argv)
 {
-    int fd, ofd, count;
+    FD_t fd, ofd;
+    int count;
     struct rpmlead lead;
     unsigned short sigtype;
-    char *rpm, *sigtarget;
+    const char *rpm;
+    const char *sigtarget;
     char tmprpm[1024];
     unsigned char buffer[8192];
     Header sig;
@@ -27,25 +23,25 @@ int doReSign(int add, char *passPhrase, char **argv)
     while (*argv) {
 	rpm = *argv++;
 	fprintf(stdout, "%s:\n", rpm);
-	if ((fd = open(rpm, O_RDONLY, 0644)) < 0) {
+	if (fdFileno(fd = fdOpen(rpm, O_RDONLY, 0644)) < 0) {
 	    fprintf(stderr, _("%s: Open failed\n"), rpm);
-	    exit(1);
+	    exit(EXIT_FAILURE);
 	}
 	if (readLead(fd, &lead)) {
 	    fprintf(stderr, _("%s: readLead failed\n"), rpm);
-	    exit(1);
+	    exit(EXIT_FAILURE);
 	}
 	if (lead.major == 1) {
 	    fprintf(stderr, _("%s: Can't sign v1.0 RPM\n"), rpm);
-	    exit(1);
+	    exit(EXIT_FAILURE);
 	}
 	if (lead.major == 2) {
 	    fprintf(stderr, _("%s: Can't re-sign v2.0 RPM\n"), rpm);
-	    exit(1);
+	    exit(EXIT_FAILURE);
 	}
 	if (rpmReadSignature(fd, &sig, lead.signature_type)) {
 	    fprintf(stderr, _("%s: rpmReadSignature failed\n"), rpm);
-	    exit(1);
+	    exit(EXIT_FAILURE);
 	}
 	if (add != ADD_SIGNATURE) {
 	    rpmFreeSignature(sig);
@@ -53,42 +49,45 @@ int doReSign(int add, char *passPhrase, char **argv)
 
 	/* Write the rest to a temp file */
 	if (makeTempFile(NULL, &sigtarget, &ofd))
-	    exit(1);
+	    exit(EXIT_FAILURE);
 
-	while ((count = read(fd, buffer, sizeof(buffer))) > 0) {
+	while ((count = fdRead(fd, buffer, sizeof(buffer))) > 0) {
 	    if (count == -1) {
 		perror(_("Couldn't read the header/archive"));
-		close(ofd);
+		fdClose(ofd);
 		unlink(sigtarget);
-		free(sigtarget);
-		exit(1);
+		xfree(sigtarget);
+		exit(EXIT_FAILURE);
 	    }
-	    if (write(ofd, buffer, count) < 0) {
+	    if (fdWrite(ofd, buffer, count) < 0) {
 		perror(_("Couldn't write header/archive to temp file"));
-		close(ofd);
+		fdClose(ofd);
 		unlink(sigtarget);
-		free(sigtarget);
-		exit(1);
+		xfree(sigtarget);
+		exit(EXIT_FAILURE);
 	    }
 	}
-	close(fd);
-	close(ofd);
+	fdClose(fd);
+	fdClose(ofd);
 
 	/* Start writing the new RPM */
-	sprintf(tmprpm, "%s.tmp", rpm);
-	ofd = open(tmprpm, O_WRONLY|O_CREAT|O_TRUNC, 0644);
+	strcpy(tmprpm, rpm);
+	strcat(tmprpm, ".XXXXXX");
+	mktemp(tmprpm);
+
+	ofd = fdOpen(tmprpm, O_WRONLY|O_CREAT|O_TRUNC, 0644);
 	lead.signature_type = RPMSIG_HEADERSIG;
 	if (writeLead(ofd, &lead)) {
 	    perror("writeLead()");
-	    close(ofd);
+	    fdClose(ofd);
 	    unlink(sigtarget);
 	    unlink(tmprpm);
-	    free(sigtarget);
-	    exit(1);
+	    xfree(sigtarget);
+	    exit(EXIT_FAILURE);
 	}
 
 	/* Generate the signature */
-	sigtype = rpmLookupSignatureType();
+	sigtype = rpmLookupSignatureType(RPMLOOKUPSIG_QUERY);
 	rpmMessage(RPMMESS_VERBOSE, _("Generating signature: %d\n"), sigtype);
 	if (add != ADD_SIGNATURE) {
 	    sig = rpmNewSignature();
@@ -99,41 +98,41 @@ int doReSign(int add, char *passPhrase, char **argv)
 	    rpmAddSignature(sig, sigtarget, sigtype, passPhrase);
 	}
 	if (rpmWriteSignature(ofd, sig)) {
-	    close(ofd);
+	    fdClose(ofd);
 	    unlink(sigtarget);
 	    unlink(tmprpm);
-	    free(sigtarget);
+	    xfree(sigtarget);
 	    rpmFreeSignature(sig);
-	    exit(1);
+	    exit(EXIT_FAILURE);
 	}
 	rpmFreeSignature(sig);
 
 	/* Append the header and archive */
-	fd = open(sigtarget, O_RDONLY);
-	while ((count = read(fd, buffer, sizeof(buffer))) > 0) {
+	fd = fdOpen(sigtarget, O_RDONLY, 0);
+	while ((count = fdRead(fd, buffer, sizeof(buffer))) > 0) {
 	    if (count == -1) {
 		perror(_("Couldn't read sigtarget"));
-		close(ofd);
-		close(fd);
+		fdClose(ofd);
+		fdClose(fd);
 		unlink(sigtarget);
 		unlink(tmprpm);
-		free(sigtarget);
-		exit(1);
+		xfree(sigtarget);
+		exit(EXIT_FAILURE);
 	    }
-	    if (write(ofd, buffer, count) < 0) {
+	    if (fdWrite(ofd, buffer, count) < 0) {
 		perror(_("Couldn't write package"));
-		close(ofd);
-		close(fd);
+		fdClose(ofd);
+		fdClose(fd);
 		unlink(sigtarget);
 		unlink(tmprpm);
-		free(sigtarget);
-		exit(1);
+		xfree(sigtarget);
+		exit(EXIT_FAILURE);
 	    }
 	}
-	close(fd);
-	close(ofd);
+	fdClose(fd);
+	fdClose(ofd);
 	unlink(sigtarget);
-	free(sigtarget);
+	xfree(sigtarget);
 
 	/* Move it in to place */
 	unlink(rpm);
@@ -143,13 +142,17 @@ int doReSign(int add, char *passPhrase, char **argv)
     return 0;
 }
 
-int doCheckSig(int flags, char **argv)
+int doCheckSig(int flags, const char **argv)
 {
-    int fd, ofd, res, res2, res3, missingKeys;
+    FD_t fd, ofd;
+    int res, res2, res3;
     struct rpmlead lead;
-    char *rpm;
-    char result[1024], * sigtarget;
+    const char *rpm;
+    char result[1024];
+    const char * sigtarget;
     unsigned char buffer[8192];
+    unsigned char missingKeys[7164];
+    unsigned char untrustedKeys[7164];
     Header sig;
     HeaderIterator sigIter;
     int_32 tag, type, count;
@@ -158,7 +161,7 @@ int doCheckSig(int flags, char **argv)
     res = 0;
     while (*argv) {
 	rpm = *argv++;
-	if ((fd = open(rpm, O_RDONLY, 0644)) < 0) {
+	if (fdFileno(fd = fdOpen(rpm, O_RDONLY, 0644)) < 0) {
 	    fprintf(stderr, _("%s: Open failed\n"), rpm);
 	    res++;
 	    continue;
@@ -178,46 +181,47 @@ int doCheckSig(int flags, char **argv)
 	    res++;
 	    continue;
 	}
-	if (! sig) {
+	if (sig == NULL) {
 	    fprintf(stderr, _("%s: No signature available\n"), rpm);
 	    res++;
 	    continue;
 	}
 	/* Write the rest to a temp file */
 	if (makeTempFile(NULL, &sigtarget, &ofd))
-	    exit(1);
-	while ((count = read(fd, buffer, sizeof(buffer))) > 0) {
+	    exit(EXIT_FAILURE);
+	while ((count = fdRead(fd, buffer, sizeof(buffer))) > 0) {
 	    if (count == -1) {
 		perror(_("Couldn't read the header/archive"));
-		close(ofd);
+		fdClose(ofd);
 		unlink(sigtarget);
-		free(sigtarget);
-		exit(1);
+		xfree(sigtarget);
+		exit(EXIT_FAILURE);
 	    }
-	    if (write(ofd, buffer, count) < 0) {
+	    if (fdWrite(ofd, buffer, count) < 0) {
 		fprintf(stderr, _("Unable to write %s"), sigtarget);
 		perror("");
-		close(ofd);
+		fdClose(ofd);
 		unlink(sigtarget);
-		free(sigtarget);
-		exit(1);
+		xfree(sigtarget);
+		exit(EXIT_FAILURE);
 	    }
 	}
-	close(fd);
-	close(ofd);
+	fdClose(fd);
+	fdClose(ofd);
+
+	res2 = 0;
+	missingKeys[0] = '\0';
+	untrustedKeys[0] = '\0';
+	sprintf(buffer, "%s:%c", rpm, (rpmIsVerbose() ? '\n' : ' ') );
 
 	sigIter = headerInitIterator(sig);
-	res2 = 0;
-	missingKeys = 0;
-	if (rpmIsVerbose()) {
-	    sprintf(buffer, "%s:\n", rpm);
-	} else {
-	    sprintf(buffer, "%s: ", rpm);
-	}
 	while (headerNextIterator(sigIter, &tag, &type, &ptr, &count)) {
-	    if ((tag == RPMSIGTAG_PGP) && !(flags & CHECKSIG_PGP)) 
+	    if ((tag == RPMSIGTAG_PGP || tag == RPMSIGTAG_PGP5)
+	    	    && !(flags & CHECKSIG_PGP)) 
 		continue;
-	    else if ((tag == RPMSIGTAG_MD5 || 
+	    if ((tag == RPMSIGTAG_GPG) && !(flags & CHECKSIG_GPG))
+		continue;
+	    if ((tag == RPMSIGTAG_MD5 || 
 		      tag == RPMSIGTAG_LEMD5_2 ||
 		      tag == RPMSIGTAG_LEMD5_1) 
 		      && !(flags & CHECKSIG_MD5)) 
@@ -229,6 +233,7 @@ int doCheckSig(int flags, char **argv)
 		    strcat(buffer, result);
 		    res2 = 1;
 		} else {
+		    char *tempKey;
 		    switch (tag) {
 		      case RPMSIGTAG_SIZE:
 			strcat(buffer, "SIZE ");
@@ -241,15 +246,43 @@ int doCheckSig(int flags, char **argv)
 			res2 = 1;
 			break;
 		      case RPMSIGTAG_PGP:
-			if (res3 == RPMSIG_NOKEY) {
-			    /* Do not consedier this a failure */
+		      case RPMSIGTAG_PGP5:
+			if (res3 == RPMSIG_NOKEY || res3 == RPMSIG_NOTTRUSTED) {
+			    /* Do not consider these a failure */
+			    int offset = 7;
 			    strcat(buffer, "(PGP) ");
-			    missingKeys = 1;
+			    tempKey = strstr(result, "Key ID");
+			    if (tempKey == NULL) {
+			        tempKey = strstr(result, "keyid:");
+				offset = 9;
+			    }
+			    if (tempKey) {
+			      if (res3 == RPMSIG_NOKEY) {
+				strcat(missingKeys, " PGP#");
+				strncat(missingKeys, tempKey + offset, 8);
+			      } else {
+			        strcat(untrustedKeys, " PGP#");
+				strncat(untrustedKeys, tempKey + offset, 8);
+			      }
+			    }
 			} else {
 			    strcat(buffer, "PGP ");
 			    res2 = 1;
 			}
 			break;
+		      case RPMSIGTAG_GPG:
+			if (res3 == RPMSIG_NOKEY) {
+			    /* Do not consider this a failure */
+			    strcat(buffer, "(GPG) ");
+			    strcat(missingKeys, " GPG#");
+			    tempKey = strstr(result, "key ID");
+			    if (tempKey)
+				strncat(missingKeys, tempKey+7, 8);
+			} else {
+			    strcat(buffer, "GPG ");
+			    res2 = 1;
+			}
+			break;
 		      default:
 			strcat(buffer, "?UnknownSignatureType? ");
 			res2 = 1;
@@ -269,8 +302,12 @@ int doCheckSig(int flags, char **argv)
 			strcat(buffer, "md5 ");
 			break;
 		      case RPMSIGTAG_PGP:
+		      case RPMSIGTAG_PGP5:
 			strcat(buffer, "pgp ");
 			break;
+		      case RPMSIGTAG_GPG:
+			strcat(buffer, "gpg ");
+			break;
 		      default:
 			strcat(buffer, "??? ");
 		    }
@@ -280,21 +317,34 @@ int doCheckSig(int flags, char **argv)
 	headerFreeIterator(sigIter);
 	res += res2;
 	unlink(sigtarget);
-	free(sigtarget);
+	xfree(sigtarget);
 
 	if (res2) {
 	    if (rpmIsVerbose()) {
-		fprintf(stderr, "%s", buffer);
+		fprintf(stderr, "%s", (char *)buffer);
 	    } else {
-		fprintf(stderr, "%s%s%s\n", buffer, _("NOT OK"),
-			missingKeys ? _(" (MISSING KEYS)") : "");
+		fprintf(stderr, "%s%s%s%s%s%s%s%s\n", (char *)buffer,
+			_("NOT OK"),
+			(missingKeys[0] != '\0') ? _(" (MISSING KEYS:") : "",
+			(char *)missingKeys,
+			(missingKeys[0] != '\0') ? _(") ") : "",
+			(untrustedKeys[0] != '\0') ? _(" (UNTRUSTED KEYS:") : "",
+			(char *)untrustedKeys,
+			(untrustedKeys[0] != '\0') ? _(")") : "");
+
 	    }
 	} else {
 	    if (rpmIsVerbose()) {
-		fprintf(stdout, "%s", buffer);
+		fprintf(stdout, "%s", (char *)buffer);
 	    } else {
-		fprintf(stdout, "%s%s%s\n", buffer, _("OK"),
-		       missingKeys ? _(" (MISSING KEYS)") : "");
+		fprintf(stdout, "%s%s%s%s%s%s%s%s\n", (char *)buffer,
+			_("OK"),
+			(missingKeys[0] != '\0') ? _(" (MISSING KEYS:") : "",
+			(char *)missingKeys,
+			(missingKeys[0] != '\0') ? _(") ") : "",
+			(untrustedKeys[0] != '\0') ? _(" (UNTRUSTED KEYS:") : "",
+			(char *)untrustedKeys,
+			(untrustedKeys[0] != '\0') ? _(")") : "");
 	    }
 	}
     }
