@@ -1,16 +1,13 @@
 /* This converts an old style (rpm 1.x) database to the new style */
 
-#include <fcntl.h>
-#include <stdlib.h>
-#include <unistd.h>
+#include "system.h"
+
+#include "build/rpmbuild.h"
 
-#include "dbindex.h"
 #include "falloc.h"
-#include "header.h"
-#include "misc.h"
 #include "oldrpmdb.h"
 #include "oldheader.h"
-#include "rpmlib.h"
+#include "lib/rpmdb.h"
 
 int convertDB(void);
 
@@ -37,7 +34,7 @@ int convertDB(void) {
     char * preun, * postun;
     int i, j;
     
-    if (exists("/var/lib/rpm/packages.rpm")) {
+    if (rpmfileexists("/var/lib/rpm/packages.rpm")) {
 	rpmError(RPMERR_NOCREATEDB, _("RPM database already exists"));
 	return 0;
     }
@@ -70,7 +67,7 @@ int convertDB(void) {
     for (label = packageLabels; label; label = label->next) {
 	if (oldrpmdbGetPackageInfo(&olddb, *label, &package)) {
 	    fprintf(stderr, _("oldrpmdbGetPackageInfo failed &olddb = %p olddb.packages = %p\n"), &olddb, olddb.packages);
-	    exit(1);
+	    exit(EXIT_FAILURE);
 	}
 
 	group = oldrpmdbGetPackageGroup(&olddb, *label);
@@ -202,13 +199,13 @@ int convertDB(void) {
 int main(int argc, char ** argv) {
     if (argc != 1) {
 	fprintf(stderr, _("rpmconvert: no arguments expected"));
-	exit(1);
+	exit(EXIT_FAILURE);
     }
 
-    rpmReadConfigFiles(NULL, NULL, NULL, 0);
+    rpmReadConfigFiles(NULL, NULL);
 
     printf(_("rpmconvert 1.0 - converting database in /var/lib/rpm\n"));
     convertDB();
 
-    exit(0);
+    exit(EXIT_SUCCESS);
 }
