@@ -1,29 +1,7 @@
-#include "config.h"
-#include "miscfn.h"
+#include "system.h"
 
-#if HAVE_ALLOCA_H
-# include <alloca.h>
-#endif
-
-#include <errno.h>
-#include <fcntl.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <unistd.h>
-#include <utime.h>
-
-#include "intl.h"
 #include "cpio.h"
 
-#if MAJOR_IN_SYSMACROS 
-#include <sys/sysmacros.h>
-#elif MAJOR_IN_MKDEV 
-#include <sys/mkdev.h>
-#endif
-
 #define CPIO_NEWC_MAGIC	"070701"
 #define CPIO_CRC_MAGIC	"070702"
 #define TRAILER		"TRAILER!!!"
@@ -32,12 +10,6 @@
    should both be the same, but really odd things are going to happen if
    that's not true! */
 
-/* We need to maintain our oun file pointer to allow padding */
-struct ourfd {
-    gzFile fd;
-    int pos;
-};
-
 struct hardLink {
     struct hardLink * next;
     char ** files;		/* there are nlink of these, used by install */
@@ -67,6 +39,8 @@ struct cpioCrcPhysicalHeader {
     char checksum[8];			/* ignored !! */
 };
 
+#define	PHYS_HDR_SIZE	110		/* don't depend on sizeof(struct) */
+
 struct cpioHeader {
     ino_t inode;
     mode_t mode;
@@ -79,47 +53,111 @@ struct cpioHeader {
     char * path;
 };
 
-static inline off_t ourread(struct ourfd * thefd, void * buf, size_t size) {
-    off_t i;
+static inline off_t saferead(CFD_t *cfd, void * vbuf, size_t amount) {
+    off_t rc = 0;
+    char * buf = vbuf;
 
-    i = gzread(thefd->fd, buf, size);
-    thefd->pos += i;
-    
+    while (amount > 0) {
+	size_t nb;
+	switch (cfd->cpioIoType) {
+	default:
+#ifdef	PARANOID
+		fprintf(stderr, "\tsaferead(%p,%p,%x)\n", cfd, vbuf, (unsigned)amount);
+		exit(1);
+		break;
+#endif
+	case cpioIoTypeDebug:
+		nb = amount;
+		fprintf(stderr, "\tsaferead(%p,%p,%x)\n", cfd, vbuf, (unsigned)amount);
+		break;
+	case cpioIoTypeFd:
+		nb = fdRead(cfd->cpioFd, buf, amount);
+		break;
+	case cpioIoTypeFp:
+		nb = fread(buf, amount, 1, cfd->cpioFp);
+		if (nb)
+			nb *= amount;
+		break;
+	case cpioIoTypeGzFd:
+		nb = gzdRead(cfd->cpioGzFd, buf, amount);
+		break;
+	}
+	if (nb <= 0)
+		return nb;
+	rc += nb;
+	if (rc >= amount)
+		break;
+	buf += nb;
+	amount -= nb;
+    }
+    return rc;
+}
+
+static inline off_t ourread(CFD_t * cfd, void * buf, size_t size) {
+    off_t i = saferead(cfd, buf, size);
+    if (i > 0)
+	cfd->cpioPos += i;
     return i;
 }
 
-static inline void padinfd(struct ourfd * fd, int modulo) {
+static inline void padinfd(CFD_t * cfd, int modulo) {
     int buf[10];
     int amount;
     
-    amount = (modulo - fd->pos % modulo) % modulo;
-    ourread(fd, buf, amount);
+    amount = (modulo - cfd->cpioPos % modulo) % modulo;
+    (void)ourread(cfd, buf, amount);
 }
 
-static inline off_t safewrite(int fd, void * vbuf, size_t amount) {
-    int rc = 0;
+static inline off_t safewrite(CFD_t *cfd, void * vbuf, size_t amount) {
+    off_t rc = 0;
     char * buf = vbuf;
 
-    while (amount && (rc = write(fd, buf, amount)) < amount && rc > 0) {
-	buf += rc;
-	amount -= rc;
+    while (amount > 0) {
+	size_t nb;
+	switch (cfd->cpioIoType) {
+	default:
+#ifdef	PARANOID
+		fprintf(stderr, "\tsafewrite(%p,%p,%x)\n", cfd, vbuf, (unsigned)amount);
+		exit(1);
+		break;
+#endif
+	case cpioIoTypeDebug:
+		nb = amount;
+		fprintf(stderr, "\tsafewrite(%p,%p,%x)\n", cfd, vbuf, (unsigned)amount);
+		break;
+	case cpioIoTypeFd:
+		nb = fdWrite(cfd->cpioFd, buf, amount);
+		break;
+	case cpioIoTypeFp:
+		nb = fwrite(buf, amount, 1, cfd->cpioFp);
+		if (nb)
+			nb *= amount;
+		break;
+	case cpioIoTypeGzFd:
+		nb = gzdWrite(cfd->cpioGzFd, buf, amount);
+		break;
+	}
+	if (nb <= 0)
+		return nb;
+	rc += nb;
+	if (rc >= amount)
+		break;
+	buf += nb;
+	amount -= nb;
     }
 
-    if (rc < 0) return rc;
-
-    return 0; 
+    return rc; 
 }
 
-static inline int padoutfd(int fd, size_t * where, int modulo) {
-    static int buf[10] = { '\0', '\0', '\0', '\0', '\0', 
-			   '\0', '\0', '\0', '\0', '\0' };
+static inline int padoutfd(CFD_t * cfd, size_t * where, int modulo) {
+    static int buf[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
     int amount;
     
     amount = (modulo - *where % modulo) % modulo;
     *where += amount;
 
-    if (safewrite(fd, buf, amount)) 
-	return CPIO_WRITE_FAILED;
+    if (safewrite(cfd, buf, amount) != amount) 
+	return CPIOERR_WRITE_FAILED;
     return 0;
 }
 
@@ -133,7 +171,7 @@ static int strntoul(const char * str, char ** endptr, int base, int num) {
 
     ret = strtoul(buf, &end, base);
     if (*end)
-	*endptr = str + (end - buf);
+	*endptr = ((char *)str) + (end - buf);	/* XXX discards const */
     else
 	*endptr = "";
 
@@ -142,23 +180,23 @@ static int strntoul(const char * str, char ** endptr, int base, int num) {
 
 #define GET_NUM_FIELD(phys, log) \
 	log = strntoul(phys, &end, 16, sizeof(phys)); \
-	if (*end) return CPIO_BAD_HEADER;
+	if (*end) return CPIOERR_BAD_HEADER;
 #define SET_NUM_FIELD(phys, val, space) \
 	sprintf(space, "%8.8lx", (unsigned long) (val)); \
 	memcpy(phys, space, 8);
 
-static int getNextHeader(struct ourfd * fd, struct cpioHeader * chPtr) {
+static int getNextHeader(CFD_t * cfd, struct cpioHeader * chPtr) {
     struct cpioCrcPhysicalHeader physHeader;
     int nameSize;
     char * end;
     int major, minor;
 
-    if (ourread(fd, &physHeader, sizeof(physHeader)) != sizeof(physHeader)) 
-	return CPIO_READ_FAILED;
+    if (ourread(cfd, &physHeader, PHYS_HDR_SIZE) != PHYS_HDR_SIZE) 
+	return CPIOERR_READ_FAILED;
 
     if (strncmp(CPIO_CRC_MAGIC, physHeader.magic, strlen(CPIO_CRC_MAGIC)) &&
 	strncmp(CPIO_NEWC_MAGIC, physHeader.magic, strlen(CPIO_NEWC_MAGIC)))
-	return CPIO_BAD_MAGIC;
+	return CPIOERR_BAD_MAGIC;
 
     GET_NUM_FIELD(physHeader.inode, chPtr->inode);
     GET_NUM_FIELD(physHeader.mode, chPtr->mode);
@@ -170,23 +208,27 @@ static int getNextHeader(struct ourfd * fd, struct cpioHeader * chPtr) {
 
     GET_NUM_FIELD(physHeader.devMajor, major);
     GET_NUM_FIELD(physHeader.devMinor, minor);
+#ifndef __LCLINT__
     chPtr->dev = makedev(major, minor);
+#endif
 
     GET_NUM_FIELD(physHeader.rdevMajor, major);
     GET_NUM_FIELD(physHeader.rdevMinor, minor);
+#ifndef __LCLINT__
     chPtr->rdev = makedev(major, minor);
+#endif
 
     GET_NUM_FIELD(physHeader.namesize, nameSize);
 
     chPtr->path = malloc(nameSize + 1);
-    if (ourread(fd, chPtr->path, nameSize) != nameSize) {
+    if (ourread(cfd, chPtr->path, nameSize) != nameSize) {
 	free(chPtr->path);
-	return CPIO_BAD_HEADER;
+	return CPIOERR_BAD_HEADER;
     }
 
     /* this is unecessary chPtr->path[nameSize] = '\0'; */
 
-    padinfd(fd, 4);
+    padinfd(cfd, 4);
 
     return 0;
 }
@@ -201,15 +243,15 @@ int cpioFileMapCmp(const void * a, const void * b) {
 /* This could trash files in the path! I'm not sure that's a good thing */
 static int createDirectory(char * path, mode_t perms) {
     struct stat sb;
-    int dounlink;
 
     if (!lstat(path, &sb)) {
+	int dounlink = 0;	/* XXX eliminate, dounlink==1 on all paths */
 	if (S_ISDIR(sb.st_mode)) {
 	    return 0;
 	} else if (S_ISLNK(sb.st_mode)) {
 	    if (stat(path, &sb)) {
 		if (errno != ENOENT) 
-		    return CPIO_STAT_FAILED;
+		    return CPIOERR_STAT_FAILED;
 		dounlink = 1;
 	    } else {
 		if (S_ISDIR(sb.st_mode))
@@ -221,15 +263,15 @@ static int createDirectory(char * path, mode_t perms) {
 	}
 
 	if (dounlink && unlink(path)) {
-	    return CPIO_UNLINK_FAILED;
+	    return CPIOERR_UNLINK_FAILED;
 	}
     }
 
     if (mkdir(path, 000))
-	return CPIO_MKDIR_FAILED;
+	return CPIOERR_MKDIR_FAILED;
 
     if (chmod(path, perms))
-	return CPIO_CHMOD_FAILED;
+	return CPIOERR_CHMOD_FAILED;
 
     return 0;
 }
@@ -243,15 +285,15 @@ static int setInfo(struct cpioHeader * hdr) {
 
     if (!S_ISLNK(hdr->mode)) {
 	if (!getuid() && chown(hdr->path, hdr->uid, hdr->gid))
-	    rc = CPIO_CHOWN_FAILED;
+	    rc = CPIOERR_CHOWN_FAILED;
 	if (!rc && chmod(hdr->path, hdr->mode & 07777))
-	    rc = CPIO_CHMOD_FAILED;
+	    rc = CPIOERR_CHMOD_FAILED;
 	if (!rc && utime(hdr->path, &stamp))
-	    rc = CPIO_UTIME_FAILED;
+	    rc = CPIOERR_UTIME_FAILED;
     } else {
 #       if ! CHOWN_FOLLOWS_SYMLINK
 	    if (!getuid() && !rc && lchown(hdr->path, hdr->uid, hdr->gid))
-		rc = CPIO_CHOWN_FAILED;
+		rc = CPIOERR_CHOWN_FAILED;
 #       endif
     }
 
@@ -302,9 +344,9 @@ static int checkDirectory(char * filename) {
     return rc;
 }
 
-static int expandRegular(struct ourfd * fd, struct cpioHeader * hdr,
+static int expandRegular(CFD_t * cfd, struct cpioHeader * hdr,
 			 cpioCallback cb, void * cbData) {
-    int out;
+    FD_t out;
     char buf[8192];
     int bytesRead;
     int left = hdr->size;
@@ -312,26 +354,41 @@ static int expandRegular(struct ourfd * fd, struct cpioHeader * hdr,
     struct cpioCallbackInfo cbInfo;
     struct stat sb;
 
-    if (!lstat(hdr->path, &sb))
-	if (unlink(hdr->path))
-	    return CPIO_UNLINK_FAILED;
+    /* Rename the old file before attempting unlink to avoid EBUSY errors */
+    if (!lstat(hdr->path, &sb)) {
+	strcpy(buf, hdr->path);
+	strcat(buf, "-RPMDELETE");
+	if (rename(hdr->path, buf)) {
+	    fprintf(stderr, _("can't rename %s to %s: %s\n"),
+		hdr->path, buf, strerror(errno));
+            return CPIOERR_UNLINK_FAILED;
+	}
 
-    out = open(hdr->path, O_CREAT | O_WRONLY, 0);
-    if (out < 0) 
-	return CPIO_OPEN_FAILED;
+	if (unlink(buf)) {
+	    fprintf(stderr, _("can't unlink %s: %s\n"),
+		buf, strerror(errno));
+#if 0
+	    return CPIOERR_UNLINK_FAILED;
+#endif
+	}
+    }
+
+    out = fdOpen(hdr->path, O_CREAT | O_WRONLY, 0);
+    if (fdFileno(out) < 0) 
+	return CPIOERR_OPEN_FAILED;
 
     cbInfo.file = hdr->path;
     cbInfo.fileSize = hdr->size;
 
     while (left) {
-	bytesRead = ourread(fd, buf, left < sizeof(buf) ? left : sizeof(buf));
+	bytesRead = ourread(cfd, buf, left < sizeof(buf) ? left : sizeof(buf));
 	if (bytesRead <= 0) {
-	    rc = CPIO_READ_FAILED;
+	    rc = CPIOERR_READ_FAILED;
 	    break;
 	}
 
-	if (write(out, buf, bytesRead) != bytesRead) {
-	    rc = CPIO_READ_FAILED;
+	if (fdWrite(out, buf, bytesRead) != bytesRead) {
+	    rc = CPIOERR_COPY_FAILED;
 	    break;
 	}
 
@@ -340,26 +397,26 @@ static int expandRegular(struct ourfd * fd, struct cpioHeader * hdr,
 	/* don't call this with fileSize == fileComplete */
 	if (!rc && cb && left) {
 	    cbInfo.fileComplete = hdr->size - left;
-	    cbInfo.bytesProcessed = fd->pos;
+	    cbInfo.bytesProcessed = cfd->cpioPos;
 	    cb(&cbInfo, cbData);
 	}
     }
 
-    close(out);
+    fdClose(out);
     
     return rc;
 }
 
-static int expandSymlink(struct ourfd * fd, struct cpioHeader * hdr) {
+static int expandSymlink(CFD_t * cfd, struct cpioHeader * hdr) {
     char buf[2048], buf2[2048];
     struct stat sb;
     int len;
 
     if ((hdr->size + 1)> sizeof(buf))
-	return CPIO_INTERNAL;
+	return CPIOERR_HDR_SIZE;
 
-    if (ourread(fd, buf, hdr->size) != hdr->size)
-	return CPIO_READ_FAILED;
+    if (ourread(cfd, buf, hdr->size) != hdr->size)
+	return CPIOERR_READ_FAILED;
 
     buf[hdr->size] = '\0';
 
@@ -373,32 +430,32 @@ static int expandSymlink(struct ourfd * fd, struct cpioHeader * hdr) {
 	}
 
 	if (unlink(hdr->path))
-	    return CPIO_UNLINK_FAILED;
+	    return CPIOERR_UNLINK_FAILED;
     }
 
     if (symlink(buf, hdr->path) < 0)
-	return CPIO_SYMLINK_FAILED;
+	return CPIOERR_SYMLINK_FAILED;
 
     return 0;
 }
 
-static int expandFifo(struct ourfd * fd, struct cpioHeader * hdr) {
+static int expandFifo(CFD_t * cfd, struct cpioHeader * hdr) {
     struct stat sb;
 
     if (!lstat(hdr->path, &sb)) {
 	if (S_ISFIFO(sb.st_mode)) return 0;
 
 	if (unlink(hdr->path))
-	    return CPIO_UNLINK_FAILED;
+	    return CPIOERR_UNLINK_FAILED;
     }
 
     if (mkfifo(hdr->path, 0))
-	return CPIO_MKFIFO_FAILED;
+	return CPIOERR_MKFIFO_FAILED;
 
     return 0; 
 }
 
-static int expandDevice(struct ourfd * fd, struct cpioHeader * hdr) {
+static int expandDevice(CFD_t * cfd, struct cpioHeader * hdr) {
     struct stat sb;
 
     if (!lstat(hdr->path, &sb)) {
@@ -406,11 +463,11 @@ static int expandDevice(struct ourfd * fd, struct cpioHeader * hdr) {
 		(sb.st_rdev == hdr->rdev))
 	    return 0;
 	if (unlink(hdr->path))
-	    return CPIO_UNLINK_FAILED;
+	    return CPIOERR_UNLINK_FAILED;
     }
 
     if (mknod(hdr->path, hdr->mode & (~0777), hdr->rdev))
-	return CPIO_MKNOD_FAILED;
+	return CPIOERR_MKNOD_FAILED;
     
     return 0;
 }
@@ -435,13 +492,13 @@ static int createLinks(struct hardLink * li, char ** failedFile) {
 	if (!lstat(li->files[i], &sb)) {
 	    if (unlink(li->files[i])) {
 		*failedFile = strdup(li->files[i]);
-		return CPIO_UNLINK_FAILED;
+		return CPIOERR_UNLINK_FAILED;
 	    }
 	}
 
 	if (link(li->files[li->createdPath], li->files[i])) {
 	    *failedFile = strdup(li->files[i]);
-	    return CPIO_LINK_FAILED;
+	    return CPIOERR_LINK_FAILED;
 	}
 
 	free(li->files[i]);
@@ -452,25 +509,24 @@ static int createLinks(struct hardLink * li, char ** failedFile) {
     return 0;
 }
 
-static int eatBytes(struct ourfd * fd, int amount) {
+static int eatBytes(CFD_t * cfd, int amount) {
     char buf[4096];
     int bite;
    
     while (amount) {
 	bite = (amount > sizeof(buf)) ? sizeof(buf) : amount;
-	if (ourread(fd, buf, bite) != bite)
-	    return CPIO_READ_FAILED;
+	if (ourread(cfd, buf, bite) != bite)
+	    return CPIOERR_READ_FAILED;
 	amount -= bite;
     }
 
     return 0;
 }
 
-int cpioInstallArchive(gzFile stream, struct cpioFileMapping * mappings, 
+int cpioInstallArchive(CFD_t *cfd, struct cpioFileMapping * mappings, 
 		       int numMappings, cpioCallback cb, void * cbData,
 		       char ** failedFile) {
     struct cpioHeader ch;
-    struct ourfd fd;
     int rc = 0;
     int linkNum = 0;
     struct cpioFileMapping * map = NULL;
@@ -481,16 +537,16 @@ int cpioInstallArchive(gzFile stream, struct cpioFileMapping * mappings,
     struct hardLink * links = NULL;
     struct hardLink * li = NULL;
 
-    fd.fd = stream;
-    fd.pos = 0;
-
+    cfd->cpioPos = 0;
     *failedFile = NULL;
 
+    ch.path = NULL;
     do {
-	if ((rc = getNextHeader(&fd, &ch))) {
-	    fprintf(stderr, _("error %d reading header: %s\n"), rc, strerror(errno));
-	    return CPIO_BAD_HEADER;
-	    exit(1);
+	if ((rc = getNextHeader(cfd, &ch))) {
+#if 0	/* XXX this is the failure point for an unreadable rpm */
+	    fprintf(stderr, _("getNextHeader: %s\n"), cpioStrerror(rc));
+#endif
+	    return rc;
 	}
 
 	if (!strcmp(ch.path, TRAILER)) {
@@ -505,7 +561,7 @@ int cpioInstallArchive(gzFile stream, struct cpioFileMapping * mappings,
 	}
 
 	if (mappings && !map) {
-	    eatBytes(&fd, ch.size);
+	    eatBytes(cfd, ch.size);
 	} else {
 	    cpioMode = ch.mode;
 
@@ -527,7 +583,6 @@ int cpioInstallArchive(gzFile stream, struct cpioFileMapping * mappings,
 	       to create those anyway */
 
 	    if (S_ISREG(ch.mode) && ch.nlink > 1) {
-		li = links;
 		for (li = links; li; li = li->next) {
 		    if (li->inode == ch.inode && li->dev == ch.dev) break;
 		}
@@ -561,26 +616,26 @@ int cpioInstallArchive(gzFile stream, struct cpioFileMapping * mappings,
 		   listed (intead of the data being given just once. This
 		   shouldn't happen, but I've made it happen w/ buggy
 		   code, so what the heck? GNU cpio handles this well fwiw */
-		if (ch.size) eatBytes(&fd, ch.size);
+		if (ch.size) eatBytes(cfd, ch.size);
 	    } else {
 		rc = checkDirectory(ch.path);
 
 		if (!rc) {
 		    if (S_ISREG(ch.mode))
-			rc = expandRegular(&fd, &ch, cb, cbData);
+			rc = expandRegular(cfd, &ch, cb, cbData);
 		    else if (S_ISDIR(ch.mode))
 			rc = createDirectory(ch.path, 000);
 		    else if (S_ISLNK(ch.mode))
-			rc = expandSymlink(&fd, &ch);
+			rc = expandSymlink(cfd, &ch);
 		    else if (S_ISFIFO(ch.mode))
-			rc = expandFifo(&fd, &ch);
+			rc = expandFifo(cfd, &ch);
 		    else if (S_ISCHR(ch.mode) || S_ISBLK(ch.mode))
-			rc = expandDevice(&fd, &ch);
+			rc = expandDevice(cfd, &ch);
 		    else if (S_ISSOCK(ch.mode)) {
 			/* this mimicks cpio but probably isnt' right */
-			rc = expandFifo(&fd, &ch);
+			rc = expandFifo(cfd, &ch);
 		    } else {
-			rc = CPIO_INTERNAL;
+			rc = CPIOERR_UNKNOWN_FILETYPE;
 		    }
 		}
 
@@ -603,24 +658,25 @@ int cpioInstallArchive(gzFile stream, struct cpioFileMapping * mappings,
 	    }
 	}
 
-	padinfd(&fd, 4);
+	padinfd(cfd, 4);
 
 	if (!rc && cb) {
 	    cbInfo.file = ch.path;
 	    cbInfo.fileSize = ch.size;
 	    cbInfo.fileComplete = ch.size;
-	    cbInfo.bytesProcessed = fd.pos;
+	    cbInfo.bytesProcessed = cfd->cpioPos;
 	    cb(&cbInfo, cbData);
 	}
 
 	free(ch.path);
+	ch.path = NULL;
     } while (1 && !rc);
 
     li = links;
     while (li && !rc) {
 	if (li->linksLeft) {
 	    if (li->createdPath == -1)
-		rc = CPIO_INTERNAL;
+		rc = CPIOERR_MISSING_HARDLINK;
 	    else 
 		rc = createLinks(li, failedFile);
 	}
@@ -645,12 +701,12 @@ int cpioInstallArchive(gzFile stream, struct cpioFileMapping * mappings,
     return rc;
 }
 
-static int writeFile(int fd, struct stat sb, struct cpioFileMapping * map, 
+static int writeFile(CFD_t *cfd, struct stat sb, struct cpioFileMapping * map, 
 		     size_t * sizeptr, int writeData) {
     struct cpioCrcPhysicalHeader hdr;
     char buf[8192], symbuf[2048];
     dev_t num;
-    int datafd;
+    FD_t datafd;
     size_t size, amount = 0;
     int rc, olderrno;
 
@@ -671,7 +727,7 @@ static int writeFile(int fd, struct stat sb, struct cpioFileMapping * map,
 
 	amount = readlink(map->fsPath, symbuf, sizeof(symbuf));
 	if (amount <= 0) {
-	    return CPIO_READLINK_FAILED;
+	    return CPIOERR_READLINK_FAILED;
 	}
 
 	sb.st_size = amount;
@@ -694,34 +750,58 @@ static int writeFile(int fd, struct stat sb, struct cpioFileMapping * map,
     num = strlen(map->archivePath) + 1; SET_NUM_FIELD(hdr.namesize, num, buf);
     memcpy(hdr.checksum, "00000000", 8);
 
-    if ((rc = safewrite(fd, &hdr, sizeof(hdr))))
+    if ((rc = safewrite(cfd, &hdr, PHYS_HDR_SIZE)) != PHYS_HDR_SIZE)
 	return rc;
-    if ((rc = safewrite(fd, map->archivePath, num)))
+    if ((rc = safewrite(cfd, map->archivePath, num)) != num)
 	return rc;
-    size = sizeof(hdr) + num;
-    if ((rc = padoutfd(fd, &size, 4)))
+    size = PHYS_HDR_SIZE + num;
+    if ((rc = padoutfd(cfd, &size, 4)))
 	return rc;
 	
     if (writeData && S_ISREG(sb.st_mode)) {
-	/* FIXME: we should use mmap here if it's available */
-	if ((datafd = open(map->fsPath, O_RDONLY)) < 0)
-	    return CPIO_OPEN_FAILED;
+	char *b;
+#if HAVE_MMAP
+	void *mapped;
+	size_t nmapped;
+#endif
+
+	if (fdFileno(datafd = fdOpen(map->fsPath, O_RDONLY, 0)) < 0)
+	    return CPIOERR_OPEN_FAILED;
+
+#if HAVE_MMAP
+	nmapped = 0;
+	mapped = mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, fdFileno(datafd), 0);
+	if (mapped != (void *)-1) {
+	    b = (char *)mapped;
+	    nmapped = sb.st_size;
+	} else
+#endif
+	{
+	    b = buf;
+	}
 
 	size += sb.st_size;
 
 	while (sb.st_size) {
-	    amount = read(datafd, buf, 
+#if HAVE_MMAP
+	  if (mapped != (void *)-1) {
+	    amount = nmapped;
+	  } else
+#endif
+	  {
+	    amount = fdRead(datafd, b, 
 		     sb.st_size > sizeof(buf) ? sizeof(buf) : sb.st_size);
 	    if (amount <= 0) {
 		olderrno = errno;
-		close(datafd);
+		fdClose(datafd);
 		errno = olderrno;
-		return CPIO_READ_FAILED;
+		return CPIOERR_READ_FAILED;
 	    }
+	  }
 
-	    if ((rc = safewrite(fd, buf, amount))) {
+	    if ((rc = safewrite(cfd, b, amount)) != amount) {
 		olderrno = errno;
-		close(datafd);
+		fdClose(datafd);
 		errno = olderrno;
 		return rc;
 	    }
@@ -729,15 +809,21 @@ static int writeFile(int fd, struct stat sb, struct cpioFileMapping * map,
 	    sb.st_size -= amount;
 	}
 
-	close(datafd);
+#if HAVE_MMAP
+	if (mapped != (void *)-1) {
+	    munmap(mapped, nmapped);
+	}
+#endif
+
+	fdClose(datafd);
     } else if (writeData && S_ISLNK(sb.st_mode)) {
-	if ((rc = safewrite(fd, symbuf, amount)))
+	if ((rc = safewrite(cfd, symbuf, amount)) != amount)
 	    return rc;
 	size += amount;
     }
 
     /* this is a noop for most file types */
-    if ((rc = padoutfd(fd, &size, 4)))
+    if ((rc = padoutfd(cfd, &size, 4)))
 	return rc;
 
     *sizeptr = size;
@@ -745,7 +831,7 @@ static int writeFile(int fd, struct stat sb, struct cpioFileMapping * map,
     return 0;
 }
 
-static int writeLinkedFile(int fd, struct hardLink * hlink, 
+static int writeLinkedFile(CFD_t *cfd, struct hardLink * hlink, 
 			   struct cpioFileMapping * mappings,
 			   cpioCallback cb, void * cbData,
 			   size_t * sizeptr,
@@ -757,7 +843,7 @@ static int writeLinkedFile(int fd, struct hardLink * hlink,
     *sizeptr = 0;
 
     for (i = hlink->nlink - 1; i > hlink->linksLeft; i--) {
-	if ((rc = writeFile(fd, hlink->sb, mappings + hlink->fileMaps[i], 
+	if ((rc = writeFile(cfd, hlink->sb, mappings + hlink->fileMaps[i], 
 			    &size, 0))) {
 	    if (failedFile) *failedFile = 
 		    mappings[hlink->fileMaps[i]].fsPath;
@@ -772,7 +858,7 @@ static int writeLinkedFile(int fd, struct hardLink * hlink,
 	}
     }
 
-    if ((rc = writeFile(fd, hlink->sb, 
+    if ((rc = writeFile(cfd, hlink->sb, 
 			mappings + hlink->fileMaps[hlink->linksLeft], 
 			&size, 1))) {
 	if (failedFile) 
@@ -790,7 +876,7 @@ static int writeLinkedFile(int fd, struct hardLink * hlink,
     return 0;
 }
 
-int cpioBuildArchive(int fd, struct cpioFileMapping * mappings, 
+int cpioBuildArchive(CFD_t *cfd, struct cpioFileMapping * mappings, 
 		     int numMappings, cpioCallback cb, void * cbData,
 		     unsigned int * archiveSize, char ** failedFile) {
     size_t size, totalsize = 0;
@@ -799,7 +885,9 @@ int cpioBuildArchive(int fd, struct cpioFileMapping * mappings,
     struct cpioCallbackInfo cbinfo;
     struct cpioCrcPhysicalHeader hdr;
     struct stat sb;
+/*@-fullinitblock@*/
     struct hardLink hlinkList = { NULL };
+/*@=fullinitblock@*/
     struct hardLink * hlink, * parent;
 
     hlinkList.next = NULL;
@@ -812,7 +900,7 @@ int cpioBuildArchive(int fd, struct cpioFileMapping * mappings,
 
 	if (rc) {
 	    if (failedFile) *failedFile = mappings[i].fsPath;
-	    return CPIO_STAT_FAILED;
+	    return CPIOERR_STAT_FAILED;
 	}
 
 	if (!S_ISDIR(sb.st_mode) && sb.st_nlink > 1) {
@@ -824,7 +912,7 @@ int cpioBuildArchive(int fd, struct cpioFileMapping * mappings,
 		hlink = malloc(sizeof(*hlink));
 		hlink->next = hlinkList.next;
 		hlinkList.next = hlink;
-		hlink->sb = sb;
+		hlink->sb = sb;		/* structure assignment */
 		hlink->dev = sb.st_dev;
 		hlink->inode = sb.st_ino;
 		hlink->nlink = sb.st_nlink;
@@ -836,7 +924,7 @@ int cpioBuildArchive(int fd, struct cpioFileMapping * mappings,
 	    hlink->fileMaps[--hlink->linksLeft] = i;
 
 	    if (!hlink->linksLeft) {
-		if ((rc = writeLinkedFile(fd, hlink, mappings, cb, cbData,
+		if ((rc = writeLinkedFile(cfd, hlink, mappings, cb, cbData,
 			 		  &size, failedFile)))
 		    return rc;
 
@@ -850,7 +938,7 @@ int cpioBuildArchive(int fd, struct cpioFileMapping * mappings,
 		free(hlink);
 	    }
 	} else {
-	    if ((rc = writeFile(fd, sb, mappings + i, &size, 1))) {
+	    if ((rc = writeFile(cfd, sb, mappings + i, &size, 1))) {
 		if (failedFile) *failedFile = mappings[i].fsPath;
 		return rc;
 	    }
@@ -866,7 +954,7 @@ int cpioBuildArchive(int fd, struct cpioFileMapping * mappings,
 
     hlink = hlinkList.next;
     while (hlink) {
-	if ((rc = writeLinkedFile(fd, hlink, mappings, cb, cbData,
+	if ((rc = writeLinkedFile(cfd, hlink, mappings, cb, cbData,
 				  &size, failedFile)))
 	    return rc;
 	free(hlink->fileMaps);
@@ -877,23 +965,75 @@ int cpioBuildArchive(int fd, struct cpioFileMapping * mappings,
 	totalsize += size;
     }
 
-    memset(&hdr, '0', sizeof(hdr));
+    memset(&hdr, '0', PHYS_HDR_SIZE);
     memcpy(hdr.magic, CPIO_NEWC_MAGIC, sizeof(hdr.magic));
     memcpy(hdr.nlink, "00000001", 8);
     memcpy(hdr.namesize, "0000000b", 8);
-    if ((rc = safewrite(fd, &hdr, sizeof(hdr))))
+    if ((rc = safewrite(cfd, &hdr, PHYS_HDR_SIZE)) != PHYS_HDR_SIZE)
 	return rc;
-    if ((rc = safewrite(fd, "TRAILER!!!", 11)))
+    if ((rc = safewrite(cfd, "TRAILER!!!", 11)) != 11)
 	return rc;
-    totalsize += sizeof(hdr) + 11;
+    totalsize += PHYS_HDR_SIZE + 11;
 
     /* GNU cpio pads to 512 bytes here, but we don't. I'm not sure if
        it matters or not */
     
-    if ((rc = padoutfd(fd, &totalsize, 4)))
+    if ((rc = padoutfd(cfd, &totalsize, 4)))
 	return rc;
 
     if (archiveSize) *archiveSize = totalsize;
 
     return 0;
 }
+
+const char * cpioStrerror(int rc)
+{
+    static char msg[256];
+    char *s;
+    int l, myerrno = errno;
+
+    strcpy(msg, "cpio: ");
+    switch (rc) {
+    default:
+	s = msg + strlen(msg);
+	sprintf(s, _("(error 0x%x)"), rc);
+	s = NULL;
+	break;
+    case CPIOERR_BAD_MAGIC:	s = _("Bad magic");		break;
+    case CPIOERR_BAD_HEADER:	s = _("Bad/unreadable  header");break;
+
+    case CPIOERR_OPEN_FAILED:	s = "open";	break;
+    case CPIOERR_CHMOD_FAILED:	s = "chmod";	break;
+    case CPIOERR_CHOWN_FAILED:	s = "chown";	break;
+    case CPIOERR_WRITE_FAILED:	s = "write";	break;
+    case CPIOERR_UTIME_FAILED:	s = "utime";	break;
+    case CPIOERR_UNLINK_FAILED:	s = "unlink";	break;
+    case CPIOERR_SYMLINK_FAILED: s = "symlink";	break;
+    case CPIOERR_STAT_FAILED:	s = "stat";	break;
+    case CPIOERR_MKDIR_FAILED:	s = "mkdir";	break;
+    case CPIOERR_MKNOD_FAILED:	s = "mknod";	break;
+    case CPIOERR_MKFIFO_FAILED:	s = "mkfifo";	break;
+    case CPIOERR_LINK_FAILED:	s = "link";	break;
+    case CPIOERR_READLINK_FAILED: s = "readlink";	break;
+    case CPIOERR_READ_FAILED:	s = "read";	break;
+    case CPIOERR_COPY_FAILED:	s = "copy";	break;
+
+    case CPIOERR_HDR_SIZE:	s = _("Header size too big");	break;
+    case CPIOERR_UNKNOWN_FILETYPE: s = _("Unknown file type");	break;
+    case CPIOERR_MISSING_HARDLINK: s = _("Missing hard link");	break;
+    case CPIOERR_INTERNAL:	s = _("Internal error");	break;
+    }
+
+    l = sizeof(msg) - strlen(msg) - 1;
+    if (s != NULL) {
+	if (l > 0) strncat(msg, s, l);
+	l -= strlen(s);
+    }
+    if (rc & CPIOERR_CHECK_ERRNO) {
+	s = _(" failed - ");
+	if (l > 0) strncat(msg, s, l);
+	l -= strlen(s);
+	if (l > 0) strncat(msg, strerror(myerrno), l);
+    }
+    return msg;
+}
