@@ -1,34 +1,40 @@
-#include "config.h"
-#include "miscfn.h"
+#include "system.h"
 
-#ifdef HAVE_DB_185_H
-#include <db_185.h>
-#else
-#include <db.h>
-#endif
-#include <errno.h>
-#include <stdlib.h>
-#include <string.h>
-
-#include "dbindex.h"
-#include "intl.h"
 #include "rpmlib.h"
 
-dbiIndex * dbiOpenIndex(char * filename, int flags, int perms) {
-    dbiIndex * db;
+int dbiIndexSetCount(dbiIndexSet set) {
+    return set.count;
+}
+
+/* structure return */
+dbiIndexRecord dbiReturnIndexRecordInstance(unsigned int recOffset, unsigned int fileNumber) {
+    dbiIndexRecord rec;
+    rec.recOffset = recOffset;
+    rec.fileNumber = fileNumber;
+    return rec;
+}
+
+unsigned int dbiIndexRecordOffset(dbiIndexSet set, int recno) {
+    return set.recs[recno].recOffset;
+}
+
+unsigned int dbiIndexRecordFileNumber(dbiIndexSet set, int recno) {
+    return set.recs[recno].fileNumber;
+}
+
+dbiIndex * dbiOpenIndex(const char * filename, int flags, int perms, DBTYPE type) {
+    dbiIndex * dbi;
         
-    db = malloc(sizeof(*db));
-    db->indexname = strdup(filename);
-    db->db = dbopen(filename, flags, perms, DB_HASH, NULL);
-    if (!db->db) {
-	free(db->indexname);
-	free(db);
-	rpmError(RPMERR_DBOPEN, _("cannot open file %s: "), filename, 
+    dbi = malloc(sizeof(*dbi));
+    dbi->db = dbopen(filename, flags, perms, type, NULL);
+    if (!dbi->db) {
+	free(dbi);
+	rpmError(RPMERR_DBOPEN, _("cannot open file %s: %s"), filename, 
 			      strerror(errno));
 	return NULL;
     }
-
-    return db;
+    dbi->indexname = strdup(filename);
+    return dbi;
 }
 
 void dbiCloseIndex(dbiIndex * dbi) {
@@ -41,11 +47,29 @@ void dbiSyncIndex(dbiIndex * dbi) {
     dbi->db->sync(dbi->db, 0);
 }
 
-int dbiSearchIndex(dbiIndex * dbi, char * str, dbiIndexSet * set) {
+int dbiGetFirstKey(dbiIndex * dbi, const char ** keyp) {
     DBT key, data;
     int rc;
 
-    key.data = str;
+    rc = dbi->db->seq(dbi->db, &key, &data, R_FIRST);
+    if (rc) {
+	return 1;
+    }
+
+    {	char *k = malloc(key.size + 1);
+	memcpy(k, key.data, key.size);
+	k[key.size] = '\0';
+	*keyp = k;
+    }
+
+    return 0;
+}
+
+int dbiSearchIndex(dbiIndex * dbi, const char * str, dbiIndexSet * set) {
+    DBT key, data;
+    int rc;
+
+    key.data = (void *)str;
     key.size = strlen(str);
 
     rc = dbi->db->get(dbi->db, &key, &data, 0);
@@ -57,19 +81,18 @@ int dbiSearchIndex(dbiIndex * dbi, char * str, dbiIndexSet * set) {
 	return 1;
     } 
 
-    set->recs = data.data;
     set->recs = malloc(data.size);
     memcpy(set->recs, data.data, data.size);
     set->count = data.size / sizeof(dbiIndexRecord);
     return 0;
 }
 
-int dbiUpdateIndex(dbiIndex * dbi, char * str, dbiIndexSet * set) {
+int dbiUpdateIndex(dbiIndex * dbi, const char * str, dbiIndexSet * set) {
    /* 0 on success */
     DBT key, data;
     int rc;
 
-    key.data = str;
+    key.data = (void *)str;
     key.size = strlen(str);
 
     if (set->count) {
@@ -107,15 +130,17 @@ int dbiAppendIndexRecord(dbiIndexSet * set, dbiIndexRecord rec) {
     return 0;
 }
 
+/* structure return */
 dbiIndexSet dbiCreateIndexRecord(void) {
     dbiIndexSet set;
 
+    set.recs = NULL;
     set.count = 0;
     return set;
 }
     
 void dbiFreeIndexRecord(dbiIndexSet set) {
-    free(set.recs);
+    if (set.recs) free(set.recs);
 }
 
 /* returns 1 on failure */
