@@ -1,90 +1,47 @@
-#include "config.h"
-#include "miscfn.h"
+#include "system.h"
 
-#if HAVE_ALLOCA_H
-# include <alloca.h>
-#endif
-
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
+#include "rpmlib.h"
 
 #include "depends.h"
-#include "intl.h"
 #include "misc.h"
-#include "rpmlib.h"
-#include "messages.h"
 
-struct availablePackage {
-    Header h;
-    char ** provides;
-    char ** files;
-    char * name, * version, * release;
-    int serial, hasSerial, providesCount, filesCount;
-    void * key;
-} ;
-
-enum indexEntryType { IET_NAME, IET_PROVIDES, IET_FILE };
-
-struct availableIndexEntry {
-    struct availablePackage * package;
-    char * entry;
-    enum indexEntryType type;
-} ;
-
-struct availableIndex {
-    struct availableIndexEntry * index ;
-    int size;
-} ;
-
-struct availableList {
-    struct availablePackage * list;
-    struct availableIndex index;
-    int size, alloced;
-};
-
-struct rpmDependencyCheck {
-    rpmdb db;					/* may be NULL */
-    int * removedPackages;
-    int numRemovedPackages, allocedRemovedPackages;
-    struct availableList addedPackages, availablePackages;
-};
-
-struct problemsSet {
-    struct rpmDependencyConflict * problems;
-    int num;
-    int alloced;
+struct orderListIndex {
+    int alIndex;
+    int orIndex;
 };
 
+static int orderListIndexCmp(const void * one, const void * two);
 static void alMakeIndex(struct availableList * al);
 static void alCreate(struct availableList * al);
 static void alFreeIndex(struct availableList * al);
 static void alFree(struct availableList * al);
-static void alAddPackage(struct availableList * al, Header h, void * key);
-
+static struct availablePackage * alAddPackage(struct availableList * al, 
+					      Header h, const void * key,
+			 		      FD_t fd, rpmRelocation * relocs);
 static int intcmp(const void * a, const void *b);
 static int indexcmp(const void * a, const void *b);
-static int unsatisfiedDepend(rpmDependencies rpmdep, char * reqName, 
+static int unsatisfiedDepend(rpmTransactionSet rpmdep, char * reqName, 
 			     char * reqVersion, int reqFlags,
 			     struct availablePackage ** suggestion);
-static int checkDependentPackages(rpmDependencies rpmdep, 
+static int checkDependentPackages(rpmTransactionSet rpmdep, 
 			    struct problemsSet * psp, char * key);
-static int checkPackageDeps(rpmDependencies rpmdep, struct problemsSet * psp,
+static int checkPackageDeps(rpmTransactionSet rpmdep, struct problemsSet * psp,
 			Header h, const char * requirement);
-static int dbrecMatchesDepFlags(rpmDependencies rpmdep, int recOffset, 
+static int dbrecMatchesDepFlags(rpmTransactionSet rpmdep, int recOffset, 
 			        char * reqVersion, int reqFlags);
 struct availablePackage * alSatisfiesDepend(struct availableList * al, 
 					    char * reqName, char * reqVersion, 
 					    int reqFlags);
-static int checkDependentConflicts(rpmDependencies rpmdep, 
+static int checkDependentConflicts(rpmTransactionSet rpmdep, 
 			    struct problemsSet * psp, char * package);
-static int checkPackageSet(rpmDependencies rpmdep, struct problemsSet * psp, 
+static int checkPackageSet(rpmTransactionSet rpmdep, struct problemsSet * psp, 
 			    char * package, dbiIndexSet * matches);
-static int addOrderedPack(rpmDependencies rpmdep, 
+static int addOrderedPack(rpmTransactionSet rpmdep, 
 			struct availablePackage * package,
-			void ** ordering, int * orderNumPtr, 
+			int * ordering, int * orderNumPtr, 
 			int * selected, int selectionClass,
 			int satisfyDepends, char ** errorStack);
+static void removePackage(rpmTransactionSet rpmdep, int dboffset, int depends);
 
 static void alCreate(struct availableList * al) {
     al->list = malloc(sizeof(*al->list) * 5);
@@ -97,7 +54,8 @@ static void alCreate(struct availableList * al) {
 
 static void alFreeIndex(struct availableList * al) {
     if (al->index.size) {
-	free(al->index.index);
+	if (al->index.index)
+		free(al->index.index);
 	al->index.index = NULL;
 	al->index.size = 0;
     }
@@ -105,20 +63,37 @@ static void alFreeIndex(struct availableList * al) {
 
 static void alFree(struct availableList * al) {
     int i;
+    rpmRelocation * r;
 
     for (i = 0; i < al->size; i++) {
 	if (al->list[i].provides)
 	    free(al->list[i].provides);
 	if (al->list[i].files)
 	    free(al->list[i].files);
+	headerFree(al->list[i].h);
+
+	if (al->list[i].relocs) {
+	    r = al->list[i].relocs;
+	    while (r->oldPath || r->newPath) {
+		if (r->oldPath) free(r->oldPath);
+		if (r->newPath) free(r->newPath);
+		r++;
+	    }
+
+	    free(al->list[i].relocs);
+	}
     }
 
     if (al->alloced) free(al->list);
     alFreeIndex(al);
 }
 
-static void alAddPackage(struct availableList * al, Header h, void * key) {
+static struct availablePackage * alAddPackage(struct availableList * al, 
+					      Header h, const void * key,
+			 		      FD_t fd, rpmRelocation * relocs) {
     struct availablePackage * p;
+    rpmRelocation * r;
+    int i;
 
     if (al->size == al->alloced) {
 	al->alloced += 5;
@@ -126,12 +101,12 @@ static void alAddPackage(struct availableList * al, Header h, void * key) {
     }
 
     p = al->list + al->size++;
-    p->h = h;
+    p->h = headerLink(h);
 
     headerGetEntry(p->h, RPMTAG_NAME, NULL, (void **) &p->name, NULL);
     headerGetEntry(p->h, RPMTAG_VERSION, NULL, (void **) &p->version, NULL);
     headerGetEntry(p->h, RPMTAG_RELEASE, NULL, (void **) &p->release, NULL);
-    p->hasSerial = headerGetEntry(h, RPMTAG_SERIAL, NULL, (void **) &p->serial, 
+    p->hasEpoch = headerGetEntry(h, RPMTAG_EPOCH, NULL, (void **) &p->epoch, 
 				  NULL);
 
     if (!headerGetEntry(h, RPMTAG_PROVIDES, NULL, (void **) &p->provides,
@@ -146,9 +121,34 @@ static void alAddPackage(struct availableList * al, Header h, void * key) {
 	p->files = NULL;
     }
 
+    /* We don't use these entries (and rpm >= 2 never have) and they are 
+       pretty misleading. Let's just get rid of them so they don't confuse
+       anyone. */
+    if (headerIsEntry(h, RPMTAG_FILEUSERNAME))
+	headerRemoveEntry(h, RPMTAG_FILEUIDS);
+    if (headerIsEntry(h, RPMTAG_FILEGROUPNAME))
+	headerRemoveEntry(h, RPMTAG_FILEGIDS);
+    
     p->key = key;
+    p->fd = fd;
 
+    if (relocs) {
+	for (i = 0, r = relocs; r->oldPath || r->newPath; i++, r++);
+	p->relocs = malloc(sizeof(*p->relocs) * (i + 1));
+
+	for (i = 0, r = relocs; r->oldPath || r->newPath; i++, r++) {
+	    p->relocs[i].oldPath = r->oldPath ? strdup(r->oldPath) : NULL;
+	    p->relocs[i].newPath = r->newPath ? strdup(r->newPath) : NULL;
+	}
+	p->relocs[i].oldPath = NULL;
+	p->relocs[i].newPath = NULL;
+    } else {
+	p->relocs = NULL;
+    }
+    
     alFreeIndex(al);
+
+    return p;
 }
 
 static void alMakeIndex(struct availableList * al) {
@@ -232,49 +232,108 @@ int intcmp(const void * a, const void *b) {
     return 1;
 }
 
-rpmDependencies rpmdepDependencies(rpmdb db) {
-    rpmDependencies rpmdep;
+rpmTransactionSet rpmtransCreateSet(rpmdb db, const char * root) {
+    rpmTransactionSet rpmdep;
+    int rootLength;
+
+    if (!root) root = "";
 
     rpmdep = malloc(sizeof(*rpmdep));
     rpmdep->db = db;
+    rpmdep->scriptFd = NULL;
     rpmdep->numRemovedPackages = 0;
     rpmdep->allocedRemovedPackages = 5;
     rpmdep->removedPackages = malloc(sizeof(int) * 
 				     rpmdep->allocedRemovedPackages);
 
+    /* This canonicalizes the root */
+    rootLength = strlen(root);
+    if (root && root[rootLength] == '/') {
+	char * newRootdir;
+
+	newRootdir = alloca(rootLength + 2);
+	strcpy(newRootdir, root);
+	newRootdir[rootLength++] = '/';
+	newRootdir[rootLength] = '\0';
+	root = newRootdir;
+    }
+
+    rpmdep->root = strdup(root);
+
     alCreate(&rpmdep->addedPackages);
     alCreate(&rpmdep->availablePackages);
 
+    rpmdep->orderAlloced = 5;
+    rpmdep->orderCount = 0;
+    rpmdep->order = malloc(sizeof(*rpmdep->order) * rpmdep->orderAlloced);
+
     return rpmdep;
 }
 
-void rpmdepUpgradePackage(rpmDependencies rpmdep, Header h, void * key) {
+int rpmtransAddPackage(rpmTransactionSet rpmdep, Header h, FD_t fd,
+			const void * key, int upgrade, rpmRelocation * relocs) {
     /* this is an install followed by uninstalls */
     dbiIndexSet matches;
     char * name;
     int count, i, j;
     char ** obsoletes;
+    int alNum;
+    int * caps;
 
-    alAddPackage(&rpmdep->addedPackages, h, key);
+    /* XXX binary rpms always have RPMTAG_SOURCERPM, source rpms do not */
+    if (headerIsEntry(h, RPMTAG_SOURCEPACKAGE))
+	return 1;
 
-    if (!rpmdep->db) return;
+    /* Make sure we've implemented all of the capabilities we need */
+    if (headerGetEntry(h, RPMTAG_CAPABILITY, NULL, (void **)&caps, &count)) {
+	if (count != 1 || *caps) {
+	    return 2;
+	}
+    }
 
-    headerGetEntry(h, RPMTAG_NAME, NULL, (void *) &name, &count);
+    /* FIXME: handling upgrades like this is *almost* okay. It doesn't
+       check to make sure we're upgrading to a newer version, and it
+       makes it difficult to generate a return code based on the number of
+       packages which failed. */
+   
+    if (rpmdep->orderCount == rpmdep->orderAlloced) {
+	rpmdep->orderAlloced += 5;
+	rpmdep->order = realloc(rpmdep->order, 
+		sizeof(*rpmdep->order) * rpmdep->orderAlloced);
+    }
+    rpmdep->order[rpmdep->orderCount].type = TR_ADDED;
+    alNum = alAddPackage(&rpmdep->addedPackages, h, key, fd, relocs) - 
+    		rpmdep->addedPackages.list;
+    rpmdep->order[rpmdep->orderCount++].u.addedIndex = alNum;
+
+    if (!upgrade || rpmdep->db == NULL) return 0;
+
+    headerGetEntry(h, RPMTAG_NAME, NULL, (void **) &name, &count);
 
     if (!rpmdbFindPackage(rpmdep->db, name, &matches))  {
-	for (i = 0; i < matches.count; i++) {
-	    rpmdepRemovePackage(rpmdep, matches.recs[i].recOffset);
+	Header h2;
+
+	for (i = 0; i < dbiIndexSetCount(matches); i++) {
+	    h2 = rpmdbGetRecord(rpmdep->db, dbiIndexRecordOffset(matches, i));
+	    if (h2) {
+		if (rpmVersionCompare(h, h2)) {
+		    removePackage(rpmdep, dbiIndexRecordOffset(matches, i), 
+				  alNum);
+		}
+		headerFree(h2);
+	    }
 	}
 
 	dbiFreeIndexRecord(matches);
     }
 
-    if (headerGetEntry(h, RPMTAG_OBSOLETES, NULL, (void *) &obsoletes, 
+    if (headerGetEntry(h, RPMTAG_OBSOLETES, NULL, (void **) &obsoletes, 
 			&count)) {
 	for (j = 0; j < count; j++) {
 	    if (!rpmdbFindPackage(rpmdep->db, obsoletes[j], &matches))  {
-		for (i = 0; i < matches.count; i++) {
-		    rpmdepRemovePackage(rpmdep, matches.recs[i].recOffset);
+		for (i = 0; i < dbiIndexSetCount(matches); i++) {
+		    removePackage(rpmdep, dbiIndexRecordOffset(matches, i),
+				  alNum);
 		}
 
 		dbiFreeIndexRecord(matches);
@@ -283,17 +342,15 @@ void rpmdepUpgradePackage(rpmDependencies rpmdep, Header h, void * key) {
 
 	free(obsoletes);
     }
+
+    return 0;
 }
 
-void rpmdepAddPackage(rpmDependencies rpmdep, Header h, void * key) {
-    alAddPackage(&rpmdep->addedPackages, h, key);
+void rpmtransAvailablePackage(rpmTransactionSet rpmdep, Header h, void * key) {
+    alAddPackage(&rpmdep->availablePackages, h, key, NULL, NULL);
 }
 
-void rpmdepAvailablePackage(rpmDependencies rpmdep, Header h, void * key) {
-    alAddPackage(&rpmdep->availablePackages, h, key);
-}
-
-void rpmdepRemovePackage(rpmDependencies rpmdep, int dboffset) {
+static void removePackage(rpmTransactionSet rpmdep, int dboffset, int depends) {
     if (rpmdep->numRemovedPackages == rpmdep->allocedRemovedPackages) {
 	rpmdep->allocedRemovedPackages += 5;
 	rpmdep->removedPackages = realloc(rpmdep->removedPackages,
@@ -301,12 +358,27 @@ void rpmdepRemovePackage(rpmDependencies rpmdep, int dboffset) {
     }
 
     rpmdep->removedPackages[rpmdep->numRemovedPackages++] = dboffset;
+
+    if (rpmdep->orderCount == rpmdep->orderAlloced) {
+	rpmdep->orderAlloced += 5;
+	rpmdep->order = realloc(rpmdep->order, 
+		sizeof(*rpmdep->order) * rpmdep->orderAlloced);
+    }
+
+    rpmdep->order[rpmdep->orderCount].type = TR_REMOVED;
+    rpmdep->order[rpmdep->orderCount].u.removed.dboffset = dboffset;
+    rpmdep->order[rpmdep->orderCount++].u.removed.dependsOnIndex = depends;
 }
 
-void rpmdepDone(rpmDependencies rpmdep) {
+void rpmtransRemovePackage(rpmTransactionSet rpmdep, int dboffset) {
+    removePackage(rpmdep, dboffset, -1);
+}
+
+void rpmtransFree(rpmTransactionSet rpmdep) {
     alFree(&rpmdep->addedPackages);
     alFree(&rpmdep->availablePackages);
     free(rpmdep->removedPackages);
+    free(rpmdep->root);
 
     free(rpmdep);
 }
@@ -327,7 +399,7 @@ void rpmdepFreeConflicts(struct rpmDependencyConflict * conflicts, int
     free(conflicts);
 }
 
-int rpmdepCheck(rpmDependencies rpmdep, 
+int rpmdepCheck(rpmTransactionSet rpmdep, 
 		struct rpmDependencyConflict ** conflicts, int * numConflicts) {
     struct availablePackage * p;
     int i, j;
@@ -372,13 +444,14 @@ int rpmdepCheck(rpmDependencies rpmdep,
 		    return 1;
 		}
 	    }
+	    free(provides);
 	}
     }
 
     /* now look at the removed packages and make sure they aren't critical */
     for (i = 0; i < rpmdep->numRemovedPackages; i++) {
 	h = rpmdbGetRecord(rpmdep->db, rpmdep->removedPackages[i]);
-	if (!h) {
+	if (h == NULL) {
 	    rpmError(RPMERR_DBCORRUPT, 
 			_("cannot read header at %d for dependency check"),
 		        rpmdep->removedPackages[i]);
@@ -437,7 +510,7 @@ int rpmdepCheck(rpmDependencies rpmdep,
 
 /* 2 == error */
 /* 1 == dependency not satisfied */
-static int unsatisfiedDepend(rpmDependencies rpmdep, char * reqName, 
+static int unsatisfiedDepend(rpmTransactionSet rpmdep, char * reqName, 
 			     char * reqVersion, int reqFlags, 
 			     struct availablePackage ** suggestion) {
     dbiIndexSet matches;
@@ -463,12 +536,13 @@ static int unsatisfiedDepend(rpmDependencies rpmdep, char * reqName,
 			  reqFlags))
 	return 0;
 
-    if (rpmdep->db) {
+    if (rpmdep->db != NULL) {
 	if (*reqName == '/') {
 	    /* reqFlags better be 0! */
 	    if (!rpmdbFindByFile(rpmdep->db, reqName, &matches)) {
-		for (i = 0; i < matches.count; i++) {
-		    if (bsearch(&matches.recs[i].recOffset, 
+		for (i = 0; i < dbiIndexSetCount(matches); i++) {
+		    unsigned int recOffset = dbiIndexRecordOffset(matches, i);
+		    if (bsearch(&recOffset, 
 				rpmdep->removedPackages, 
 				rpmdep->numRemovedPackages, 
 				sizeof(int), intcmp)) 
@@ -477,41 +551,43 @@ static int unsatisfiedDepend(rpmDependencies rpmdep, char * reqName,
 		}
 
 		dbiFreeIndexRecord(matches);
-		if (i < matches.count) return 0;
+		if (i < dbiIndexSetCount(matches)) return 0;
 	    }
-	} else {
-	    if (!reqFlags && !rpmdbFindByProvides(rpmdep->db, reqName, 
-						  &matches)) {
-		for (i = 0; i < matches.count; i++) {
-		    if (bsearch(&matches.recs[i].recOffset, 
-				rpmdep->removedPackages, 
-				rpmdep->numRemovedPackages, 
-				sizeof(int), intcmp)) 
-			continue;
+	}
+
+	if (!reqFlags && !rpmdbFindByProvides(rpmdep->db, reqName, 
+					      &matches)) {
+	    for (i = 0; i < dbiIndexSetCount(matches); i++) {
+		unsigned int recOffset = dbiIndexRecordOffset(matches, i);
+		if (bsearch(&recOffset,
+			    rpmdep->removedPackages, 
+			    rpmdep->numRemovedPackages, 
+			    sizeof(int), intcmp)) 
+		    continue;
+		break;
+	    }
+
+	    dbiFreeIndexRecord(matches);
+	    if (i < dbiIndexSetCount(matches)) return 0;
+	}
+
+	if (!rpmdbFindPackage(rpmdep->db, reqName, &matches)) {
+	    for (i = 0; i < dbiIndexSetCount(matches); i++) {
+		unsigned int recOffset = dbiIndexRecordOffset(matches, i);
+		if (bsearch(&recOffset,
+			    rpmdep->removedPackages, 
+			    rpmdep->numRemovedPackages, 
+			    sizeof(int), intcmp)) 
+		    continue;
+
+		if (dbrecMatchesDepFlags(rpmdep, recOffset, 
+					 reqVersion, reqFlags)) {
 		    break;
 		}
-
-		dbiFreeIndexRecord(matches);
-		if (i < matches.count) return 0;
 	    }
 
-	    if (!rpmdbFindPackage(rpmdep->db, reqName, &matches)) {
-		for (i = 0; i < matches.count; i++) {
-		    if (bsearch(&matches.recs[i].recOffset, 
-				rpmdep->removedPackages, 
-				rpmdep->numRemovedPackages, 
-				sizeof(int), intcmp)) 
-			continue;
-
-		    if (dbrecMatchesDepFlags(rpmdep, matches.recs[i].recOffset, 
-					     reqVersion, reqFlags)) {
-			break;
-		    }
-		}
-
-		dbiFreeIndexRecord(matches);
-		if (i < matches.count) return 0;
-	    }
+	    dbiFreeIndexRecord(matches);
+	    if (i < dbiIndexSetCount(matches)) return 0;
 	}
     }
 
@@ -522,18 +598,19 @@ static int unsatisfiedDepend(rpmDependencies rpmdep, char * reqName,
     return 1;
 }
 
-static int checkPackageSet(rpmDependencies rpmdep, struct problemsSet * psp, 
+static int checkPackageSet(rpmTransactionSet rpmdep, struct problemsSet * psp, 
 			    char * package, dbiIndexSet * matches) {
     int i;
     Header h;
 
     for (i = 0; i < matches->count; i++) {
-	if (bsearch(&matches->recs[i].recOffset, rpmdep->removedPackages, 
+	unsigned int recOffset = dbiIndexRecordOffset(*matches, i);
+	if (bsearch(&recOffset, rpmdep->removedPackages, 
 		    rpmdep->numRemovedPackages, sizeof(int), intcmp)) 
 	    continue;
 
-	h = rpmdbGetRecord(rpmdep->db, matches->recs[i].recOffset);
-	if (!h) {
+	h = rpmdbGetRecord(rpmdep->db, recOffset);
+	if (h == NULL) {
 	    rpmError(RPMERR_DBCORRUPT, 
                      _("cannot read header at %d for dependency check"),
 		     rpmdep->removedPackages[i]);
@@ -551,7 +628,7 @@ static int checkPackageSet(rpmDependencies rpmdep, struct problemsSet * psp,
     return 0;
 }
 
-static int checkDependentPackages(rpmDependencies rpmdep, 
+static int checkDependentPackages(rpmTransactionSet rpmdep, 
 			    struct problemsSet * psp, char * key) {
     dbiIndexSet matches;
     int rc;
@@ -566,23 +643,24 @@ static int checkDependentPackages(rpmDependencies rpmdep,
     return rc;
 }
 
-static int checkDependentConflicts(rpmDependencies rpmdep, 
+static int checkDependentConflicts(rpmTransactionSet rpmdep, 
 			    struct problemsSet * psp, char * package) {
     dbiIndexSet matches;
     int rc;
 
-    if (!rpmdep->db) return 0;
+    if (rpmdep->db == NULL) return 0;
 
     if (rpmdbFindByConflicts(rpmdep->db, package, &matches)) {
 	return 0;
     }
 
     rc = checkPackageSet(rpmdep, psp, package, &matches);
+    dbiFreeIndexRecord(matches);
 
     return rc;
 }
 
-static int checkPackageDeps(rpmDependencies rpmdep, struct problemsSet * psp,
+static int checkPackageDeps(rpmTransactionSet rpmdep, struct problemsSet * psp,
 			Header h, const char * requirement) {
     char ** requires, ** requiresVersion;
     char * name, * version, * release;
@@ -636,7 +714,7 @@ static int checkPackageDeps(rpmDependencies rpmdep, struct problemsSet * psp,
 		psp->problems = realloc(psp->problems, sizeof(*psp->problems) * 
 			    psp->alloced);
 	    }
-	    psp->problems[psp->num].byHeader = headerCopy(h);
+	    psp->problems[psp->num].byHeader = headerLink(h);
 	    psp->problems[psp->num].byName = strdup(name);
 	    psp->problems[psp->num].byVersion = strdup(version);
 	    psp->problems[psp->num].byRelease = strdup(release);
@@ -679,7 +757,7 @@ static int checkPackageDeps(rpmDependencies rpmdep, struct problemsSet * psp,
 		psp->problems = realloc(psp->problems, sizeof(*psp->problems) * 
 			    psp->alloced);
 	    }
-	    psp->problems[psp->num].byHeader = headerCopy(h);
+	    psp->problems[psp->num].byHeader = headerLink(h);
 	    psp->problems[psp->num].byName = strdup(name);
 	    psp->problems[psp->num].byVersion = strdup(version);
 	    psp->problems[psp->num].byRelease = strdup(release);
@@ -709,38 +787,39 @@ static int checkPackageDeps(rpmDependencies rpmdep, struct problemsSet * psp,
     return ourrc;
 }
 
-int headerMatchesDepFlags(Header h, char * reqInfo, int reqFlags) {
-    char * name, * version, * release, * chptr;
-    char * reqVersion = reqInfo;
-    char * reqRelease = NULL;
+int headerMatchesDepFlags(Header h, const char * reqInfo, int reqFlags) {
+    const char * name, * version, * release, * chptr;
+    const char * reqVersion = reqInfo;
+    const char * reqRelease = NULL;
     int type, count;
-    int_32 * serial;
+    int_32 * epoch;
     char buf[20];
     int result = 0;
     int sense;
 
-    headerGetEntry(h, RPMTAG_NAME, &type, (void *) &name, &count);
+    headerGetEntry(h, RPMTAG_NAME, &type, (void **) &name, &count);
 
     if (!(reqFlags & RPMSENSE_SENSEMASK) || !reqInfo || !strlen(reqInfo)) {
 	return 1;
     }
 
     if (reqFlags & RPMSENSE_SERIAL) {
-	if (!headerGetEntry(h, RPMTAG_SERIAL, &type, (void *) &serial, &count)) {
+	if (!headerGetEntry(h, RPMTAG_EPOCH, &type, (void **) &epoch, &count)) {
 	    return 0;
 	}
-	sprintf(buf, "%d", *serial);
+	sprintf(buf, "%d", *epoch);
 	version = buf;
     } else {
-	headerGetEntry(h, RPMTAG_VERSION, &type, (void *) &version, &count);
+	headerGetEntry(h, RPMTAG_VERSION, &type, (void **) &version, &count);
 	chptr = strrchr(reqInfo, '-');
 	if (chptr) {
-	    reqVersion = alloca(strlen(reqInfo) + 1);
-	    strcpy(reqVersion, reqInfo);
-	    reqVersion[chptr - reqInfo] = '\0';
+	    char *rv = alloca(strlen(reqInfo) + 1);
+	    strcpy(rv, reqInfo);
+	    rv[chptr - reqInfo] = '\0';
+	    reqVersion = rv;
 	    reqRelease = reqVersion + (chptr - reqInfo) + 1;
 	    if (*reqRelease) 
-		headerGetEntry(h, RPMTAG_RELEASE, &type, (void *) &release, &count);
+		headerGetEntry(h, RPMTAG_RELEASE, &type, (void **) &release, &count);
 	    else
 		reqRelease = NULL;
 	}
@@ -763,13 +842,13 @@ int headerMatchesDepFlags(Header h, char * reqInfo, int reqFlags) {
     return result;
 }
 
-static int dbrecMatchesDepFlags(rpmDependencies rpmdep, int recOffset, 
+static int dbrecMatchesDepFlags(rpmTransactionSet rpmdep, int recOffset, 
 			        char * reqVersion, int reqFlags) {
     Header h;
     int rc;
 
     h = rpmdbGetRecord(rpmdep->db, recOffset);
-    if (!h) {
+    if (h == NULL) {
 	rpmMessage(RPMMESS_DEBUG, _("dbrecMatchesDepFlags() failed to read header"));
 	return 0;
     }
@@ -790,9 +869,9 @@ static int dbrecMatchesDepFlags(rpmDependencies rpmdep, int recOffset,
    the current selection pass is included as a separate parameter, and is
    incremented when satisfying a prerequisite */
 
-static int addOrderedPack(rpmDependencies rpmdep, 
+static int addOrderedPack(rpmTransactionSet rpmdep, 
 			struct availablePackage * package,
-			void ** ordering, int * orderNumPtr, 
+			int * ordering, int * orderNumPtr, 
 			int * selected, int selectionClass,
 			int satisfyDepends, char ** errorStack) {
     char ** requires, ** requiresVersion;
@@ -873,18 +952,21 @@ static int addOrderedPack(rpmDependencies rpmdep,
     }
 
     /* whew -- add this package */
-    ordering[(*orderNumPtr)++] = package->key;
+    ordering[(*orderNumPtr)++] = packageNum;
     selected[packageNum] = -1;
 
     return 0;
 }
 
-int rpmdepOrder(rpmDependencies rpmdep, void *** keysListPtr) {
-    int i;
+int rpmdepOrder(rpmTransactionSet rpmdep) {
+    int i, j;
     int * selected;
-    void ** order;
-    int orderNum;
+    int * ordering;
+    int orderingCount;
     char ** errorStack;
+    struct transactionElement * newOrder;
+    int newOrderCount = 0;
+    struct orderListIndex * orderList, * needle, key;
 
     alMakeIndex(&rpmdep->addedPackages);
     alMakeIndex(&rpmdep->availablePackages);
@@ -895,21 +977,79 @@ int rpmdepOrder(rpmDependencies rpmdep, void *** keysListPtr) {
     errorStack = alloca(sizeof(*errorStack) * (rpmdep->addedPackages.size + 1));
     *errorStack++ = NULL;
 
-    order = malloc(sizeof(*order) * (rpmdep->addedPackages.size + 1));
-    orderNum = 0;
+    ordering = alloca(sizeof(*ordering) * (rpmdep->addedPackages.size + 1));
+    orderingCount = 0;
 
     for (i = 0; i < rpmdep->addedPackages.size; i++) {
 	if (!selected[i]) {
 	    if (addOrderedPack(rpmdep, rpmdep->addedPackages.list + i,
-			       order, &orderNum, selected, 1, 0, errorStack)) {
-		free(order);
+			       ordering, &orderingCount, selected, 1, 0, 
+			       errorStack)) {
 		return 1;
 	    }
 	}
     }
 
-    order[orderNum] = NULL;
-    *keysListPtr = order;
+    /* The order ends up as installed packages followed by removed packages,
+       with removes for upgrades immediately follwing the installation of
+       the new package. This would be easier if we could sort the 
+       addedPackages array, but we store indexes into it in various places. */
+    orderList = malloc(sizeof(*orderList) * rpmdep->addedPackages.size);
+    for (i = 0, j = 0; i < rpmdep->orderCount; i++) {
+	if (rpmdep->order[i].type == TR_ADDED) {
+	    orderList[j].alIndex = rpmdep->order[i].u.addedIndex;
+	    orderList[j].orIndex = i;
+	    j++;
+	}
+    }
+    if (j > rpmdep->addedPackages.size) abort();
+
+    qsort(orderList, rpmdep->addedPackages.size, sizeof(*orderList), 
+	  orderListIndexCmp);
+
+    newOrder = malloc(sizeof(*newOrder) * rpmdep->orderCount);
+    for (i = 0, newOrderCount = 0; i < orderingCount; i++) {
+	key.alIndex = ordering[i];
+	needle = bsearch(&key, orderList, rpmdep->addedPackages.size,
+			 sizeof(key), orderListIndexCmp);
+	/* bsearch should never, ever fail */
+	
+	newOrder[newOrderCount++] = rpmdep->order[needle->orIndex];
+	for (j = needle->orIndex + 1; j < rpmdep->orderCount; j++) {
+	    if (rpmdep->order[j].type == TR_REMOVED &&
+		rpmdep->order[j].u.removed.dependsOnIndex == needle->alIndex) {
+		newOrder[newOrderCount++] = rpmdep->order[j];
+	    } else {
+		break;
+	    }
+	}
+    }
+
+    for (i = 0; i < rpmdep->orderCount; i++) {
+	if (rpmdep->order[i].type == TR_REMOVED && 
+	    rpmdep->order[i].u.removed.dependsOnIndex == -1)  {
+	    newOrder[newOrderCount++] = rpmdep->order[i];
+	}
+    }
+
+    if (newOrderCount != rpmdep->orderCount) abort();
+
+    free(rpmdep->order);
+    rpmdep->order = newOrder;
+    rpmdep->orderAlloced = rpmdep->orderCount;
+    free(orderList);
+
+    return 0;
+}
+
+static int orderListIndexCmp(const void * one, const void * two) {
+    const struct orderListIndex * a = one;
+    const struct orderListIndex * b = two;
+
+    if (a->alIndex < b->alIndex)
+	return -1;
+    if (a->alIndex > b->alIndex)
+	return 1;
 
     return 0;
 }
