@@ -1,12 +1,5 @@
-#include "config.h"
-#include "miscfn.h"
+#include "system.h"
 
-#include <stdlib.h>
-#include <string.h>
-#include <sys/stat.h>
-
-#include "header.h"
-#include "intl.h"
 #include "rpmlib.h"
 
 static char * permsFormat(int_32 type, const void * data, 
@@ -193,14 +186,14 @@ static char * depflagsFormat(int_32 type, const void * data,
 
 static int fsnamesTag(Header h, int_32 * type, void ** data, int_32 * count,
 		      int * freeData) {
-    char ** list;
+    const char ** list;
 
     if (rpmGetFilesystemList(&list, count)) {
 	return 1;
     }
 
     *type = RPM_STRING_ARRAY_TYPE;
-    *((char ***) data) = list;
+    *((const char ***) data) = list;
 
     *freeData = 0;
 
@@ -228,7 +221,7 @@ static int instprefixTag(Header h, int_32 * type, void ** data, int_32 * count,
 
 static int fssizesTag(Header h, int_32 * type, void ** data, int_32 * count,
 		      int * freeData) {
-    char ** filenames;
+    const char ** filenames;
     int_32 * filesizes;
     uint_32 * usages;
     int numFiles;
@@ -326,7 +319,6 @@ static int triggertypeTag(Header h, int_32 * type, void ** data,
     int_32 * indices, * flags;
     char ** conds, ** s;
     int i, j;
-    char buf[5];
     int numScripts, numNames;
 
     if (!headerGetEntry(h, RPMTAG_TRIGGERINDEX, NULL, (void **) &indices, 
