@@ -7,27 +7,13 @@
 /* Data read from file descriptors is expected to be in          */
 /* network byte order and is converted on the fly to host order. */
 
-#include "config.h"
-#include "miscfn.h"
+#include "system.h"
 
-#if HAVE_ALLOCA_H
-# include <alloca.h>
-#endif
-
-#include <stdlib.h>
-#include <ctype.h>
-#include <stdlib.h>
-#include <string.h>
-#include <sys/time.h>
-#include <sys/types.h>
-#include <unistd.h>
 #include <netinet/in.h>
 
+#include "rpmio.h"
 #include "header.h"
 
-#include "intl.h"
-#include "tread.h"
-
 #define INDEX_MALLOC_SIZE 8
 
 #define PARSER_BEGIN 	0
@@ -56,6 +42,7 @@ struct headerToken {
 
     int sorted;  
     int langNum;
+    int usageCount;
 };
 
 struct entryInfo {
@@ -118,7 +105,8 @@ static struct indexEntry *findEntry(Header h, int_32 tag, int_32 type);
 static void * grabData(int_32 type, void * p, int_32 c, int * lengthPtr);
 static int dataLength(int_32 type, void * p, int_32 count, int onDisk);
 static void copyEntry(struct indexEntry * entry, 
-			int_32 *type, void **p, int_32 *c);
+			int_32 *type, void **p, int_32 *c,
+			int minimizeMemory);
 static void freeFormat(struct sprintfToken * format, int num);
 static void findTag(char * name, const struct headerTagTableEntry * tags, 
 		    const struct headerSprintfExtension * extensions,
@@ -152,7 +140,9 @@ static char * shescapeFormat(int_32 type, const void * data,
 static int getExtension(Header h, headerTagTagFunction fn, int_32 * typeptr,
 			void ** data, int_32 * countptr, 
 			struct extensionCache * ext);
-char *headerFindI18NString(Header h, struct indexEntry *entry);
+static char *headerFindI18NString(Header h, struct indexEntry *entry);
+static int intGetEntry(Header h, int_32 tag, int_32 *type, void **p, int_32 *c,
+		       int mimMem);
 
 const struct headerSprintfExtension headerDefaultFormats[] = {
     { HEADER_EXT_FORMAT, "octal", { octalFormat } },
@@ -205,7 +195,7 @@ int headerNextIterator(HeaderIterator iter,
     if (tag) {
 	*tag = h->index[slot].info.tag;
     }
-    copyEntry(h->index + slot, type, p, c);
+    copyEntry(h->index + slot, type, p, c, 0);
 
     return 1;
 }
@@ -240,9 +230,11 @@ Header headerCopy(Header h)
     HeaderIterator headerIter;
     Header res = headerNew();
    
+#if 0	/* XXX harmless, but headerInitIterator() does this anyways */
     /* Sort the index -- not really necessary but some old apps may depend
        on this and it certainly won't hurt anything */
     headerSort(h);
+#endif
     headerIter = headerInitIterator(h);
 
     while (headerNextIterator(headerIter, &tag, &type, &ptr, &count)) {
@@ -264,7 +256,7 @@ Header headerCopy(Header h)
 /*                                                                  */
 /********************************************************************/
 
-void headerWrite(int fd, Header h, int magicp)
+void headerWrite(FD_t fd, Header h, int magicp)
 {
     void * p;
     int length;
@@ -273,53 +265,52 @@ void headerWrite(int fd, Header h, int magicp)
     p = doHeaderUnload(h, &length);
 
     if (magicp) {
-	write(fd, header_magic, sizeof(header_magic));
+	(void)fdWrite(fd, header_magic, sizeof(header_magic));
 	l = htonl(0);
-	write(fd, &l, sizeof(l));
+	(void)fdWrite(fd, &l, sizeof(l));
     }
     
-    write(fd, p, length);
+    (void)fdWrite(fd, p, length);
 
     free(p);
 }
 
-Header headerRead(int fd, int magicp)
+Header headerRead(FD_t fd, int magicp)
 {
+    int_32 block[40];
     int_32 reserved;
     int_32 * p;
     int_32 il, dl;
     int_32 magic;
     Header h;
-    void * block;
+    void * dataBlock;
     int totalSize;
+    int i;
+
+    i = 2;
+    if (magicp == HEADER_MAGIC_YES)
+	i += 2;
+
+    if (timedRead(fd, block, i * sizeof(*block)) != (i * sizeof(*block)))
+	return NULL;
+    i = 0;
 
     if (magicp == HEADER_MAGIC_YES) {
-	if (timedRead(fd, &magic, sizeof(magic)) != sizeof(magic))
-	    return NULL;
+	magic = block[i++];
 	if (memcmp(&magic, header_magic, sizeof(magic))) {
 	    return NULL;
 	}
 
-	if (timedRead(fd, &reserved, sizeof(reserved)) != sizeof(reserved))
-	    return NULL;
+	reserved = block[i++];
     }
     
-    /* First read the index length (count of index entries) */
-    if (timedRead(fd, &il, sizeof(il)) != sizeof(il)) 
-	return NULL;
-
-    il = ntohl(il);
-
-    /* Then read the data length (number of bytes) */
-    if (timedRead(fd, &dl, sizeof(dl)) != sizeof(dl)) 
-	return NULL;
-
-    dl = ntohl(dl);
+    il = ntohl(block[i++]);
+    dl = ntohl(block[i++]);
 
     totalSize = sizeof(int_32) + sizeof(int_32) + 
 		(il * sizeof(struct entryInfo)) + dl;
 
-    block = p = malloc(totalSize);
+    dataBlock = p = malloc(totalSize);
     *p++ = htonl(il);
     *p++ = htonl(dl);
 
@@ -327,14 +318,14 @@ Header headerRead(int fd, int magicp)
     if (timedRead(fd, p, totalSize) != totalSize)
 	return NULL;
     
-    h = headerLoad(block);
+    h = headerLoad(dataBlock);
 
-    free(block);
+    free(dataBlock);
 
     return h;
 }
 
-void headerGzWrite(gzFile fd, Header h, int magicp)
+void headerGzWrite(FD_t fd, Header h, int magicp)
 {
     void * p;
     int length;
@@ -343,17 +334,17 @@ void headerGzWrite(gzFile fd, Header h, int magicp)
     p = doHeaderUnload(h, &length);
 
     if (magicp) {
-	gzwrite(fd, header_magic, sizeof(header_magic));
+	gzdWrite(fd, header_magic, sizeof(header_magic));
 	l = htonl(0);
-	gzwrite(fd, &l, sizeof(l));
+	gzdWrite(fd, &l, sizeof(l));
     }
     
-    gzwrite(fd, p, length);
+    gzdWrite(fd, p, length);
 
     free(p);
 }
 
-Header headerGzRead(gzFile fd, int magicp)
+Header headerGzRead(FD_t fd, int magicp)
 {
     int_32 reserved;
     int_32 * p;
@@ -364,24 +355,24 @@ Header headerGzRead(gzFile fd, int magicp)
     int totalSize;
 
     if (magicp == HEADER_MAGIC_YES) {
-	if (gzread(fd, &magic, sizeof(magic)) != sizeof(magic))
+	if (gzdRead(fd, &magic, sizeof(magic)) != sizeof(magic))
 	    return NULL;
 	if (memcmp(&magic, header_magic, sizeof(magic))) {
 	    return NULL;
 	}
 
-	if (gzread(fd, &reserved, sizeof(reserved)) != sizeof(reserved))
+	if (gzdRead(fd, &reserved, sizeof(reserved)) != sizeof(reserved))
 	    return NULL;
     }
     
     /* First read the index length (count of index entries) */
-    if (gzread(fd, &il, sizeof(il)) != sizeof(il)) 
+    if (gzdRead(fd, &il, sizeof(il)) != sizeof(il)) 
 	return NULL;
 
     il = ntohl(il);
 
     /* Then read the data length (number of bytes) */
-    if (gzread(fd, &dl, sizeof(dl)) != sizeof(dl)) 
+    if (gzdRead(fd, &dl, sizeof(dl)) != sizeof(dl)) 
 	return NULL;
 
     dl = ntohl(dl);
@@ -394,7 +385,7 @@ Header headerGzRead(gzFile fd, int magicp)
     *p++ = htonl(dl);
 
     totalSize -= sizeof(int_32) + sizeof(int_32);
-    if (gzread(fd, p, totalSize) != totalSize)
+    if (gzdRead(fd, p, totalSize) != totalSize)
 	return NULL;
     
     h = headerLoad(block);
@@ -432,6 +423,7 @@ Header headerLoad(void *pv)
     h->indexAlloced = il;
     h->indexUsed = il;
     h->index = malloc(il * sizeof(struct indexEntry));
+    h->usageCount = 1;
 
     /* This assumes you only headerLoad() something you headerUnload()-ed */
     h->sorted = 1;
@@ -446,8 +438,8 @@ Header headerLoad(void *pv)
 	entry->info.count = htonl(pe->count);
 	entry->info.offset = -1;
 
-        if (entry->info.type < RPM_MIN_TYPE ||
-            entry->info.type > RPM_MAX_TYPE) return NULL;
+	if (entry->info.type < RPM_MIN_TYPE ||
+	    entry->info.type > RPM_MAX_TYPE) return NULL;
 
 	src = dataStart + htonl(pe->offset);
 	entry->length = dataLength(entry->info.type, src, 
@@ -569,11 +561,9 @@ void *headerUnload(Header h) {
 
 void headerDump(Header h, FILE *f, int flags, 
 		const struct headerTagTableEntry * tags) {
-    int i, c, ct;
+    int i;
     struct indexEntry *p;
     const struct headerTagTableEntry * tage;
-    char *dp;
-    char ch;
     char *type, *tag;
 
     /* First write out the length of the index (count of index entries) */
@@ -607,20 +597,21 @@ void headerDump(Header h, FILE *f, int flags,
 	    tag = tage->name;
 
 	fprintf(f, "Entry      : %.3d (%d)%-14s %-18s 0x%.8x %.8d\n", i,
-		p->info.tag, tag, type, (uint_32) p->info.offset, (uint_32) 
+		p->info.tag, tag, type, (unsigned) p->info.offset, (int) 
 		p->info.count);
 
 	if (flags & HEADER_DUMP_INLINE) {
+	    char *dp = p->data;
+	    int c = p->info.count;
+	    int ct = 0;
+
 	    /* Print the data inline */
-	    dp = p->data;
-	    c = p->info.count;
-	    ct = 0;
 	    switch (p->info.type) {
 	    case RPM_INT32_TYPE:
 		while (c--) {
 		    fprintf(f, "       Data: %.3d 0x%08x (%d)\n", ct++,
-			    (uint_32) *((int_32 *) dp),
-			    (uint_32) *((int_32 *) dp));
+			    (unsigned) *((int_32 *) dp),
+			    (int) *((int_32 *) dp));
 		    dp += sizeof(int_32);
 		}
 		break;
@@ -628,16 +619,16 @@ void headerDump(Header h, FILE *f, int flags,
 	    case RPM_INT16_TYPE:
 		while (c--) {
 		    fprintf(f, "       Data: %.3d 0x%04x (%d)\n", ct++,
-			    (short int) *((int_16 *) dp),
-			    (short int) *((int_16 *) dp));
+			    (unsigned) (*((int_16 *) dp) & 0xffff),
+			    (int) *((int_16 *) dp));
 		    dp += sizeof(int_16);
 		}
 		break;
 	    case RPM_INT8_TYPE:
 		while (c--) {
 		    fprintf(f, "       Data: %.3d 0x%02x (%d)\n", ct++,
-			    (char) *((int_8 *) dp),
-			    (char) *((int_8 *) dp));
+			    (unsigned) (*((int_8 *) dp) & 0xff),
+			    (int) *((int_8 *) dp));
 		    dp += sizeof(int_8);
 		}
 		break;
@@ -645,7 +636,7 @@ void headerDump(Header h, FILE *f, int flags,
 	      while (c > 0) {
 		  fprintf(f, "       Data: %.3d ", ct);
 		  while (c--) {
-		      fprintf(f, "%02x ", (unsigned char) *(int_8 *)dp);
+		      fprintf(f, "%02x ", (unsigned) (*(int_8 *)dp & 0xff));
 		      ct++;
 		      dp += sizeof(int_8);
 		      if (! (ct % 8)) {
@@ -657,9 +648,9 @@ void headerDump(Header h, FILE *f, int flags,
 		break;
 	    case RPM_CHAR_TYPE:
 		while (c--) {
-		    ch = (char) *((char *) dp);
+		    char ch = (char) *((char *) dp);
 		    fprintf(f, "       Data: %.3d 0x%2x %c (%d)\n", ct++,
-			    ch,
+			    (unsigned)(ch & 0xff),
 			    (isprint(ch) ? ch : ' '),
 			    (char) *((char *) dp));
 		    dp += sizeof(char);
@@ -729,7 +720,8 @@ int headerIsEntry(Header h, int_32 tag)
 }
 
 static void copyEntry(struct indexEntry * entry, 
-			int_32 *type, void **p, int_32 *c) {
+			int_32 *type, void **p, int_32 *c,
+			int minimizeMemory) {
     int i, tableSize;
     char ** ptrEntry;
     char * chptr;
@@ -753,9 +745,14 @@ static void copyEntry(struct indexEntry * entry,
       case RPM_I18NSTRING_TYPE:
 	i = entry->info.count;
 	tableSize = i * sizeof(char *);
-	ptrEntry = *p = malloc(tableSize + entry->length);
-	chptr = ((char *) *p) + tableSize;
-	memcpy(chptr, entry->data, entry->length);
+	if (minimizeMemory) {
+	    ptrEntry = *p = malloc(tableSize);
+	    chptr = entry->data;
+	} else {
+	    ptrEntry = *p = malloc(tableSize + entry->length);
+	    chptr = ((char *) *p) + tableSize;
+	    memcpy(chptr, entry->data, entry->length);
+	}
 	while (i--) {
 	    *ptrEntry++ = chptr;
 	    chptr = strchr(chptr, 0);
@@ -771,31 +768,31 @@ static void copyEntry(struct indexEntry * entry,
 int headerGetRawEntry(Header h, int_32 tag, int_32 *type, void **p, int_32 *c) {
     struct indexEntry * entry;
 
-    if (!p) return headerIsEntry(h, tag);
+    if (p == NULL) return headerIsEntry(h, tag);
 
     /* First find the tag */
     entry = findEntry(h, tag, RPM_NULL_TYPE);
     if (!entry) {
-	*p = NULL;
+	if (p) *p = NULL;
+	if (c) *c = 0;
 	return 0;
     }
 
-    copyEntry(entry, type, p, c);
+    copyEntry(entry, type, p, c, 0);
 
     return 1;
 }
 
-int headerGetEntry(Header h, int_32 tag, int_32 * type, void **p, int_32 * c)
-{
+static int intGetEntry(Header h, int_32 tag, int_32 *type, void **p, int_32 *c,
+		       int minMem) {
     struct indexEntry * entry;
     char * chptr;
 
-    if (!p) return headerIsEntry(h, tag);
-
     /* First find the tag */
     entry = findEntry(h, tag, RPM_NULL_TYPE);
     if (!entry) {
-	*p = NULL;
+	if (p) *p = NULL;
+	if (c) *c = 0;
 	return 0;
     }
 
@@ -807,12 +804,21 @@ int headerGetEntry(Header h, int_32 tag, int_32 * type, void **p, int_32 * c)
 
 	*p = chptr;
     } else {
-	copyEntry(entry, type, p, c);
+	copyEntry(entry, type, p, c, minMem);
     }
 
     return 1;
 }
 
+int headerGetEntryMinMemory(Header h, int_32 tag, int_32 *type, void **p, 
+			    int_32 *c) {
+    return intGetEntry(h, tag, type, p, c, 1);
+}
+
+int headerGetEntry(Header h, int_32 tag, int_32 * type, void **p, int_32 * c) {
+    return intGetEntry(h, tag, type, p, c, 0);
+}
+
 /********************************************************************/
 /*                                                                  */
 /* Header creation and deletion                                     */
@@ -829,6 +835,7 @@ Header headerNew()
 
     h->sorted = 0;
     h->langNum = -1;
+    h->usageCount = 1;
 
     return (Header) h;
 }
@@ -837,6 +844,7 @@ void headerFree(Header h)
 {
     int i;
 
+    if (--h->usageCount) return;
     for (i = 0; i < h->indexUsed; i++)
 	free(h->index[i].data);
 
@@ -844,6 +852,12 @@ void headerFree(Header h)
     free(h);
 }
 
+Header headerLink(Header h)
+{
+    h->usageCount++;
+    return h;
+}
+
 unsigned int headerSizeof(Header h, int magicp)
 {
     unsigned int size;
@@ -879,6 +893,7 @@ static int dataLength(int_32 type, void * p, int_32 count, int onDisk) {
     int thisLen, length, i;
     char ** src, * chptr;
 
+    length = 0;
     switch (type) {
       case RPM_STRING_TYPE:
 	if (count == 1) {
@@ -1001,6 +1016,26 @@ int headerAddEntry(Header h, int_32 tag, int_32 type, void *p, int_32 c)
     return 1;
 }
 
+char **
+headerGetLangs(Header h)
+{
+    char **s, *e, **table;
+    int i, type, count;
+
+    if (!headerGetRawEntry(h, HEADER_I18NTABLE, &type, (void **)&s, &count))
+	return NULL;
+
+    if ((table = (char **)calloc((count+1), sizeof(char *))) == NULL)
+	return NULL;
+
+    for (i = 0, e = *s; i < count > 0; i++, e += strlen(e)+1) {
+	table[i] = e;
+    }
+    table[count] = NULL;
+
+    return table;
+}
+
 int headerAddI18NString(Header h, int_32 tag, char * string, char * lang) {
     struct indexEntry * table, * entry;
     char * charArray[2];
@@ -1067,24 +1102,39 @@ int headerAddI18NString(Header h, int_32 tag, char * string, char * lang) {
 	entry->length += length;
 	entry->info.count = langNum + 1;
     } else {
-	/* tricky... we need to add one to the middle (or beginning). This
-	   will replace an existing entry as well */
-	buf = malloc(strlen(string) + entry->length);
+	char *b, *be, *e, *ee, *t;
+	size_t bn, sn, en;
 
-	chptr = entry->data, length = 0;
-	for (i = 0; i < langNum; i++) 
-	    length += strlen(chptr + length) + 1;
+	/* Set beginning/end pointers to previous data */
+	b = be = e = ee = entry->data;
+	for (i = 0; i < table->info.count; i++) {
+	    if (i == langNum)
+		be = ee;
+	    ee += strlen(ee) + 1;
+	    if (i == langNum)
+		e  = ee;
+	}
 
-	memcpy(buf, entry->data, length);
-	strcpy(buf + length, string);
-	i = strlen(chptr + length) + 1;
-	memcpy(buf + length + strlen(string) + 1,
-	       ((char *) entry->data) + length + i, 
-	       entry->length - length - i);
+	/* Get storage for new buffer */
+	bn = (be-b);
+	sn = strlen(string) + 1;
+	en = (ee-e);
+	length = bn + sn + en;
+	t = buf = malloc(length);
 
+	/* Copy values into new storage */
+	memcpy(t, b, bn);
+	t += bn;
+	memcpy(t, string, sn);
+	t += sn;
+	memcpy(t, e, en);
+	t += en;
+
+	/* Replace I18N string array */
+	entry->length -= strlen(be) + 1;
+	entry->length += sn;
 	free(entry->data);
 	entry->data = buf;
-	entry->length += length;
     }
 
     return 0;
@@ -1423,8 +1473,8 @@ static int parseFormat(char * str, const struct headerTagTableEntry * tags,
 		format[currToken].u.tag.arrayCount = 0;
 
 		chptr = start;
-		while (*chptr && *chptr != '{') chptr++;
-		if (!*chptr) {
+		while (*chptr && *chptr != '{' && *chptr != '%') chptr++;
+		if (!*chptr || *chptr == '%') {
 		    *error = _("missing { after %");
 		    freeFormat(format, numTokens);
 		    return 1;
@@ -1843,7 +1893,7 @@ static struct extensionCache * allocateExtensionCache(
     return calloc(i, sizeof(struct extensionCache));
 }
 
-void freeExtensionCache(const struct headerSprintfExtension * extensions,
+static void freeExtensionCache(const struct headerSprintfExtension * extensions,
 		        struct extensionCache * cache) {
     const struct headerSprintfExtension * ext = extensions;
     int i = 0;
@@ -1903,6 +1953,7 @@ char * headerSprintf(Header h, const char * origFmt,
 
 	    strcat(answer, piece);
 	    answerLength += pieceLength;
+	    free(piece);
 	}
     }
 
@@ -1949,7 +2000,6 @@ static char * realDateFormat(int_32 type, const void * data,
 			     char * formatPrefix, int padding, int element,
 			     char * strftimeFormat) {
     char * val;
-    time_t dateint;
     struct tm * tstruct;
     char buf[50];
 
@@ -1961,9 +2011,10 @@ static char * realDateFormat(int_32 type, const void * data,
 	strcat(formatPrefix, "s");
 
 	/* this is important if sizeof(int_32) ! sizeof(time_t) */
-	dateint = *((int_32 *) data);
-	tstruct = localtime(&dateint);
-	strftime(buf, sizeof(buf) - 1, strftimeFormat, tstruct);
+	{   time_t dateint = *((int_32 *) data);
+	    tstruct = localtime(&dateint);
+	}
+	(void)strftime(buf, sizeof(buf) - 1, strftimeFormat, tstruct);
 	sprintf(val, formatPrefix, buf);
     }
 
@@ -2066,7 +2117,7 @@ void headerResetLang(Header h) {
     headerSetLangPath(h, getenv("LANG"));
 }
 
-char *headerFindI18NString(Header h, struct indexEntry *entry)
+static char *headerFindI18NString(Header h, struct indexEntry *entry)
 {
     char * lang, * buf, * chptr, * start, * next, * resptr;
     struct indexEntry * table;
@@ -2110,3 +2161,24 @@ char *headerFindI18NString(Header h, struct indexEntry *entry)
 
     return entry->data;
 }
+
+void headerCopyTags(Header headerFrom, Header headerTo, int *tagstocopy)
+{
+    int *p;
+
+    if (headerFrom == headerTo)
+	return;
+
+    for (p = tagstocopy; *p != 0; p++) {
+	char *s;
+	int type, count;
+	if (headerIsEntry(headerTo, *p))
+	    continue;
+	if (!headerGetEntry(headerFrom, *p, &type, (void **) &s, &count))
+	    continue;
+	headerAddEntry(headerTo, *p, type, s, count);
+	if (s != NULL &&
+	   (type == RPM_STRING_ARRAY_TYPE || type == RPM_I18NSTRING_TYPE))
+	    free(s);
+    }
+}
