@@ -1,17 +1,8 @@
-#include "miscfn.h"
+#include "system.h"
 
-#if HAVE_ALLOCA_H
-# include <alloca.h>
-#endif
-
-#include <string.h>
-
-#include "intl.h"
+#include "lookup.h"
 #include "rpmlib.h"
 
-static int findMatches(rpmdb db, char * name, char * version, char * release,
-		       dbiIndexSet * matches);
-
 /* 0 found matches */
 /* 1 no matches */
 /* 2 error */
@@ -28,7 +19,7 @@ int rpmdbFindByHeader(rpmdb db, Header h, dbiIndexSet * matches) {
 /* 0 found matches */
 /* 1 no matches */
 /* 2 error */
-int rpmdbFindByLabel(rpmdb db, char * arg, dbiIndexSet * matches) {
+int rpmdbFindByLabel(rpmdb db, const char * arg, dbiIndexSet * matches) {
     char * localarg, * chptr;
     char * release;
     int rc;
@@ -64,8 +55,9 @@ int rpmdbFindByLabel(rpmdb db, char * arg, dbiIndexSet * matches) {
 /* 0 found matches */
 /* 1 no matches */
 /* 2 error */
-static int findMatches(rpmdb db, char * name, char * version, char * release,
-		       dbiIndexSet * matches) {
+int findMatches(rpmdb db, const char * name, const char * version,
+			const char * release, dbiIndexSet * matches)
+{
     int gotMatches;
     int rc;
     int i;
@@ -86,7 +78,7 @@ static int findMatches(rpmdb db, char * name, char * version, char * release,
     for (i = 0; i < matches->count; i++) {
 	if (matches->recs[i].recOffset) {
 	    h = rpmdbGetRecord(db, matches->recs[i].recOffset);
-	    if (!h) {
+	    if (h == NULL) {
 		rpmError(RPMERR_DBCORRUPT, 
 			 _("cannot read header at %d for lookup"), 
 			matches->recs[i].recOffset);
