@@ -1,10 +1,8 @@
+#include "system.h"
+
 #include <stdarg.h>
-#include <stdio.h>
 
-#include "intl.h"
-#include "messages.h"
 #include "rpmlib.h"
-#include "intl.h"
 
 static int minLevel = RPMMESS_NORMAL;
 
@@ -31,7 +29,7 @@ int rpmIsVerbose(void)
     return (minLevel <= RPMMESS_VERBOSE);
 }
 
-void rpmMessage(int level, char * format, ...) {
+void rpmMessage(int level, const char * format, ...) {
     va_list args;
 
     va_start(args, format);
