@@ -1,32 +1,15 @@
-#include "config.h"
-#include "miscfn.h"
+#include "system.h"
 
-#if HAVE_ALLOCA_H
-# include <alloca.h>
-#endif
-
-#include <errno.h>
-#include <ctype.h>
-#include <fcntl.h>
-#include <grp.h>
-#include <pwd.h>
-#include <stdlib.h>
-#include <sys/time.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <stdio.h>
-#include <string.h>
-#include <unistd.h>
-
-#include "intl.h"
-#include "misc.h"
 #include "rpmlib.h"
-#include "messages.h"
+#include "rpmmacro.h"	/* XXX for rpmGetPath */
+
+#include "misc.h"
 
 char * RPMVERSION = VERSION;	/* just to put a marker in librpm.a */
 
-char ** splitString(char * str, int length, char sep) {
-    char * s, * source, * dest;
+char ** splitString(const char * str, int length, char sep) {
+    const char * source;
+    char * s, * dest;
     char ** list;
     int i;
     int fields;
@@ -64,7 +47,7 @@ void freeSplitString(char ** list) {
     free(list);
 }
 
-int exists(char * filespec) {
+int rpmfileexists(const char * filespec) {
     struct stat buf;
 
     if (stat(filespec, &buf)) {
@@ -78,24 +61,30 @@ int exists(char * filespec) {
     return 1;
 }
 
-int rpmvercmp(char * one, char * two) {
-    int num1, num2;
+/* compare alpha and numeric segments of two versions */
+/* return 1: a is newer than b */
+/*        0: a and b are the same version */
+/*       -1: b is newer than a */
+int rpmvercmp(const char * a, const char * b) {
     char oldch1, oldch2;
     char * str1, * str2;
+    char * one, * two;
     int rc;
     int isnum;
     
-    if (!strcmp(one, two)) return 0;
+    /* easy comparison to see if versions are identical */
+    if (!strcmp(a, b)) return 0;
 
-    str1 = alloca(strlen(one) + 1);
-    str2 = alloca(strlen(two) + 1);
+    str1 = alloca(strlen(a) + 1);
+    str2 = alloca(strlen(b) + 1);
 
-    strcpy(str1, one);
-    strcpy(str2, two);
+    strcpy(str1, a);
+    strcpy(str2, b);
 
     one = str1;
     two = str2;
 
+    /* loop through each version segment of str1 and str2 and compare them */
     while (*one && *two) {
 	while (*one && !isalnum(*one)) one++;
 	while (*two && !isalnum(*two)) two++;
@@ -103,6 +92,9 @@ int rpmvercmp(char * one, char * two) {
 	str1 = one;
 	str2 = two;
 
+	/* grab first completely alpha or completely numeric segment */
+	/* leave one and two pointing to the start of the alpha or numeric */
+	/* segment and walk str1 and str2 to end of segment */
 	if (isdigit(*str1)) {
 	    while (*str1 && isdigit(*str1)) str1++;
 	    while (*str2 && isdigit(*str2)) str2++;
@@ -113,35 +105,52 @@ int rpmvercmp(char * one, char * two) {
 	    isnum = 0;
 	}
 		
+	/* save character at the end of the alpha or numeric segment */
+	/* so that they can be restored after the comparison */
 	oldch1 = *str1;
 	*str1 = '\0';
 	oldch2 = *str2;
 	*str2 = '\0';
 
+	/* take care of the case where the two version segments are */
+	/* different types: one numeric and one alpha */
 	if (one == str1) return -1;	/* arbitrary */
 	if (two == str2) return -1;
 
 	if (isnum) {
-	    num1 = atoi(one);
-	    num2 = atoi(two);
+	    /* this used to be done by converting the digit segments */
+	    /* to ints using atoi() - it's changed because long  */
+	    /* digit segments can overflow an int - this should fix that. */
+	  
+	    /* throw away any leading zeros - it's a number, right? */
+	    while (*one == '0') one++;
+	    while (*two == '0') two++;
 
-	    if (num1 < num2) 
-		return -1;
-	    else if (num1 > num2)
-		return 1;
-	} else {
-	    rc = strcmp(one, two);
-	    if (rc) return rc;
+	    /* whichever number has more digits wins */
+	    if (strlen(one) > strlen(two)) return 1;
+	    if (strlen(two) > strlen(one)) return -1;
 	}
+
+	/* strcmp will return which one is greater - even if the two */
+	/* segments are alpha or if they are numeric.  don't return  */
+	/* if they are equal because there might be more segments to */
+	/* compare */
+	rc = strcmp(one, two);
+	if (rc) return rc;
 	
+	/* restore character that was replaced by null above */
 	*str1 = oldch1;
 	one = str1;
 	*str2 = oldch2;
 	two = str2;
     }
 
+    /* this catches the case where all numeric and alpha segments have */
+    /* compared identically but the segment sepparating characters were */
+    /* different */
     if ((!*one) && (!*two)) return 0;
 
+    /* whichever version still has characters left over wins */
     if (!*one) return -1; else return 1;
 }
 
@@ -331,45 +340,67 @@ char * gidToGname(gid_t gid) {
     }
 }
 
-int makeTempFile(char * prefix, char ** fnptr, int * fdptr) {
-    char * fn;
-    int fd;
+int makeTempFile(const char * prefix, const char ** fnptr, FD_t * fdptr) {
+    const char * fn;
+    FD_t fd;
     int ran;
-    char * tmpdir = rpmGetVar(RPMVAR_TMPPATH);
     struct stat sb, sb2;
 
     if (!prefix) prefix = "";
 
-    fn = malloc(strlen(prefix) + 25 + strlen(tmpdir));
+    fn = NULL;
 
     srand(time(NULL));
     ran = rand() % 100000;
 
-    /* maybe this should use link/stat? */
+     /* maybe this should use link/stat? */
 
     do {
-	sprintf(fn, "%s%s/rpm-tmp.%d", prefix, tmpdir, ran++);
-	fd = open(fn, O_CREAT | O_RDWR | O_EXCL, 0700);
-    } while (fd < 0 && errno == EEXIST);
+	char tfn[32];
+	sprintf(tfn, "rpm-tmp.%d", ran++);
+	if (fn)	xfree(fn);
+	fn = rpmGetPath(prefix, "%{_tmppath}/", tfn, NULL);
+	fd = fdOpen(fn, O_CREAT | O_RDWR | O_EXCL, 0700);
+    } while (fdFileno(fd) < 0 && errno == EEXIST);
 
     if (!stat(fn, &sb) && S_ISLNK(sb.st_mode)) {
 	rpmError(RPMERR_SCRIPT, _("error creating temporary file %s"), fn);
+	xfree(fn);
 	return 1;
     }
 
     if (sb.st_nlink != 1) {
 	rpmError(RPMERR_SCRIPT, _("error creating temporary file %s"), fn);
+	xfree(fn);
 	return 1;
     }
 
-    fstat(fd, &sb2);
+    fstat(fdFileno(fd), &sb2);
     if (sb2.st_ino != sb.st_ino || sb2.st_dev != sb.st_dev) {
 	rpmError(RPMERR_SCRIPT, _("error creating temporary file %s"), fn);
+	xfree(fn);
 	return 1;
     }
 
-    if (fnptr) *fnptr = fn;
+    if (fnptr)
+	*fnptr = fn;
+    else
+	xfree(fn);
     *fdptr = fd;
 
     return 0;
 }
+
+char * currentDirectory(void) {
+    int currDirLen;
+    char * currDir;
+
+    currDirLen = 50;
+    currDir = malloc(currDirLen);
+    while (!getcwd(currDir, currDirLen) && errno == ERANGE) {
+	currDirLen += 50;
+	currDir = realloc(currDir, currDirLen);
+    }
+
+    return currDir;
+}
