@@ -19,7 +19,7 @@ struct oldrpmFileInfo {
 } ;
 
 void oldrpmfileFromSpecLine(char * str, struct oldrpmFileInfo * fi);
-void oldrpmfileFromInfoLine(char * path, char * state, char * str, 
+void oldrpmfileFromInfoLine(char * path, char * state, char * str,
 			struct oldrpmFileInfo * fi);
 void oldrpmfileFree(struct oldrpmFileInfo * fi);
 char * oldrpmfileToInfoStr(struct oldrpmFileInfo * fi);
@@ -51,10 +51,10 @@ struct oldrpmHeaderSpec {
     struct oldrpmFileInfo * files;
 } ;
 
-char * oldhdrReadFromStream(int fd, struct oldrpmHeader * header);
+char * oldhdrReadFromStream(FD_t fd, struct oldrpmHeader * header);
 char * oldhdrReadFromFile(char * filename, struct oldrpmHeader * header);
 char * oldhdrParseSpec(struct oldrpmHeader * header, struct oldrpmHeaderSpec * spec);
 void   oldhdrFree(struct oldrpmHeader * header);
 void   oldhdrSpecFree(struct oldrpmHeaderSpec * spec);
 
-#endif
+#endif	/* H_OLDHEADER */
