@@ -1,68 +1,63 @@
-#include "config.h"
+#include "system.h"
 
-#include <errno.h>
-#include <fcntl.h>
-#include <stdio.h>
-#include <string.h>
-#include <sys/stat.h>	/* for mkdir(2) !?! */
-#include <unistd.h>
-
-#if HAVE_ALLOCA_H
-#include <alloca.h>
-#endif
-
-#include "intl.h"
-#include "messages.h"
-#include "rpmdb.h"
 #include "rpmlib.h"
+#include "rpmmacro.h"	/* XXX for rpmGetPath */
 
-int rpmdbRebuild(char * rootdir) {
+#include "rpmdb.h"
+
+int rpmdbRebuild(const char * rootdir) {
     rpmdb olddb, newdb;
-    char * dbpath, * newdbpath;
+    const char * dbpath = NULL;
+    const char * newdbpath = NULL;
+    const char * newrootdbpath = NULL;
     int recnum; 
     Header h;
     int failed = 0;
+    int rc = 0;
+    char tfn[64];
 
     rpmMessage(RPMMESS_DEBUG, _("rebuilding database in rootdir %s\n"), rootdir);
 
-    dbpath = rpmGetVar(RPMVAR_DBPATH);
-    if (!dbpath) {
+    dbpath = rpmGetPath("%{_dbpath}", NULL);
+    if (!dbpath || dbpath[0] == '%') {
 	rpmMessage(RPMMESS_DEBUG, _("no dbpath has been set"));
 	return 1;
     }
 
-    newdbpath = alloca(strlen(dbpath) + 50 + strlen(rootdir));
-    sprintf(newdbpath, "%s/%s/rebuilddb.%d", rootdir, dbpath, (int) getpid());
+    sprintf(tfn, "rebuilddb.%d", (int) getpid());
+    newrootdbpath = rpmGetPath(rootdir, dbpath, tfn, NULL);
+    newdbpath = newrootdbpath + strlen(rootdir);
 
-    if (!access(newdbpath, F_OK)) {
+    if (!access(newrootdbpath, F_OK)) {
 	rpmError(RPMERR_MKDIR, _("temporary database %s already exists"),
-	      newdbpath);
+	      newrootdbpath);
     }
 
-    rpmMessage(RPMMESS_DEBUG, _("creating directory: %s\n"), newdbpath);
-    if (mkdir(newdbpath, 0755)) {
+    rpmMessage(RPMMESS_DEBUG, _("creating directory: %s\n"), newrootdbpath);
+    if (mkdir(newrootdbpath, 0755)) {
 	rpmError(RPMERR_MKDIR, _("error creating directory %s: %s"),
-	      newdbpath, strerror(errno));
+	      newrootdbpath, strerror(errno));
     }
 
-    sprintf(newdbpath, "%s/rebuilddb.%d", dbpath, (int) getpid());
-
     rpmMessage(RPMMESS_DEBUG, _("opening old database\n"));
-    if (openDatabase(rootdir, dbpath, &olddb, O_RDONLY, 0644, 0)) {
-	return 1;
+    if (openDatabase(rootdir, dbpath, &olddb, O_RDONLY, 0644, 
+		     RPMDB_FLAG_MINIMAL)) {
+	rc = 1;
+	goto exit;
     }
 
     rpmMessage(RPMMESS_DEBUG, _("opening new database\n"));
     if (openDatabase(rootdir, newdbpath, &newdb, O_RDWR | O_CREAT, 0644, 0)) {
-	return 1;
+	rc = 1;
+	goto exit;
     }
 
     recnum = rpmdbFirstRecNum(olddb);
     while (recnum > 0) {
-	if (!(h = rpmdbGetRecord(olddb, recnum))) {
-	    rpmError(RPMERR_INTERNAL, 
-		     _("record number %d in database is bad -- skipping it"), 
-		     recnum);
+	if ((h = rpmdbGetRecord(olddb, recnum)) == NULL) {
+	    rpmError(RPMERR_INTERNAL,
+			_("record number %d in database is bad -- skipping it"),
+			recnum);
 	    break;
 	} else {
 	    /* let's sanity check this record a bit, otherwise just skip it */
@@ -72,16 +67,18 @@ int rpmdbRebuild(char * rootdir) {
 		headerIsEntry(h, RPMTAG_RELEASE) &&
 		headerIsEntry(h, RPMTAG_BUILDTIME)) {
 		if (rpmdbAdd(newdb, h)) {
-		    rpmError(RPMERR_INTERNAL, 
-			     _("cannot add record originally at %d"), recnum);
+		    rpmError(RPMERR_INTERNAL,
+			_("cannot add record originally at %d"), recnum);
 		    failed = 1;
 		    break;
 		}
 	    } else {
-		rpmError(RPMERR_INTERNAL, 
-		     _("record number %d in database is bad -- skipping it"), 
-		     recnum);
+		rpmError(RPMERR_INTERNAL,
+			_("record number %d in database is bad -- skipping it"), 
+			recnum);
 	    }
+
+	    headerFree(h);
 	}
 	recnum = rpmdbNextRecNum(olddb, recnum);
     }
@@ -94,20 +91,26 @@ int rpmdbRebuild(char * rootdir) {
 		"remains in place\n"));
 
 	rpmdbRemoveDatabase(rootdir, newdbpath);
-	return 1;
+	rc = 1;
+	goto exit;
     } else {
 	if (rpmdbMoveDatabase(rootdir, newdbpath, dbpath)) {
 	    rpmMessage(RPMMESS_ERROR, _("failed to replace old database with new "
 			"database!\n"));
 	    rpmMessage(RPMMESS_ERROR, _("replaces files in %s with files from %s "
 			"to recover"), dbpath, newdbpath);
-	    return 1;
+	    rc = 1;
+	    goto exit;
 	}
-	if (rmdir(newdbpath))
-	    rpmMessage(RPMERR_RMDIR, _("failed to remove %s: %s\n"),
-			newdbpath, strerror(errno));
+	if (rmdir(newrootdbpath))
+	    rpmMessage(RPMMESS_ERROR, _("failed to remove directory %s: %s\n"),
+			newrootdbpath, strerror(errno));
     }
+    rc = 0;
 
+exit:
+    if (dbpath)		xfree(dbpath);
+    if (newrootdbpath)	xfree(newrootdbpath);
 
-    return 0;
+    return rc;
 }
