@@ -1,25 +1,16 @@
-#include "config.h"
+#include "system.h"
 
-#if HAVE_ALLOCA_H
-# include <alloca.h>
-#endif
-
-#include <fcntl.h>
-#include <stdlib.h>
-#include <string.h>
 #include <sys/file.h>
 #include <signal.h>
 #include <sys/signal.h>
-#include <unistd.h>
 
-#include "dbindex.h"
-#include "falloc.h"
-#include "header.h"
-#include "intl.h"
-#include "misc.h"
-#include "rpmdb.h"
 #include "rpmlib.h"
-#include "messages.h"
+
+#include "falloc.h"
+#include "fprint.h"
+#include "misc.h"
+#include "rpmmacro.h"	/* XXX for rpmGetPath */
+#include "rpmdb.h"
 
 /* XXX the signal handling in here is not thread safe */
 
@@ -38,42 +29,84 @@ struct rpmdb_s {
     dbiIndex * requiredbyIndex, * conflictsIndex, * triggerIndex;
 };
 
+struct intMatch {
+    dbiIndexRecord rec;
+    int fpNum;
+};
+
 static void removeIndexEntry(dbiIndex * dbi, char * name, dbiIndexRecord rec,
 		             int tolerant, char * idxName);
 static int addIndexEntry(dbiIndex * idx, char * index, unsigned int offset,
 		         unsigned int fileNumber);
 static void blockSignals(void);
 static void unblockSignals(void);
+static int intMatchCmp(const void * one, const void * two);
 
 static sigset_t signalMask;
 
-int rpmdbOpen (char * prefix, rpmdb *rpmdbp, int mode, int perms) {
-    char * dbpath;
+static int intMatchCmp(const void * one, const void * two) {
+    const struct intMatch * a = one;
+    const struct intMatch * b = two;
 
-    dbpath = rpmGetVar(RPMVAR_DBPATH);
-    if (!dbpath) {
+    if (a->rec.recOffset < b->rec.recOffset)
+	return -1;
+    else if (a->rec.recOffset > b->rec.recOffset)
+	return 1;
+
+    return 0;
+}
+
+int rpmdbOpenForTraversal(const char * prefix, rpmdb * rpmdbp) {
+    const char * dbpath;
+
+    dbpath = rpmGetPath("%{_dbpath}", NULL);
+    if (dbpath == NULL || dbpath[0] == '%') {
 	rpmMessage(RPMMESS_DEBUG, _("no dbpath has been set"));
 	return 1;
     }
 
-    return openDatabase(prefix, dbpath, rpmdbp, mode, perms, 0);
+    if (openDatabase(prefix, dbpath, rpmdbp, O_RDONLY, 0644, 
+		     RPMDB_FLAG_MINIMAL)) {
+	return 1;
+    }
+    xfree(dbpath);
+    return 0;
 }
 
-int rpmdbInit (char * prefix, int perms) {
+int rpmdbOpen (const char * prefix, rpmdb *rpmdbp, int mode, int perms) {
+    const char * dbpath;
+    int rc;
+
+    dbpath = rpmGetPath("%{_dbpath}", NULL);
+    if (dbpath == NULL || dbpath[0] == '%') {
+	rpmMessage(RPMMESS_DEBUG, _("no dbpath has been set"));
+	return 1;
+    }
+
+    rc = openDatabase(prefix, dbpath, rpmdbp, mode, perms, 0);
+    xfree(dbpath);
+    return rc;
+}
+
+int rpmdbInit (const char * prefix, int perms) {
     rpmdb db;
-    char * dbpath;
+    const char * dbpath;
+    int rc;
 
-    dbpath = rpmGetVar(RPMVAR_DBPATH);
-    if (!dbpath) {
+    dbpath = rpmGetPath("%{_dbpath}", NULL);
+    if (dbpath == NULL || dbpath[0] == '%') {
 	rpmMessage(RPMMESS_DEBUG, _("no dbpath has been set"));
 	return 1;
     }
 
-    return openDatabase(prefix, dbpath, &db, O_CREAT | O_RDWR, perms, 1);
+    rc = openDatabase(prefix, dbpath, &db, O_CREAT | O_RDWR, perms, 
+			RPMDB_FLAG_JUSTCHECK);
+    xfree(dbpath);
+    return rc;
 }
 
-static int openDbFile(char * prefix, char * dbpath, char * shortName, 
-		      int justCheck, int perms, dbiIndex ** db){
+static int openDbFile(const char * prefix, const char * dbpath, const char * shortName, 
+		      int justCheck, int perms, dbiIndex ** db, DBTYPE type){
     char * filename = alloca(strlen(prefix) + strlen(dbpath) + 
 			     strlen(shortName) + 20);
 
@@ -83,8 +116,8 @@ static int openDbFile(char * prefix, char * dbpath, char * shortName,
     strcat(filename, dbpath);
     strcat(filename, shortName);
 
-    if (!justCheck || !exists(filename)) {
-	*db = dbiOpenIndex(filename, perms, 0644);
+    if (!justCheck || !rpmfileexists(filename)) {
+	*db = dbiOpenIndex(filename, perms, 0644, type);
 	if (!*db) {
 	    return 1;
 	}
@@ -93,12 +126,18 @@ static int openDbFile(char * prefix, char * dbpath, char * shortName,
     return 0;
 }
 
-int openDatabase(char * prefix, char * dbpath, rpmdb *rpmdbp, int mode, 
-		 int perms, int justcheck) {
+int openDatabase(const char * prefix, const char * dbpath, rpmdb *rpmdbp, int mode, 
+		 int perms, int flags) {
     char * filename;
     struct rpmdb_s db;
     int i, rc;
     struct flock lockinfo;
+    int justcheck = flags & RPMDB_FLAG_JUSTCHECK;
+    int minimal = flags & RPMDB_FLAG_MINIMAL;
+    const char * akey;
+
+    if (mode & O_WRONLY) 
+	return 1;
 
     /* we should accept NULL as a valid prefix */
     if (!prefix) prefix="";
@@ -114,21 +153,19 @@ int openDatabase(char * prefix, char * dbpath, rpmdb *rpmdbp, int mode,
     
     filename = alloca(strlen(prefix) + strlen(dbpath) + 40);
 
-    if (mode & O_WRONLY) 
-	return 1;
-
     strcpy(filename, prefix); 
     strcat(filename, dbpath);
 
-    rpmMessage(RPMMESS_DEBUG, _("opening database in %s\n"), filename);
+    rpmMessage(RPMMESS_DEBUG, _("opening database mode 0x%x in %s\n"),
+	mode, filename);
 
     strcat(filename, "packages.rpm");
 
     memset(&db, 0, sizeof(db));
 
-    if (!justcheck || !exists(filename)) {
+    if (!justcheck || !rpmfileexists(filename)) {
 	db.pkgs = faOpen(filename, mode, 0644);
-	if (!db.pkgs) {
+	if (db.pkgs == NULL) {
 	    rpmError(RPMERR_DBOPEN, _("failed to open %s\n"), filename);
 	    return 1;
 	}
@@ -141,41 +178,60 @@ int openDatabase(char * prefix, char * dbpath, rpmdb *rpmdbp, int mode,
 	
 	if (mode & O_RDWR) {
 	    lockinfo.l_type = F_WRLCK;
-	    if (fcntl(db.pkgs->fd, F_SETLK, (void *) &lockinfo)) {
+	    if (faFcntl(db.pkgs, F_SETLK, (void *) &lockinfo)) {
 		rpmError(RPMERR_FLOCK, _("cannot get %s lock on database"), 
 			 _("exclusive"));
 		return 1;
 	    } 
 	} else {
 	    lockinfo.l_type = F_RDLCK;
-	    if (fcntl(db.pkgs->fd, F_SETLK, (void *) &lockinfo)) {
+	    if (faFcntl(db.pkgs, F_SETLK, (void *) &lockinfo)) {
 		rpmError(RPMERR_FLOCK, _("cannot get %s lock on database"), 
 			 _("shared"));
 		return 1;
 	    } 
 	}
     }
-    
+
     rc = openDbFile(prefix, dbpath, "nameindex.rpm", justcheck, mode,
-		    &db.nameIndex);
+		    &db.nameIndex, DB_HASH);
+
+    if (minimal) {
+	*rpmdbp = malloc(sizeof(struct rpmdb_s));
+	**rpmdbp = db;
+	return 0;
+    }
+
     if (!rc)
 	rc = openDbFile(prefix, dbpath, "fileindex.rpm", justcheck, mode,
-			&db.fileIndex);
+			&db.fileIndex, DB_HASH);
+
+    /* We used to store the fileindexes as complete paths, rather then
+       plain basenames. Let's see which version we are... */
+    if (!justcheck && !dbiGetFirstKey(db.fileIndex, &akey)) {
+	if (strchr(akey, '/')) {
+	    rpmError(RPMERR_OLDDB, _("old format database is present; "
+			"use --rebuilddb to generate a new format database"));
+	    rc |= 1;
+	}
+	xfree(akey);
+    }
+
     if (!rc)
 	rc = openDbFile(prefix, dbpath, "providesindex.rpm", justcheck, mode,
-			&db.providesIndex);
+			&db.providesIndex, DB_HASH);
     if (!rc)
 	rc = openDbFile(prefix, dbpath, "requiredby.rpm", justcheck, mode,
-			&db.requiredbyIndex);
+			&db.requiredbyIndex, DB_HASH);
     if (!rc)
 	rc = openDbFile(prefix, dbpath, "conflictsindex.rpm", justcheck, mode,
-			&db.conflictsIndex);
+			&db.conflictsIndex, DB_HASH);
     if (!rc)
 	rc = openDbFile(prefix, dbpath, "groupindex.rpm", justcheck, mode,
-			&db.groupIndex);
+			&db.groupIndex, DB_HASH);
     if (!rc)
 	rc = openDbFile(prefix, dbpath, "triggerindex.rpm", justcheck, mode,
-			&db.triggerIndex);
+			&db.triggerIndex, DB_HASH);
 
     if (rc) {
 	faClose(db.pkgs);
@@ -200,7 +256,7 @@ int openDatabase(char * prefix, char * dbpath, rpmdb *rpmdbp, int mode,
 }
 
 void rpmdbClose (rpmdb db) {
-    if (db->pkgs) faClose(db->pkgs);
+    if (db->pkgs != NULL) faClose(db->pkgs);
     if (db->fileIndex) dbiCloseIndex(db->fileIndex);
     if (db->groupIndex) dbiCloseIndex(db->groupIndex);
     if (db->nameIndex) dbiCloseIndex(db->nameIndex);
@@ -221,56 +277,86 @@ int rpmdbNextRecNum(rpmdb db, unsigned int lastOffset) {
 }
 
 Header rpmdbGetRecord(rpmdb db, unsigned int offset) {
-    lseek(db->pkgs->fd, offset, SEEK_SET);
+    (void)faLseek(db->pkgs, offset, SEEK_SET);
 
-    return headerRead(db->pkgs->fd, HEADER_MAGIC_NO);
+    return headerRead(faFileno(db->pkgs), HEADER_MAGIC_NO);
 }
 
-int rpmdbFindByFile(rpmdb db, char * filespec, dbiIndexSet * matches) {
-    char * fs;
-    char * src, * dst;
+int rpmdbFindByFile(rpmdb db, const char * filespec, dbiIndexSet * matches) {
+    const char * basename;
+    fingerPrint fp1, fp2;
+    dbiIndexSet allMatches;
+    int i, rc;
+    Header h;
+    char ** fileList;
 
-    /* we try and canonicalize the filespec a bit before doing the search */
+    fp1 = fpLookup(filespec, 0);
 
-    fs = alloca(strlen(filespec) + 1);
-    for (src = filespec, dst = fs; *src; ) { 
-	switch (*src) {
-	  case '/':
-	    if ((dst == fs) || (*(dst - 1) != '/'))
-		*dst++ = *src;
-	    src++;
-	    break;
-	  default:
-	    *dst++ = *src++;
+    basename = strrchr(filespec, '/');
+    if (!basename) 
+	basename = filespec;
+    else
+	basename++;
+
+    rc = dbiSearchIndex(db->fileIndex, basename, &allMatches);
+    if (rc) return rc;
+
+    *matches = dbiCreateIndexRecord();
+    i = 0;
+    while (i < allMatches.count) {
+	if (!(h = rpmdbGetRecord(db, allMatches.recs[i].recOffset))) {
+	    i++;
+	    continue;
 	}
-    }
-    *dst-- = '\0';
-    if (*dst == '/') *dst = '\0';
 
-    return dbiSearchIndex(db->fileIndex, fs, matches);
+	headerGetEntryMinMemory(h, RPMTAG_FILENAMES, NULL, 
+				(void **) &fileList, NULL);
+
+	do {
+	    fp2 = fpLookup(fileList[allMatches.recs[i].fileNumber], 1);
+	    if (FP_EQUAL(fp1, fp2)) 
+		dbiAppendIndexRecord(matches, allMatches.recs[i]);
+
+	    i++;
+	} while ((i < allMatches.count) && 
+			((i == 0) || (allMatches.recs[i].recOffset == 
+				allMatches.recs[i - 1].recOffset)));
+
+	free(fileList);
+	headerFree(h);
+    }
+
+    dbiFreeIndexRecord(allMatches);
+
+    if (!matches->count) {
+	dbiFreeIndexRecord(*matches);
+	return 1;
+    }
+
+    return 0;
 }
 
-int rpmdbFindByProvides(rpmdb db, char * filespec, dbiIndexSet * matches) {
+int rpmdbFindByProvides(rpmdb db, const char * filespec, dbiIndexSet * matches) {
     return dbiSearchIndex(db->providesIndex, filespec, matches);
 }
 
-int rpmdbFindByRequiredBy(rpmdb db, char * filespec, dbiIndexSet * matches) {
+int rpmdbFindByRequiredBy(rpmdb db, const char * filespec, dbiIndexSet * matches) {
     return dbiSearchIndex(db->requiredbyIndex, filespec, matches);
 }
 
-int rpmdbFindByConflicts(rpmdb db, char * filespec, dbiIndexSet * matches) {
+int rpmdbFindByConflicts(rpmdb db, const char * filespec, dbiIndexSet * matches) {
     return dbiSearchIndex(db->conflictsIndex, filespec, matches);
 }
 
-int rpmdbFindByTriggeredBy(rpmdb db, char * filespec, dbiIndexSet * matches) {
+int rpmdbFindByTriggeredBy(rpmdb db, const char * filespec, dbiIndexSet * matches) {
     return dbiSearchIndex(db->triggerIndex, filespec, matches);
 }
 
-int rpmdbFindByGroup(rpmdb db, char * group, dbiIndexSet * matches) {
+int rpmdbFindByGroup(rpmdb db, const char * group, dbiIndexSet * matches) {
     return dbiSearchIndex(db->groupIndex, group, matches);
 }
 
-int rpmdbFindPackage(rpmdb db, char * name, dbiIndexSet * matches) {
+int rpmdbFindPackage(rpmdb db, const char * name, dbiIndexSet * matches) {
     return dbiSearchIndex(db->nameIndex, name, matches);
 }
 
@@ -311,12 +397,13 @@ int rpmdbRemove(rpmdb db, unsigned int offset, int tolerant) {
     char ** fileList, ** providesList, ** requiredbyList;
     char ** conflictList, ** triggerList;
     int i;
+    char * basename;
 
-    rec.recOffset = offset;
-    rec.fileNumber = 0;
+    /* structure assignment */
+    rec = dbiReturnIndexRecordInstance(offset, 0);
 
     h = rpmdbGetRecord(db, offset);
-    if (!h) {
+    if (h == NULL) {
 	rpmError(RPMERR_DBCORRUPT, _("cannot read header at %d for uninstall"),
 	      offset);
 	return 1;
@@ -390,9 +477,17 @@ int rpmdbRemove(rpmdb db, unsigned int offset, int tolerant) {
     if (headerGetEntry(h, RPMTAG_FILENAMES, &type, (void **) &fileList, 
 	 &count)) {
 	for (i = 0; i < count; i++) {
-	    rpmMessage(RPMMESS_DEBUG, _("removing file index for %s\n"), fileList[i]);
-	    rec.fileNumber = i;
-	    removeIndexEntry(db->fileIndex, fileList[i], rec, tolerant, 
+	    basename = strrchr(fileList[i], '/');
+	    if (!basename) 
+		basename = fileList[i];
+	    else
+		basename++;
+
+	    rpmMessage(RPMMESS_DEBUG, _("removing file index for %s\n"), 
+			basename);
+	    /* structure assignment */
+	    rec = dbiReturnIndexRecordInstance(offset, i);
+	    removeIndexEntry(db->fileIndex, basename, rec, tolerant, 
 			     "file index");
 	}
 	free(fileList);
@@ -419,8 +514,7 @@ static int addIndexEntry(dbiIndex * idx, char * index, unsigned int offset,
     dbiIndexRecord irec;   
     int rc;
 
-    irec.recOffset = offset;
-    irec.fileNumber = fileNumber;
+    irec = dbiReturnIndexRecordInstance(offset, fileNumber);
 
     rc = dbiSearchIndex(idx, index, &set);
     if (rc == -1)  		/* error */
@@ -448,6 +542,7 @@ int rpmdbAdd(rpmdb db, Header dbentry) {
     int triggerCount = 0;
     int type;
     int rc = 0;
+    char * basename;
 
     headerGetEntry(dbentry, RPMTAG_NAME, &type, (void **) &name, &count);
     headerGetEntry(dbentry, RPMTAG_GROUP, &type, (void **) &group, &count);
@@ -479,9 +574,9 @@ int rpmdbAdd(rpmdb db, Header dbentry) {
 	if (count) free(fileList);
 	return 1;
     }
-    lseek(db->pkgs->fd, dboffset, SEEK_SET);
+    (void)faLseek(db->pkgs, dboffset, SEEK_SET);
 
-    headerWrite(db->pkgs->fd, dbentry, HEADER_MAGIC_NO);
+    headerWrite(faFileno(db->pkgs), dbentry, HEADER_MAGIC_NO);
 
     /* Now update the appropriate indexes */
     if (addIndexEntry(db->nameIndex, name, dboffset, 0))
@@ -507,8 +602,15 @@ int rpmdbAdd(rpmdb db, Header dbentry) {
     for (i = 0; i < providesCount; i++)
 	rc += addIndexEntry(db->providesIndex, providesList[i], dboffset, 0);
 
-    for (i = 0; i < count; i++) 
-	rc += addIndexEntry(db->fileIndex, fileList[i], dboffset, i);
+    for (i = 0; i < count; i++) {
+	basename = strrchr(fileList[i], '/');
+	if (!basename) 
+	    basename = fileList[i];
+	else
+	    basename++;
+	if (*basename)
+	    rc += addIndexEntry(db->fileIndex, basename, dboffset, i);
+    }
 
     dbiSyncIndex(db->nameIndex);
     dbiSyncIndex(db->groupIndex);
@@ -533,7 +635,7 @@ int rpmdbUpdateRecord(rpmdb db, int offset, Header newHeader) {
     int oldSize;
 
     oldHeader = rpmdbGetRecord(db, offset);
-    if (!oldHeader) {
+    if (oldHeader == NULL) {
 	rpmError(RPMERR_DBCORRUPT, _("cannot read header at %d for update"),
 		offset);
 	return 1;
@@ -552,9 +654,9 @@ int rpmdbUpdateRecord(rpmdb db, int offset, Header newHeader) {
     } else {
 	blockSignals();
 
-	lseek(db->pkgs->fd, offset, SEEK_SET);
+	(void)faLseek(db->pkgs, offset, SEEK_SET);
 
-	headerWrite(db->pkgs->fd, newHeader, HEADER_MAGIC_NO);
+	headerWrite(faFileno(db->pkgs), newHeader, HEADER_MAGIC_NO);
 
 	unblockSignals();
     }
@@ -573,7 +675,7 @@ static void unblockSignals(void) {
     sigprocmask(SIG_SETMASK, &signalMask, NULL);
 }
 
-void rpmdbRemoveDatabase(char * rootdir, char * dbpath) { 
+void rpmdbRemoveDatabase(const char * rootdir, const char * dbpath) { 
     int i;
     char * filename;
 
@@ -610,7 +712,7 @@ void rpmdbRemoveDatabase(char * rootdir, char * dbpath) {
     unlink(filename);
 }
 
-int rpmdbMoveDatabase(char * rootdir, char * olddbpath, char * newdbpath) {
+int rpmdbMoveDatabase(const char * rootdir, const char * olddbpath, const char * newdbpath) {
     int i;
     char * ofilename, * nfilename;
     int rc = 0;
@@ -670,3 +772,96 @@ int rpmdbMoveDatabase(char * rootdir, char * olddbpath, char * newdbpath) {
 
     return rc;
 }
+
+int rpmdbFindFpList(rpmdb db, fingerPrint * fpList, dbiIndexSet * matchList, 
+		    int numItems) {
+    struct intMatch * intMatches, * im;
+    int i, j;
+    dbiIndexSet matches;
+    int intMatchesAlloced, numIntMatches;
+    int start, end;
+    int num, rc;
+    const char ** fullfl, **fl;
+    int_32 fc;
+    fingerPrint * fps;
+    Header h;
+
+    /* this may be worth batching by basename, but probably not as
+       basenames are quite unique as it is */
+
+    intMatchesAlloced = numItems;
+    numIntMatches = 0;
+    im = intMatches = malloc(sizeof(*intMatches) * intMatchesAlloced);
+
+    for (i = 0; i < numItems; i++, im++) {
+        rc = dbiSearchIndex(db->fileIndex, fpList[i].basename, &matches);
+	if (rc == 2) {
+	    im = intMatches;
+	    free(intMatches);
+	    return 1;
+	} else if (!rc) {
+	    if ((numIntMatches + matches.count) >= intMatchesAlloced) {
+		intMatchesAlloced += matches.count;
+		intMatchesAlloced += intMatchesAlloced / 5;
+		intMatches = realloc(intMatches, 
+				     sizeof(*intMatches) * intMatchesAlloced);
+	    }
+
+	    for (j = 0; j < matches.count; j++) {
+		intMatches[numIntMatches].rec = matches.recs[j];
+		intMatches[numIntMatches++].fpNum = i;
+	    }
+
+	    dbiFreeIndexRecord(matches);
+	}
+    }
+
+    qsort(intMatches, numIntMatches, sizeof(*intMatches), intMatchCmp);
+    /* intMatches is now sorted by (recnum, filenum) */
+
+    for (i = 0; i < numItems; i++)
+	matchList[i] = dbiCreateIndexRecord();
+
+    start = 0;
+    while (start < numIntMatches) {
+	im = intMatches + start;
+	end = start + 1;
+	while (end < numIntMatches && 
+	    (im->rec.recOffset == intMatches[end].rec.recOffset))
+	    end++;
+
+	num = end - start;
+	h = rpmdbGetRecord(db, im->rec.recOffset);
+	if (!h) {
+	    free(intMatches);
+	    return 1;
+	}
+
+	headerGetEntryMinMemory(h, RPMTAG_FILENAMES, NULL, 
+				(void **) &fullfl, &fc);
+
+	fl = malloc(sizeof(*fl) * num);
+	for (i = 0; i < num; i++)
+	    fl[i] = fullfl[im[i].rec.fileNumber];
+	free(fullfl);
+	fps = malloc(sizeof(*fps) * num);
+	fpLookupList(fl, fps, num, 1);
+	free(fl);
+
+	for (i = 0; i < num; i++) {
+	    j = im[i].fpNum;
+	    if (FP_EQUAL(fps[i], fpList[j]))
+		dbiAppendIndexRecord(&matchList[j], im[i].rec);
+	}
+
+	headerFree(h);
+
+	free(fps);
+
+	start = end;
+    }
+
+    free(intMatches);
+
+    return 0;
+}
