@@ -1,27 +1,6 @@
-#include "config.h"
-#include "miscfn.h"
+#include "system.h"
 
-#if HAVE_ALLOCA_H
-# include <alloca.h>
-#endif
-
-#include <ctype.h>
-#include <errno.h>
-#include <fcntl.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <sys/time.h>
-#include <sys/resource.h>
-#include <sys/stat.h>
-#include <sys/utsname.h>
-#include <unistd.h>
-
-#include "intl.h"
-#include "messages.h"
-#include "misc.h"
-#include "rpmlib.h"
-#include "intl.h"
+#include <stdarg.h>
 
 #if HAVE_SYS_SYSTEMCFG_H
 #include <sys/systemcfg.h>
@@ -29,6 +8,13 @@
 #define __power_pc() 0
 #endif
 
+#include "rpmlib.h"
+#include "rpmmacro.h"
+
+#include "misc.h"
+
+static const char *defrcfiles = LIBRPMRC_FILENAME ":/etc/rpmrc:~/.rpmrc";
+
 struct machCacheEntry {
     char * name;
     int count;
@@ -54,14 +40,14 @@ struct machEquivTable {
 struct rpmvarValue {
     char * value;
     /* eventually, this arch will be replaced with a generic condition */
-    char * arch;		
+    char * arch;
     struct rpmvarValue * next;
 };
 
 struct rpmOption {
     char * name;
     int var;
-    int archSpecific, required;
+    int archSpecific, required, macroize, localize;
     struct rpmOptionValue * value;
 };
 
@@ -76,7 +62,7 @@ struct canonEntry {
     short num;
 };
 
-/* tags are 'key'canon, 'key'translate, 'key'compat 
+/* tags are 'key'canon, 'key'translate, 'key'compat
 
    for giggles, 'key'_canon, 'key'_compat, and 'key'_canon will also work */
 struct tableType {
@@ -90,6 +76,7 @@ struct tableType {
     int canonsLength;
 };
 
+/*@-fullinitblock@*/
 static struct tableType tables[RPM_MACHTABLE_COUNT] = {
     { "arch", 1, 0 },
     { "os", 1, 0 },
@@ -98,44 +85,15 @@ static struct tableType tables[RPM_MACHTABLE_COUNT] = {
 };
 
 /* this *must* be kept in alphabetical order */
+/* The order of the flags is archSpecific, required, macroize, localize */
+
 static struct rpmOption optionTable[] = {
-    { "builddir",		RPMVAR_BUILDDIR,		0, 0 },
-    { "buildroot",              RPMVAR_BUILDROOT,               0, 0 },
-    { "buildshell",             RPMVAR_BUILDSHELL,              0, 0 },
-    { "bzip2bin",		RPMVAR_BZIP2BIN,		0, 1 },
-    { "dbpath",			RPMVAR_DBPATH,			0, 1 },
-    { "defaultdocdir",		RPMVAR_DEFAULTDOCDIR,		0, 0 },
-    { "distribution",		RPMVAR_DISTRIBUTION,		0, 0 },
-    { "excludedocs",	        RPMVAR_EXCLUDEDOCS,             0, 0 },
-    { "fixperms",		RPMVAR_FIXPERMS,		0, 1 },
-    { "ftpport",		RPMVAR_FTPPORT,			0, 0 },
-    { "ftpproxy",		RPMVAR_FTPPROXY,		0, 0 },
-    { "gzipbin",		RPMVAR_GZIPBIN,			0, 1 },
-    { "instchangelog",		RPMVAR_INSTCHANGELOG,		0, 0 },
-    { "langpatt",               RPMVAR_LANGPATT,                0, 0 },
-    { "messagelevel",		RPMVAR_MESSAGELEVEL,		0, 0 },
-    { "netsharedpath",		RPMVAR_NETSHAREDPATH,		0, 0 },
-    { "optflags",		RPMVAR_OPTFLAGS,		1, 0 },
-    { "packager",               RPMVAR_PACKAGER,                0, 0 },
-    { "pgp_name",               RPMVAR_PGP_NAME,                0, 0 },
-    { "pgp_path",               RPMVAR_PGP_PATH,                0, 0 },
-    { "provides",               RPMVAR_PROVIDES,                0, 0 },
-    { "require_distribution",	RPMVAR_REQUIREDISTRIBUTION,	0, 0 },
-    { "require_icon",		RPMVAR_REQUIREICON,		0, 0 },
-    { "require_vendor",		RPMVAR_REQUIREVENDOR,		0, 0 },
-/* root is obsolete - use buildroot instead                             */
-/*    { "root",			RPMVAR_ROOT,			0, 0 }, */
-    { "rpmdir",			RPMVAR_RPMDIR,			0, 0 },
-    { "rpmfilename",		RPMVAR_RPMFILENAME,		0, 1 },
-    { "signature",		RPMVAR_SIGTYPE,			0, 0 },
-    { "sourcedir",		RPMVAR_SOURCEDIR,		0, 0 },
-    { "specdir",		RPMVAR_SPECDIR,			0, 0 },
-    { "srcrpmdir",		RPMVAR_SRPMDIR,			0, 0 },
-    { "timecheck",		RPMVAR_TIMECHECK,		0, 0 },
-    { "tmppath",		RPMVAR_TMPPATH,			0, 1 },
-    { "topdir",			RPMVAR_TOPDIR,			0, 0 },
-    { "vendor",			RPMVAR_VENDOR,			0, 0 },
+    { "include",		RPMVAR_INCLUDE,			0, 1,	0, 2 },
+    { "macrofiles",		RPMVAR_MACROFILES,		0, 0,	0, 1 },
+    { "optflags",		RPMVAR_OPTFLAGS,		1, 0,	1, 0 },
+    { "provides",               RPMVAR_PROVIDES,                0, 0,	0, 0 },
 };
+/*@=fullinitblock@*/
 static int optionTableSize = sizeof(optionTable) / sizeof(*optionTable);
 
 #define OS	0
@@ -147,49 +105,55 @@ static struct rpmvarValue values[RPMVAR_NUM];
 
 /* prototypes */
 static void defaultMachine(char ** arch, char ** os);
-static int doReadRC(int fd, char * filename);
+static int doReadRC(FD_t fd, const char * filename);
 static int optionCompare(const void * a, const void * b);
 static int addCanon(struct canonEntry **table, int *tableLen, char *line,
-		   char *fn, int lineNum);
+			const char *fn, int lineNum);
 static int addDefault(struct defaultEntry **table, int *tableLen, char *line,
-		      char *fn, int lineNum);
+			const char *fn, int lineNum);
 static void freeRpmVar(struct rpmvarValue * orig);
-static void rpmSetVarArch(int var, char * val, char * arch);
+static void rpmSetVarArch(int var, const char * val, const char * arch);
 static struct canonEntry *lookupInCanonTable(char *name,
 					   struct canonEntry *table,
 					   int tableLen);
-static char *lookupInDefaultTable(char *name,
+static const char *lookupInDefaultTable(const char *name,
 				  struct defaultEntry *table,
 				  int tableLen);
+
+static void setVarDefault(int var, const char *macroname, const char *val, const char *body);
+static void setPathDefault(int var, const char *macroname, const char *subdir);
 static void setDefaults(void);
-static void setPathDefault(int var, char * s);
-static void rebuildCompatTables(int type, char * name);
+
+static void rebuildCompatTables(int type, char *name);
 
 /* compatiblity tables */
-static int machCompatCacheAdd(char * name, char * fn, int linenum,
+static int machCompatCacheAdd(char * name, const char * fn, int linenum,
 				struct machCache * cache);
-static struct machCacheEntry * machCacheFindEntry(struct machCache * cache, 
+static struct machCacheEntry * machCacheFindEntry(struct machCache * cache,
 						  char * key);
 static struct machEquivInfo * machEquivSearch(
 		struct machEquivTable * table, char * name);
 static void machAddEquiv(struct machEquivTable * table, char * name,
 			   int distance);
-static void machCacheEntryVisit(struct machCache * cache, 
-				  struct machEquivTable * table, 
+static void machCacheEntryVisit(struct machCache * cache,
+				  struct machEquivTable * table,
 				  char * name,
 	  			  int distance);
-static void machFindEquivs(struct machCache * cache, 
+static void machFindEquivs(struct machCache * cache,
 			     struct machEquivTable * table,
 			     char * key);
 
-
 static int optionCompare(const void * a, const void * b) {
     return strcasecmp(((struct rpmOption *) a)->name,
 		      ((struct rpmOption *) b)->name);
 }
 
-static struct machCacheEntry * machCacheFindEntry(struct machCache * cache, 
-						  char * key) {
+static void rpmRebuildTargetVars(const char **target, const char ** canontarget);
+
+
+static struct machCacheEntry * machCacheFindEntry(struct machCache * cache,
+						  char * key)
+{
     int i;
 
     for (i = 0; i < cache->size; i++)
@@ -198,22 +162,23 @@ static struct machCacheEntry * machCacheFindEntry(struct machCache * cache,
     return NULL;
 }
 
-static int machCompatCacheAdd(char * name, char * fn, int linenum,
-				struct machCache * cache) {
+static int machCompatCacheAdd(char * name, const char * fn, int linenum,
+				struct machCache * cache)
+{
     char * chptr, * equivs;
     int delEntry = 0;
     int i;
     struct machCacheEntry * entry = NULL;
 
     while (*name && isspace(*name)) name++;
-  
+
     chptr = name;
     while (*chptr && *chptr != ':') chptr++;
     if (!*chptr) {
 	rpmError(RPMERR_RPMRC, _("missing second ':' at %s:%d"), fn, linenum);
 	return 1;
     } else if (chptr == name) {
-	rpmError(RPMERR_RPMRC, _("missing architecture name at %s:%d"), fn, 
+	rpmError(RPMERR_RPMRC, _("missing architecture name at %s:%d"), fn,
 			     linenum);
 	return 1;
     }
@@ -237,7 +202,7 @@ static int machCompatCacheAdd(char * name, char * fn, int linenum,
     }
 
     if (!entry) {
-	cache->cache = realloc(cache->cache, 
+	cache->cache = realloc(cache->cache,
 			       (cache->size + 1) * sizeof(*cache->cache));
 	entry = cache->cache + cache->size++;
 	entry->name = strdup(name);
@@ -246,28 +211,27 @@ static int machCompatCacheAdd(char * name, char * fn, int linenum,
     }
 
     if (delEntry) return 0;
-	
-    chptr = strtok(equivs, " ");
-    while (chptr) {
-	if (strlen(chptr)) {		/* does strtok() return "" ever?? */
-	    if (entry->count)
-		entry->equivs = realloc(entry->equivs, sizeof(*entry->equivs) 
+
+    while ((chptr = strtok(equivs, " ")) != NULL) {
+	equivs = NULL;
+	if (chptr[0] == '\0')	/* does strtok() return "" ever?? */
+	    continue;
+	if (entry->count)
+	    entry->equivs = realloc(entry->equivs, sizeof(*entry->equivs)
 					* (entry->count + 1));
-	    else
-		entry->equivs = malloc(sizeof(*entry->equivs));
+	else
+	    entry->equivs = malloc(sizeof(*entry->equivs));
 
-	    entry->equivs[entry->count] = strdup(chptr);
-	    entry->count++;
-	}
-
-	chptr = strtok(NULL, " ");
+	entry->equivs[entry->count] = strdup(chptr);
+	entry->count++;
     }
 
     return 0;
 }
 
 static struct machEquivInfo * machEquivSearch(
-		struct machEquivTable * table, char * name) {
+		struct machEquivTable * table, char * name)
+{
     int i;
 
 /*
@@ -278,16 +242,16 @@ static struct machEquivInfo * machEquivSearch(
  * XXX used by rpmInstallPackage->{os,arch}Okay->rpmMachineScore->.
  * XXX to verify correct arch/os from headers.
  */
-
     for (i = 0; i < table->count; i++)
-	if (!strcasecmp(table->list[i].name, name)) 
+	if (!strcasecmp(table->list[i].name, name))
 	    return table->list + i;
 
     return NULL;
 }
 
 static void machAddEquiv(struct machEquivTable * table, char * name,
-			   int distance) {
+			   int distance)
+{
     struct machEquivInfo * equiv;
 
     equiv = machEquivSearch(table, name);
@@ -297,16 +261,17 @@ static void machAddEquiv(struct machEquivTable * table, char * name,
 				    * sizeof(*table->list));
 	else
 	    table->list = malloc(sizeof(*table->list));
-    
+
 	table->list[table->count].name = strdup(name);
 	table->list[table->count++].score = distance;
     }
 }
 
-static void machCacheEntryVisit(struct machCache * cache, 
-				  struct machEquivTable * table, 
+static void machCacheEntryVisit(struct machCache * cache,
+				  struct machEquivTable * table,
 				  char * name,
-	  			  int distance) {
+	  			  int distance)
+{
     struct machCacheEntry * entry;
     int i;
 
@@ -324,9 +289,10 @@ static void machCacheEntryVisit(struct machCache * cache,
     }
 }
 
-static void machFindEquivs(struct machCache * cache, 
+static void machFindEquivs(struct machCache * cache,
 			     struct machEquivTable * table,
-			     char * key) {
+			     char * key)
+{
     int i;
 
     for (i = 0; i < cache->size; i++)
@@ -342,10 +308,11 @@ static void machFindEquivs(struct machCache * cache,
 }
 
 static int addCanon(struct canonEntry **table, int *tableLen, char *line,
-		    char *fn, int lineNum) {
+		    const char *fn, int lineNum)
+{
     struct canonEntry *t;
     char *s, *s1;
-    
+
     if (! *tableLen) {
 	*tableLen = 2;
 	*table = malloc(2 * sizeof(struct canonEntry));
@@ -388,9 +355,10 @@ static int addCanon(struct canonEntry **table, int *tableLen, char *line,
 }
 
 static int addDefault(struct defaultEntry **table, int *tableLen, char *line,
-		      char *fn, int lineNum) {
+			const char *fn, int lineNum)
+{
     struct defaultEntry *t;
-    
+
     if (! *tableLen) {
 	*tableLen = 1;
 	*table = malloc(sizeof(struct defaultEntry));
@@ -403,7 +371,7 @@ static int addDefault(struct defaultEntry **table, int *tableLen, char *line,
     t->name = strtok(line, ": \t");
     t->defName = strtok(NULL, " \t");
     if (! (t->name && t->defName)) {
-	rpmError(RPMERR_RPMRC, _("Incomplete default line at %s:%d"), 
+	rpmError(RPMERR_RPMRC, _("Incomplete default line at %s:%d"),
 		 fn, lineNum);
 	return RPMERR_RPMRC;
     }
@@ -432,7 +400,7 @@ static struct canonEntry *lookupInCanonTable(char *name,
     return NULL;
 }
 
-static char *lookupInDefaultTable(char *name, struct defaultEntry *table,
+static const char *lookupInDefaultTable(const char *name, struct defaultEntry *table,
 				  int tableLen) {
     while (tableLen) {
 	tableLen--;
@@ -444,48 +412,126 @@ static char *lookupInDefaultTable(char *name, struct defaultEntry *table,
     return name;
 }
 
-int rpmReadConfigFiles(char * file, char * arch, char * os, int building) {
+int rpmReadConfigFiles(const char * file, const char * target)
+{
+
+    /* Preset target macros */
+    rpmRebuildTargetVars(&target, NULL);
+
+    /* Read the files */
     if (rpmReadRC(file)) return -1;
 
-    if (building)
-	rpmSetTables(RPM_MACHTABLE_BUILDARCH, RPM_MACHTABLE_BUILDOS);
+    /* Reset target macros */
+    rpmRebuildTargetVars(&target, NULL);
 
-    rpmSetMachine(arch, os);
+    /* Finally set target platform */
+    {	const char *cpu = rpmExpand("%{_target_cpu}", NULL);
+	const char *os = rpmExpand("%{_target_os}", NULL);
+	rpmSetMachine(cpu, os);
+	xfree(cpu);
+	xfree(os);
+    }
 
     return 0;
 }
 
-static void setPathDefault(int var, char * s) {
-    char * topdir;
-    char * fn;
-
-    if (rpmGetVar(var)) return;
-
-    topdir = rpmGetVar(RPMVAR_TOPDIR);
-    if (!topdir) topdir = rpmGetVar(RPMVAR_TMPPATH);
-	
-    fn = alloca(strlen(topdir) + strlen(s) + 2);
-    strcpy(fn, topdir);
-    if (fn[strlen(topdir) - 1] != '/')
-	strcat(fn, "/");
-    strcat(fn, s);
-   
-    rpmSetVar(var, fn);
+static void setVarDefault(int var, const char *macroname, const char *val, const char *body)
+{
+    if (var >= 0) {	/* XXX Dying ... */
+	if (rpmGetVar(var)) return;
+	rpmSetVar(var, val);
+    }
+    if (body == NULL)
+	body = val;
+    addMacro(NULL, macroname, NULL, body, RMIL_DEFAULT);
 }
 
+static void setPathDefault(int var, const char *macroname, const char *subdir)
+{
+
+    if (var >= 0) {	/* XXX Dying ... */
+	const char * topdir;
+	char * fn;
+
+	if (rpmGetVar(var)) return;
+
+	topdir = rpmGetPath("%{_topdir}", NULL);
+
+	fn = alloca(strlen(topdir) + strlen(subdir) + 2);
+	strcpy(fn, topdir);
+	if (fn[strlen(topdir) - 1] != '/')
+	    strcat(fn, "/");
+	strcat(fn, subdir);
+
+	rpmSetVar(var, fn);
+	if (topdir)	xfree(topdir);
+    }
+
+    if (macroname != NULL) {
+#define	_TOPDIRMACRO	"%{_topdir}/"
+	char *body = alloca(sizeof(_TOPDIRMACRO) + strlen(subdir));
+	strcpy(body, _TOPDIRMACRO);
+	strcat(body, subdir);
+	addMacro(NULL, macroname, NULL, body, RMIL_DEFAULT);
+#undef _TOPDIRMACRO
+    }
+}
+
+static const char *prescriptenviron = "\n\
+RPM_SOURCE_DIR=\"%{_sourcedir}\"\n\
+RPM_BUILD_DIR=\"%{_builddir}\"\n\
+RPM_OPT_FLAGS=\"%{optflags}\"\n\
+RPM_ARCH=\"%{_arch}\"\n\
+RPM_OS=\"%{_os}\"\n\
+export RPM_SOURCE_DIR RPM_BUILD_DIR RPM_OPT_FLAGS RPM_ARCH RPM_OS\n\
+RPM_DOC_DIR=\"%{_docdir}\"\n\
+export RPM_DOC_DIR\n\
+RPM_PACKAGE_NAME=\"%{name}\"\n\
+RPM_PACKAGE_VERSION=\"%{version}\"\n\
+RPM_PACKAGE_RELEASE=\"%{release}\"\n\
+export RPM_PACKAGE_NAME RPM_PACKAGE_VERSION RPM_PACKAGE_RELEASE\n\
+%{?buildroot:RPM_BUILD_ROOT=\"%{buildroot}\"\n\
+export RPM_BUILD_ROOT\n}\
+";
+
 static void setDefaults(void) {
-    rpmSetVar(RPMVAR_OPTFLAGS, "-O2");
-    rpmSetVar(RPMVAR_SIGTYPE, "none");
-    rpmSetVar(RPMVAR_DEFAULTDOCDIR, "/usr/doc");
-    rpmSetVar(RPMVAR_TOPDIR, "/usr/src/redhat");
-    rpmSetVar(RPMVAR_BUILDSHELL, "/bin/sh");
+
+    addMacro(NULL, "_usr", NULL, "/usr", RMIL_DEFAULT);
+    addMacro(NULL, "_var", NULL, "/var", RMIL_DEFAULT);
+
+    addMacro(NULL, "_preScriptEnvironment",NULL, prescriptenviron,RMIL_DEFAULT);
+
+    setVarDefault(-1,			"_topdir",
+		"/usr/src/redhat",	"%{_usr}/src/redhat");
+    setVarDefault(-1,			"_tmppath",
+		"/var/tmp",		"%{_var}/tmp");
+    setVarDefault(-1,			"_dbpath",
+		"/var/lib/rpm",		"%{_var}/lib/rpm");
+    setVarDefault(-1,			"_defaultdocdir",
+		"/usr/doc",		"%{_usr}/doc");
+
+    setVarDefault(-1,			"_rpmfilename",
+	"%%{ARCH}/%%{NAME}-%%{VERSION}-%%{RELEASE}.%%{ARCH}.rpm",NULL);
+
+    setVarDefault(RPMVAR_OPTFLAGS,	"optflags",
+		"-O2",			NULL);
+    setVarDefault(-1,			"sigtype",
+		"none",			NULL);
+    setVarDefault(-1,			"_buildshell",
+		"/bin/sh",		NULL);
+
+    setPathDefault(-1,			"_builddir",	"BUILD");
+    setPathDefault(-1,			"_rpmdir",	"RPMS");
+    setPathDefault(-1,			"_srcrpmdir",	"SRPMS");
+    setPathDefault(-1,			"_sourcedir",	"SOURCES");
+    setPathDefault(-1,			"_specdir",	"SPECS");
+
 }
 
-int rpmReadRC(char * file) {
-    int fd;
-    char * fn;
-    char * home;
-    int rc = 0;
+int rpmReadRC(const char * rcfiles)
+{
+    char *myrcfiles, *r, *re;
+    int rc;
     static int first = 1;
 
     if (first) {
@@ -493,183 +539,255 @@ int rpmReadRC(char * file) {
 	first = 0;
     }
 
-    fd = open(LIBRPMRC_FILENAME, O_RDONLY);
-    if (fd >= 0) {
-	rc = doReadRC(fd, LIBRPMRC_FILENAME);
-	close(fd);
-	if (rc) return rc;
-    } else {
-	rpmError(RPMERR_RPMRC, _("Unable to open %s for reading: %s."), 
-		 LIBRPMRC_FILENAME, strerror(errno));
-	return 1;
-    }
-    
-    if (file) 
-	fn = file;
-    else
-	fn = "/etc/rpmrc";
+    if (rcfiles == NULL)
+	rcfiles = defrcfiles;
 
-    fd = open(fn, O_RDONLY);
-    if (fd >= 0) {
-	rc = doReadRC(fd, fn);
-	close(fd);
-	if (rc) return rc;
-    } else if (file) {
-	rpmError(RPMERR_RPMRC, _("Unable to open %s for reading: %s."), file,
-		 strerror(errno));
-	return 1;
-    }
+    /* Read each file in rcfiles. */
+    rc = 0;
+    for (r = myrcfiles = strdup(rcfiles); *r != '\0'; r = re) {
+	char fn[4096];
+	FD_t fd;
 
-    if (!file) {
-	home = getenv("HOME");
-	if (home) {
-	    fn = alloca(strlen(home) + 8);
-	    strcpy(fn, home);
-	    strcat(fn, "/.rpmrc");
-	    fd = open(fn, O_RDONLY);
-	    if (fd >= 0) {
-		rc |= doReadRC(fd, fn);
-		close(fd);
-		if (rc) return rc;
+	/* Get pointer to rest of files */
+	if ((re = strchr(r, ':')) != NULL)
+	    *re++ = '\0';
+	else
+	    re = r + strlen(r);
+
+	/* Expand ~/ to $HOME/ */
+	fn[0] = '\0';
+	if (r[0] == '~' && r[1] == '/') {
+	    char *home = getenv("HOME");
+	    if (home == NULL) {
+		rpmError(RPMERR_RPMRC, _("Cannot expand %s"), r);
+		rc = 1;
+		break;
 	    }
+	    strcpy(fn, home);
+	    r++;
 	}
+	strcat(fn, r);
+
+	/* Read another rcfile */
+	fd = fdOpen(fn, O_RDONLY, 0);
+	if (fdFileno(fd) < 0) {
+	    /* XXX Only /usr/lib/rpm/rpmrc must exist in default rcfiles list */
+	    if (rcfiles == defrcfiles && myrcfiles != r)
+		continue;
+	    rpmError(RPMERR_RPMRC, _("Unable to open %s for reading: %s."),
+		 fn, strerror(errno));
+	    rc = 1;
+	    break;
+	}
+	rc = doReadRC(fd, fn);
+	fdClose(fd);
+ 	if (rc) break;
     }
+    if (myrcfiles)	free(myrcfiles);
+    if (rc)
+	return rc;
 
-    rpmSetMachine(NULL, NULL);
+    rpmSetMachine(NULL, NULL);	/* XXX WTFO? Why bother? */
 
-    setPathDefault(RPMVAR_BUILDDIR, "BUILD");    
-    setPathDefault(RPMVAR_RPMDIR, "RPMS");    
-    setPathDefault(RPMVAR_SRPMDIR, "SRPMS");    
-    setPathDefault(RPMVAR_SOURCEDIR, "SOURCES");    
-    setPathDefault(RPMVAR_SPECDIR, "SPECS");
+    if ((r = rpmGetVar(RPMVAR_MACROFILES)) != NULL)
+	initMacros(NULL, r);
 
-    return 0;
+    return rc;
 }
 
-static int doReadRC(int fd, char * filename) {
-    struct stat sb;
-    char * buf;
-    char * start, * chptr, * next, * rest;
+static int doReadRC(FD_t fd, const char * filename)
+{
+    char *s, *se, *next;
     int linenum = 0;
     struct rpmOption searchOption, * option;
-    int i;
-    int gotit;
+    int rc;
 
-    fstat(fd, &sb);
-    next = buf = alloca(sb.st_size + 2);
-    if (read(fd, buf, sb.st_size) != sb.st_size) {
+  { struct stat sb;
+    fstat(fdFileno(fd), &sb);
+    next = alloca(sb.st_size + 2);
+    if (fdRead(fd, next, sb.st_size) != sb.st_size) {
 	rpmError(RPMERR_RPMRC, _("Failed to read %s: %s."), filename,
 		 strerror(errno));
 	return 1;
     }
-    buf[sb.st_size] = '\n';
-    buf[sb.st_size + 1] = '\0';
+    next[sb.st_size] = '\n';
+    next[sb.st_size + 1] = '\0';
+  }
 
     while (*next) {
 	linenum++;
 
-	chptr = start = next;
-	while (*chptr != '\n') chptr++;
+	se = s = next;
+	while (*se && *se != '\n') se++;
+	if (*se) *se++ = '\0';
+	next = se;
 
-	*chptr = '\0';
-	next = chptr + 1;
-
-	while (isspace(*start)) start++;
+	while (*s && isspace(*s)) s++;
 
 	/* we used to allow comments to begin anywhere, but not anymore */
-	if (*start == '#' || !*start) continue;
+	if (*s == '#' || *s == '\0') continue;
 
-	chptr = start;
-	while (!isspace(*chptr) && *chptr != ':' && *chptr) chptr++;
+	se = s;
+	while (*se && !isspace(*se) && *se != ':') se++;
 
-	if (isspace(*chptr)) {
-	    *chptr++ = '\0';
-	    while (isspace(*chptr) && *chptr != ':' && *chptr) chptr++;
+	if (isspace(*se)) {
+	    *se++ = '\0';
+	    while (*se && isspace(*se) && *se != ':') se++;
 	}
 
-	if (*chptr != ':') {
-	    rpmError(RPMERR_RPMRC, _("missing ':' at %s:%d"), 
+	if (*se != ':') {
+	    rpmError(RPMERR_RPMRC, _("missing ':' at %s:%d"),
 		     filename, linenum);
 	    return 1;
 	}
+	*se++ = '\0';
+	while (*se && isspace(*se)) se++;
 
-	*chptr++ = '\0';
-
-	searchOption.name = start;
+	/* Find keyword in table */
+	searchOption.name = s;
 	option = bsearch(&searchOption, optionTable, optionTableSize,
 			 sizeof(struct rpmOption), optionCompare);
 
-	if (option) {
-	    start = chptr;
-	    while (isspace(*start) && *start) start++;
+	if (option) {	/* For configuration variables  ... */
+	    const char *arch, *val, *fn;
 
-	    if (! *start) {
-		rpmError(RPMERR_RPMRC, _("missing argument for %s at %s:%d"), 
+	    arch = val = fn = NULL;
+	    if (*se == '\0') {
+		rpmError(RPMERR_RPMRC, _("missing argument for %s at %s:%d"),
 		      option->name, filename, linenum);
 		return 1;
 	    }
 
-	    if (option->archSpecific) {
-		chptr = start;
-		while (!isspace(*chptr) && *chptr) chptr++;
+	    switch (option->var) {
+	    case RPMVAR_INCLUDE:
+	      {	FD_t fdinc;
 
-		if (!*chptr) {
-		    rpmError(RPMERR_RPMRC, 
-				_("missing architecture for %s at %s:%d"), 
-			  	option->name, filename, linenum);
+		s = se;
+		while (*se && !isspace(*se)) se++;
+		if (*se) *se++ = '\0';
+
+		rpmRebuildTargetVars(NULL, NULL);
+
+		fn = rpmGetPath(s, NULL);
+		if (fn == NULL || *fn == '\0') {
+		    rpmError(RPMERR_RPMRC, _("%s expansion failed at %s:%d \"%s\""),
+			option->name, filename, linenum, s);
+		    if (fn) xfree(fn);
 		    return 1;
 		}
 
-		*chptr++ = '\0';
-
-		while (isspace(*chptr) && *chptr) chptr++;
-		if (!*chptr) {
-		    rpmError(RPMERR_RPMRC, 
-				_("missing argument for %s at %s:%d"), 
-			  	option->name, filename, linenum);
+		fdinc = fdOpen(fn, O_RDONLY, 0);
+		if (fdFileno(fdinc) < 0) {
+		    rpmError(RPMERR_RPMRC, _("cannot open %s at %s:%d"),
+			fn, filename, linenum);
+		    rc = 1;
+		} else {
+		    rc = doReadRC(fdinc, fn);
+		    fdClose(fdinc);
+		}
+		if (fn) xfree(fn);
+		if (rc) return rc;
+		continue;	/* XXX don't save include value as var/macro */
+	      }	break;
+	    case RPMVAR_MACROFILES:
+		fn = rpmGetPath(se, NULL);
+		if (fn == NULL || *fn == '\0') {
+		    rpmError(RPMERR_RPMRC, _("%s expansion failed at %s:%d \"%s\""),
+			option->name, filename, linenum, fn);
+		    if (fn) xfree(fn);
 		    return 1;
 		}
-		
-		rpmSetVarArch(option->var, chptr, start);
-	    } else {
-		rpmSetVarArch(option->var, start, NULL);
+		se = (char *)fn;
+		break;
+	    case RPMVAR_PROVIDES:
+	      {	char *t;
+		s = rpmGetVar(RPMVAR_PROVIDES);
+		if (s == NULL) s = "";
+		fn = t = malloc(strlen(s) + strlen(se) + 2);
+		while (*s) *t++ = *s++;
+		*t++ = ' ';
+		while (*se) *t++ = *se++;
+		*t++ = '\0';
+		se = (char *)fn;
+	      }	break;
+	    default:
+		break;
 	    }
-	} else {
+
+	    if (option->archSpecific) {
+		arch = se;
+		while (*se && !isspace(*se)) se++;
+		if (*se == '\0') {
+		    rpmError(RPMERR_RPMRC,
+				_("missing architecture for %s at %s:%d"),
+			  	option->name, filename, linenum);
+		    return 1;
+		}
+		*se++ = '\0';
+		while (*se && isspace(*se)) se++;
+		if (*se == '\0') {
+		    rpmError(RPMERR_RPMRC,
+				_("missing argument for %s at %s:%d"),
+			  	option->name, filename, linenum);
+		    return 1;
+		}
+	    }
+	
+	    val = se;
+
+	    /* Only add macros if appropriate for this arch */
+	    if (option->macroize &&
+	      (arch == NULL || !strcmp(arch, current[ARCH]))) {
+		char *n, *name;
+		n = name = malloc(strlen(option->name)+2);
+		if (option->localize)
+		    *n++ = '_';
+		strcpy(n, option->name);
+		addMacro(NULL, name, NULL, val, RMIL_RPMRC);
+		free(name);
+	    }
+	    rpmSetVarArch(option->var, val, arch);
+	    if (fn) xfree(fn);
+
+	} else {	/* For arch/os compatibilty tables ... */
+	    int gotit;
+	    int i;
+
 	    gotit = 0;
 
 	    for (i = 0; i < RPM_MACHTABLE_COUNT; i++) {
-		if (!strncmp(tables[i].key, start, strlen(tables[i].key)))
+		if (!strncmp(tables[i].key, s, strlen(tables[i].key)))
 		    break;
 	    }
 
 	    if (i < RPM_MACHTABLE_COUNT) {
-		rest = start + strlen(tables[i].key);
+		char *rest = s + strlen(tables[i].key);
 		if (*rest == '_') rest++;
 
 		if (!strcmp(rest, "compat")) {
-		    if (machCompatCacheAdd(chptr, filename, linenum, 
+		    if (machCompatCacheAdd(se, filename, linenum,
 						&tables[i].cache))
 			return 1;
 		    gotit = 1;
 		} else if (tables[i].hasTranslate &&
 			   !strcmp(rest, "translate")) {
-		    if (addDefault(&tables[i].defaults, 
+		    if (addDefault(&tables[i].defaults,
 				   &tables[i].defaultsLength,
-				   chptr, filename, linenum))
+				   se, filename, linenum))
 			return 1;
 		    gotit = 1;
 		} else if (tables[i].hasCanon &&
 			   !strcmp(rest, "canon")) {
 		    if (addCanon(&tables[i].canons, &tables[i].canonsLength,
-				 chptr, filename, linenum))
+				 se, filename, linenum))
 			return 1;
 		    gotit = 1;
 		}
 	    }
 
 	    if (!gotit) {
-		rpmError(RPMERR_RPMRC, _("bad option '%s' at %s:%d"), 
-			    start, filename, linenum);
+		rpmError(RPMERR_RPMRC, _("bad option '%s' at %s:%d"),
+			    s, filename, linenum);
 	    }
 	}
     }
@@ -682,7 +800,7 @@ static void defaultMachine(char ** arch, char ** os) {
     static int gotDefaults = 0;
     char * chptr;
     struct canonEntry * canon;
-    
+
     if (!gotDefaults) {
 	uname(&un);
 
@@ -707,7 +825,7 @@ static void defaultMachine(char ** arch, char ** os) {
                  }
               sprintf(un.sysname,"sunos%s",un.release);
            }
-           
+
            else /* Solaris 2.x: n.x.x becomes n-3.x.x */
               sprintf(un.sysname,"solaris%1d%s",atoi(un.release)-3,un.release+1+(atoi(un.release)/10));
         }
@@ -727,15 +845,14 @@ static void defaultMachine(char ** arch, char ** os) {
         else if ((!strncmp(un.machine, "34", 2) || \
                  !strncmp(un.machine, "33", 2)) && \
                  !strncmp(un.release, "4.0", 3)) {
-	   char * prelid=NULL;
-	   int fd, irelid;
            /* we are on ncr-sysv4 */
-           fd = open("/etc/.relid", O_RDONLY, 0700);
-           if (fd >0) {
+	   char *prelid = NULL;
+           FD_t fd = fdOpen("/etc/.relid", O_RDONLY, 0700);
+           if (fdFileno(fd) > 0) {
               chptr = (char *) calloc(256,1);
               if (chptr != NULL) {
-                 irelid = read(fd, (void *)chptr, 256);
-                 close(fd);
+                 int irelid = fdRead(fd, (void *)chptr, 256);
+                 fdClose(fd);
                  /* example: "112393 RELEASE 020200 Version 01 OS" */
                  if (irelid > 0) {
                     if ((prelid=strstr(chptr, "RELEASE "))){
@@ -745,18 +862,17 @@ static void defaultMachine(char ** arch, char ** os) {
 		 }
                  free (chptr);
               }
-           }           
+           }
            if (!prelid)
               /* parsing /etc/.relid file failed */
               strcpy(un.sysname,"ncr-sysv4");
            /* wrong, just for now, find out how to look for i586 later*/
            strcpy(un.machine,"i486");
         }
-#endif
-                   
-	/* get rid of the slashes in the sysname */
-	chptr = un.machine;
-	while (*chptr++)
+#endif	/* __linux__ */
+
+	/* get rid of the hyphens in the sysname */
+	for (chptr = un.machine; *chptr; chptr++)
 	    if (*chptr == '/') *chptr = '-';
 
 #	if defined(__MIPSEL__) || defined(__MIPSEL) || defined(_MIPSEL)
@@ -807,7 +923,7 @@ static void defaultMachine(char ** arch, char ** os) {
 		    strcpy(un.machine, "hppa2.0");
 #	    endif
 	}
-#	endif
+#	endif	/* hpux */
 
 	/* the uname() result goes through the arch_canon table */
 	canon = lookupInCanonTable(un.machine,
@@ -821,6 +937,7 @@ static void defaultMachine(char ** arch, char ** os) {
 				   tables[RPM_MACHTABLE_INSTOS].canonsLength);
 	if (canon)
 	    strcpy(un.sysname, canon->short_name);
+	gotDefaults = 1;
     }
 
     if (arch) *arch = un.machine;
@@ -851,51 +968,32 @@ char *rpmGetVar(int var)
     return rpmGetVarArch(var, NULL);
 }
 
-int rpmGetBooleanVar(int var) {
-    char * val;
-    int num;
-    char * chptr;
-
-    val = rpmGetVar(var);
-    if (!val) return 0;
-
-    if (val[0] == 'y' || val[0] == 'Y') return 1;
-
-    num = strtol(val, &chptr, 0);
-    if (chptr && *chptr == '\0') {
-	return num != 0;
-    }
-
-    return 0;
-}
-
-void rpmSetVar(int var, char *val) {
-    freeRpmVar(&values[var]);
-   
-    values[var].arch = NULL;
-    values[var].next = NULL;
-
-    if (val) 
-	values[var].value = strdup(val);
-    else
-	values[var].value = NULL;
-}
-
 /* this doesn't free the passed pointer! */
 static void freeRpmVar(struct rpmvarValue * orig) {
     struct rpmvarValue * next, * var = orig;
 
     while (var) {
 	next = var->next;
-	if (var->arch) free(var->arch);
-	if (var->value) free(var->value);
+	if (var->arch) {
+	    free(var->arch);
+	    var->arch = NULL;
+	}
+	if (var->value) {
+	    free(var->value);
+	    var->value = NULL;
+	}
 
 	if (var != orig) free(var);
 	var = next;
     }
 }
 
-static void rpmSetVarArch(int var, char * val, char * arch) {
+void rpmSetVar(int var, const char *val) {
+    freeRpmVar(&values[var]);
+    values[var].value = (val ? strdup(val) : NULL);
+}
+
+static void rpmSetVarArch(int var, const char * val, const char * arch) {
     struct rpmvarValue * next = values + var;
 
     if (next->value) {
@@ -922,10 +1020,7 @@ static void rpmSetVarArch(int var, char * val, char * arch) {
     }
 
     next->value = strdup(val);
-    if (arch)
-	next->arch = strdup(arch);
-    else
-	next->arch = NULL;
+    next->arch = (arch ? strdup(arch) : NULL);
 }
 
 void rpmSetTables(int archTable, int osTable) {
@@ -945,57 +1040,60 @@ void rpmSetTables(int archTable, int osTable) {
 }
 
 int rpmMachineScore(int type, char * name) {
-    struct machEquivInfo * info;
-
-    info = machEquivSearch(&tables[type].equiv, name);
-    if (info) 
-	return info->score;
-    else
-	return 0;
+    struct machEquivInfo * info = machEquivSearch(&tables[type].equiv, name);
+    return (info != NULL ? info->score : 0);
 }
 
 void rpmGetMachine(char **arch, char **os)
 {
-    if (arch) {
+    if (arch)
 	*arch = current[ARCH];
-    }
 
-    if (os) {
+    if (os)
 	*os = current[OS];
-    }
 }
 
-void rpmSetMachine(char * arch, char * os) {
-    int transOs = os == NULL;
-    int transArch = arch == NULL;
-    char * realArch, * realOs;
+void rpmSetMachine(const char * arch, const char * os) {
+    char * host_cpu, * host_os;
 
-    defaultMachine(&realArch, &realOs);
+    defaultMachine(&host_cpu, &host_os);
 
-    if (!arch)
-	arch = realArch;
-    if (!os)
-	os = realOs;
-
-    if (transArch && tables[currTables[ARCH]].hasTranslate)
-	arch = lookupInDefaultTable(arch,
+    if (arch == NULL) {
+	arch = host_cpu;
+	if (tables[currTables[ARCH]].hasTranslate)
+	    arch = lookupInDefaultTable(arch,
 			    tables[currTables[ARCH]].defaults,
 			    tables[currTables[ARCH]].defaultsLength);
-    if (transOs && tables[currTables[OS]].hasTranslate)
-	os = lookupInDefaultTable(os,
+    }
+
+    if (os == NULL) {
+	os = host_os;
+	if (tables[currTables[OS]].hasTranslate)
+	    os = lookupInDefaultTable(os,
 			    tables[currTables[OS]].defaults,
 			    tables[currTables[OS]].defaultsLength);
+    }
 
     if (!current[ARCH] || strcmp(arch, current[ARCH])) {
 	if (current[ARCH]) free(current[ARCH]);
 	current[ARCH] = strdup(arch);
-	rebuildCompatTables(ARCH, realArch);
+	rebuildCompatTables(ARCH, host_cpu);
     }
 
     if (!current[OS] || strcmp(os, current[OS])) {
 	if (current[OS]) free(current[OS]);
 	current[OS] = strdup(os);
-	rebuildCompatTables(OS, realOs);
+	/*
+	 * XXX Capitalizing the 'L' is needed to insure that old
+	 * XXX os-from-uname (e.g. "Linux") is compatible with the new
+	 * XXX os-from-platform (e.g "linux" from "sparc-*-linux").
+	 * XXX A copy of this string is embedded in headers and is
+	 * XXX used by rpmInstallPackage->{os,arch}Okay->rpmMachineScore->
+	 * XXX to verify correct arch/os from headers.
+	 */
+	if (!strcmp(current[OS], "linux"))
+	    *current[OS]= 'L';
+	rebuildCompatTables(OS, host_os);
     }
 }
 
@@ -1012,7 +1110,7 @@ static void getMachineInfo(int type, char ** name, int * num) {
     /* use the normal canon tables, even if we're looking up build stuff */
     if (which >= 2) which -= 2;
 
-    canon = lookupInCanonTable(current[type], 
+    canon = lookupInCanonTable(current[type],
 			       tables[which].canons,
 			       tables[which].canonsLength);
 
@@ -1038,11 +1136,97 @@ void rpmGetOsInfo(char ** name, int * num) {
     getMachineInfo(OS, name, num);
 }
 
+void rpmRebuildTargetVars(const char **buildtarget, const char ** canontarget)
+{
+
+    char *ca = NULL, *co = NULL, *ct = NULL;
+    int x;
+
+    /* Rebuild the compat table to recalculate the current target arch.  */
+
+    rpmSetMachine(NULL, NULL);
+    rpmSetTables(RPM_MACHTABLE_INSTARCH, RPM_MACHTABLE_INSTOS);
+    rpmSetTables(RPM_MACHTABLE_BUILDARCH, RPM_MACHTABLE_BUILDOS);
+
+    if (buildtarget && *buildtarget) {
+	char *c;
+	/* Set arch and os from specified build target */
+	ca = strdup(*buildtarget);
+	if ((c = strchr(ca, '-')) != NULL) {
+	    *c++ = '\0';
+	    
+	    if ((co = strrchr(c, '-')) == NULL) {
+		co = c;
+	    } else {
+		if (!strcasecmp(co, "-gnu"))
+		    *co = '\0';
+		if ((co = strrchr(c, '-')) == NULL)
+		    co = c;
+		else
+		    co++;
+	    }
+	    if (co != NULL) co = strdup(co);
+	}
+    } else {
+	/* Set build target from rpm arch and os */
+	rpmGetArchInfo(&ca,NULL);
+	if (ca)	ca = strdup(ca);
+	rpmGetOsInfo(&co,NULL);
+	if (co)	co = strdup(co);
+    }
+
+    /* If still not set, Set target arch/os from default uname(2) values */
+    if (ca == NULL) {
+	defaultMachine(&ca, NULL);
+	ca = strdup(ca);
+     }
+    for (x = 0; ca[x]; x++)
+	ca[x] = tolower(ca[x]);
+
+    if (co == NULL) {
+	defaultMachine(NULL, &co);
+	co = strdup(co);
+    }
+    for (x = 0; co[x]; x++)
+	co[x] = tolower(co[x]);
+
+    /* XXX For now, set canonical target to arch-os */
+    if (ct == NULL) {
+	ct = malloc(strlen(ca) + sizeof("-") + strlen(co));
+	sprintf(ct, "%s-%s", ca, co);
+    }
+
+/*
+ * XXX All this macro pokery/jiggery could be achieved by doing a delayed
+ *	initMacros(NULL, PER-PLATFORM-MACRO-FILE-NAMES);
+ */
+    delMacro(NULL, "_target");
+    addMacro(NULL, "_target", NULL, ct, RMIL_RPMRC);
+    delMacro(NULL, "_target_cpu");
+    addMacro(NULL, "_target_cpu", NULL, ca, RMIL_RPMRC);
+    delMacro(NULL, "_target_os");
+    addMacro(NULL, "_target_os", NULL, co, RMIL_RPMRC);
+/*
+ * XXX Make sure that per-arch optflags is initialized correctly.
+ */
+  { char *optflags = rpmGetVarArch(RPMVAR_OPTFLAGS, ca);
+    if (optflags != NULL) {
+	delMacro(NULL, "optflags");
+	addMacro(NULL, "optflags", NULL, optflags, RMIL_RPMRC);
+    }
+  }
+
+    if (canontarget)
+	*canontarget = ct;
+    else
+	free(ct);
+    free(ca);
+    free(co);
+}
+
 int rpmShowRC(FILE *f)
 {
     struct rpmOption *opt;
-    int count = 0;
-    char *s;
     int i;
     struct machEquivTable * equivTable;
 
@@ -1065,7 +1249,7 @@ int rpmShowRC(FILE *f)
     fprintf(f, "\n");
 
     rpmSetTables(RPM_MACHTABLE_INSTARCH, RPM_MACHTABLE_INSTOS);
-    rpmSetMachine(NULL, NULL);
+    rpmSetMachine(NULL, NULL);	/* XXX WTFO? Why bother? */
 
     fprintf(f, "install arch          : %s\n", current[ARCH]);
     fprintf(f, "install os            : %s\n", current[OS]);
@@ -1082,14 +1266,14 @@ int rpmShowRC(FILE *f)
 	fprintf(f," %s", equivTable->list[i].name);
     fprintf(f, "\n");
 
-    fprintf(f, "RPMRC VALUES:\n");
-    opt = optionTable;
-    while (count < optionTableSize) {
-	s = rpmGetVar(opt->var);
-	fprintf(f, "%-21s : %s\n", opt->name, s ? s : "(not set)");
-	opt++;
-	count++;
+    fprintf(f, "\nRPMRC VALUES:\n");
+    for (i = 0, opt = optionTable; i < optionTableSize; i++, opt++) {
+	char *s = rpmGetVar(opt->var);
+	if (s != NULL || rpmGetVerbosity() < RPMMESS_NORMAL)
+	    fprintf(f, "%-21s : %s\n", opt->name, s ? s : "(not set)");
     }
-    
+
+    dumpMacroTable(NULL, f);
+
     return 0;
 }
