@@ -1,4 +1,5 @@
 #ifndef H_SIGNATURE
+#define	H_SIGNATURE
 
 /* signature.h - generate and verify signatures */
 
@@ -33,19 +34,27 @@
 Header rpmNewSignature(void);
 
 /* If an old-style signature is found, we emulate a new style one */
-int rpmReadSignature(int fd, Header *header, short sig_type);
-int rpmWriteSignature(int fd, Header header);
+int rpmReadSignature(FD_t fd, Header *header, short sig_type);
+int rpmWriteSignature(FD_t fd, Header header);
 
 /* Generate a signature of data in file, insert in header */
-int rpmAddSignature(Header header, char *file,
-		    int_32 sigTag, char *passPhrase);
+int rpmAddSignature(Header header, const char *file,
+		    int_32 sigTag, const char *passPhrase);
 
 /******************************************************************/
 
+/* Possible actions for rpmLookupSignatureType() */
+#define RPMLOOKUPSIG_QUERY	0	/* Lookup type in effect          */
+#define RPMLOOKUPSIG_DISABLE	1	/* Disable (--sign was not given) */
+#define RPMLOOKUPSIG_ENABLE	2	/* Re-enable %_signature          */
+
 /* Return type of signature in effect for building */
-int rpmLookupSignatureType(void);
+int rpmLookupSignatureType(int action);
 
 /* Utility to read a pass phrase from the user */
-char *rpmGetPassPhrase(char *prompt);
+char *rpmGetPassPhrase(const char *prompt, const int sigTag);
 
-#endif
+/* Return path to pgp executable of given type, or NULL when not found */
+const char *rpmDetectPGPVersion(int sigType);
+
+#endif	/* H_SIGNATURE */
