@@ -1,9 +1,4 @@
-#include "config.h"
-#include "miscfn.h"
-
-#include <stdlib.h>
-#include <ctype.h>
-#include <string.h>
+#include "system.h"
 
 #include "stringbuf.h"
 
@@ -63,7 +58,7 @@ char *getStringBuf(StringBuf sb)
     return sb->buf;
 }
 
-void appendStringBufAux(StringBuf sb, char *s, int nl)
+void appendStringBufAux(StringBuf sb, const char *s, int nl)
 {
     int l;
 
