@@ -1,3 +1,7 @@
+/* (C) 1998 Red Hat Software, Inc. -- Licensing details are in the COPYING
+   file accompanying popt source distributions, available from 
+   ftp://ftp.redhat.com/pub/code/popt */
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -27,7 +31,7 @@ static char * strerror(int errno) {
     if ((0 <= errno) && (errno < sys_nerr))
 	return sys_errlist[errno];
     else
-	return "unknown errno";
+	return POPT_("unknown errno");
 }
 #endif
 
@@ -37,6 +41,25 @@ void poptSetExecPath(poptContext con, const char * path, int allowAbsolute) {
     con->execAbsolute = allowAbsolute;
 }
 
+static void invokeCallbacks(poptContext con, const struct poptOption * table,
+			    int post) {
+    const struct poptOption * opt = table;
+    poptCallbackType cb;
+    
+    while (opt->longName || opt->shortName || opt->arg) {
+	if ((opt->argInfo & POPT_ARG_MASK) == POPT_ARG_INCLUDE_TABLE) {
+	    invokeCallbacks(con, opt->arg, post);
+	} else if (((opt->argInfo & POPT_ARG_MASK) == POPT_ARG_CALLBACK) &&
+		   ((!post && (opt->argInfo & POPT_CBFLAG_PRE)) ||
+		    ( post && (opt->argInfo & POPT_CBFLAG_POST)))) {
+	    cb = (poptCallbackType)opt->arg;
+	    cb(con, post ? POPT_CALLBACK_REASON_POST : POPT_CALLBACK_REASON_PRE,
+	       NULL, NULL, opt->descrip);
+	}
+	opt++;
+    }
+}
+
 poptContext poptGetContext(char * name, int argc, char ** argv, 
 			   const struct poptOption * options, int flags) {
     poptContext con = malloc(sizeof(*con));
@@ -63,10 +86,14 @@ poptContext poptGetContext(char * name, int argc, char ** argv,
     if (name)
 	con->appName = strcpy(malloc(strlen(name) + 1), name);
 
+    invokeCallbacks(con, con->options, 0);
+
     return con;
 }
 
 void poptResetContext(poptContext con) {
+    int i;
+
     con->os = con->optionStack;
     con->os->currAlias = NULL;
     con->os->nextCharArg = NULL;
@@ -77,6 +104,10 @@ void poptResetContext(poptContext con) {
     con->nextLeftover = 0;
     con->restLeftover = 0;
     con->doExec = NULL;
+
+    for (i = 0; i < con->finalArgvCount; i++)
+	free(con->finalArgv[i]);
+
     con->finalArgvCount = 0;
 }
 
@@ -129,7 +160,8 @@ static int handleAlias(poptContext con, char * longName, char shortName,
     if (con->os->currAlias && con->os->currAlias->longName && longName &&
 	!strcmp(con->os->currAlias->longName, longName)) 
 	return 0;
-    if (con->os->currAlias && shortName == con->os->currAlias->shortName)
+    if (con->os->currAlias && shortName && 
+	    shortName == con->os->currAlias->shortName)
 	return 0;
 
     i = con->numAliases - 1;
@@ -197,7 +229,18 @@ static void execCommand(poptContext con) {
 #ifdef __hpux
     setresuid(getuid(), getuid(),-1);
 #else
+/*
+ * XXX " ... on BSD systems setuid() should be preferred over setreuid()"
+ * XXX 	sez' Timur Bakeyev <mc@bat.ru>
+ * XXX	from Norbert Warmuth <nwarmuth@privat.circular.de>
+ */
+#if defined(HAVE_SETUID)
+    setuid(getuid());
+#elif defined (HAVE_SETREUID)
     setreuid(getuid(), getuid()); /*hlauer: not portable to hpux9.01 */
+#else
+    ; /* Can't drop privileges */
+#endif
 #endif
 
     execvp(argv[0], argv);
@@ -205,21 +248,31 @@ static void execCommand(poptContext con) {
 
 static const struct poptOption * findOption(const struct poptOption * table,
 					    const char * longName,
-					    const char shortName,
+					    char shortName,
 					    poptCallbackType * callback,
-					    void ** callbackData) {
+					    void ** callbackData,
+					    int singleDash) {
     const struct poptOption * opt = table;
     const struct poptOption * opt2;
     const struct poptOption * cb = NULL;
 
+    /* This happens when a single - is given */
+    if (singleDash && !shortName && !*longName)
+	shortName = '-';
+
     while (opt->longName || opt->shortName || opt->arg) {
 	if ((opt->argInfo & POPT_ARG_MASK) == POPT_ARG_INCLUDE_TABLE) {
 	    opt2 = findOption(opt->arg, longName, shortName, callback, 
-			      callbackData);
-	    if (opt2) return opt2;
+			      callbackData, singleDash);
+	    if (opt2) {
+		if (*callback && !*callbackData)
+		    *callbackData = opt->descrip;
+		return opt2;
+	    }
 	} else if ((opt->argInfo & POPT_ARG_MASK) == POPT_ARG_CALLBACK) {
 	    cb = opt;
 	} else if (longName && opt->longName && 
+		   (!singleDash || (opt->argInfo & POPT_ARGFLAG_ONEDASH)) &&
 		   !strcmp(longName, opt->longName)) {
 	    break;
 	} else if (shortName && shortName == opt->shortName) {
@@ -229,11 +282,12 @@ static const struct poptOption * findOption(const struct poptOption * table,
     }
 
     if (!opt->longName && !opt->shortName) return NULL;
+    *callbackData = NULL;
+    *callback = NULL;
     if (cb) {
-	*callback = cb->arg;
-	*callbackData = cb->descrip;
-    } else {
-	*callback = NULL;
+	*callback = (poptCallbackType)cb->arg;
+	if (!(cb->argInfo & POPT_CBFLAG_INC_DATA))
+	    *callbackData = cb->descrip;
     }
 
     return opt;
@@ -251,12 +305,14 @@ int poptGetNextOpt(poptContext con) {
     int i;
     poptCallbackType cb;
     void * cbData;
+    int singleDash;
 
     while (!done) {
 	while (!con->os->nextCharArg && con->os->next == con->os->argc 
 		&& con->os > con->optionStack)
 	    con->os--;
 	if (!con->os->nextCharArg && con->os->next == con->os->argc) {
+	    invokeCallbacks(con, con->options, 1);
 	    if (con->doExec) execCommand(con);
 	    return -1;
 	}
@@ -283,8 +339,12 @@ int poptGetNextOpt(poptContext con) {
 	    if (optString[1] == '-' && !optString[2]) {
 		con->restLeftover = 1;
 		continue;
-	    } else if (optString[1] == '-') {
-		optString += 2;
+	    } else {
+		optString++;
+		if (*optString == '-')
+		    singleDash = 0, optString++;
+		else
+		    singleDash = 1;
 
 		if (handleAlias(con, optString, '\0', NULL))
 		    continue;
@@ -298,9 +358,12 @@ int poptGetNextOpt(poptContext con) {
 		    *chptr = '\0';
 		}
 
-		opt = findOption(con->options, optString, '\0', &cb, &cbData);
-		if (!opt) return POPT_ERROR_BADOPT;
-	    } else 
+		opt = findOption(con->options, optString, '\0', &cb, &cbData,
+				 singleDash);
+		if (!opt && !singleDash) return POPT_ERROR_BADOPT;
+	    }
+
+	    if (!opt)
 		con->os->nextCharArg = origOptString + 1;
 	}
 
@@ -317,7 +380,8 @@ int poptGetNextOpt(poptContext con) {
 	    if (handleExec(con, NULL, *origOptString))
 		continue;
 
-	    opt = findOption(con->options, NULL, *origOptString, &cb, &cbData);
+	    opt = findOption(con->options, NULL, *origOptString, &cb, 
+			     &cbData, 0);
 	    if (!opt) return POPT_ERROR_BADOPT;
 
 	    origOptString++;
@@ -325,9 +389,11 @@ int poptGetNextOpt(poptContext con) {
 		con->os->nextCharArg = origOptString;
 	}
 
-	if (opt->arg && (opt->argInfo & POPT_ARG_MASK) == POPT_ARG_NONE) 
+	if (opt->arg && (opt->argInfo & POPT_ARG_MASK) == POPT_ARG_NONE) {
 	    *((int *)opt->arg) = 1;
-	else if ((opt->argInfo & POPT_ARG_MASK) != POPT_ARG_NONE) {
+	} else if ((opt->argInfo & POPT_ARG_MASK) == POPT_ARG_VAL) {
+	    if (opt->arg) *((int *) opt->arg) = opt->val;
+	} else if ((opt->argInfo & POPT_ARG_MASK) != POPT_ARG_NONE) {
 	    if (longArg) {
 		con->os->nextArg = longArg;
 	    } else if (con->os->nextCharArg) {
@@ -352,7 +418,7 @@ int poptGetNextOpt(poptContext con) {
 		  case POPT_ARG_INT:
 		  case POPT_ARG_LONG:
 		    aLong = strtol(con->os->nextArg, &end, 0);
-		    if (*end) 
+		    if (!(end && *end == '\0')) 
 			return POPT_ERROR_BADNUMBER;
 
 		    if (aLong == LONG_MIN || aLong == LONG_MAX)
@@ -367,8 +433,7 @@ int poptGetNextOpt(poptContext con) {
 		    break;
 
 		  default:
-		    /* XXX I18N? */
-		    fprintf(stdout, "option type (%d) not implemented in popt\n",
+		    fprintf(stdout, POPT_("option type (%d) not implemented in popt\n"),
 		      opt->argInfo & POPT_ARG_MASK);
 		    exit(1);
 		}
@@ -376,8 +441,8 @@ int poptGetNextOpt(poptContext con) {
 	}
 
 	if (cb)
-	    cb(con, opt, con->os->nextArg, cbData);
-	else if (opt->val) 
+	    cb(con, POPT_CALLBACK_REASON_OPTION, opt, con->os->nextArg, cbData);
+	else if (opt->val && ((opt->argInfo & POPT_ARG_MASK) != POPT_ARG_VAL))
 	    done = 1;
 
 	if ((con->finalArgvCount + 2) >= (con->finalArgvAlloced)) {
@@ -394,7 +459,8 @@ int poptGetNextOpt(poptContext con) {
 	else 
 	    sprintf(con->finalArgv[i], "-%c", opt->shortName);
 
-	if (opt->arg && (opt->argInfo & POPT_ARG_MASK) != POPT_ARG_NONE) 
+	if (opt->arg && (opt->argInfo & POPT_ARG_MASK) != POPT_ARG_NONE
+		     && (opt->argInfo & POPT_ARG_MASK) != POPT_ARG_VAL) 
 	    con->finalArgv[con->finalArgvCount++] = strdup(con->os->nextArg);
     }
 
@@ -447,6 +513,7 @@ void poptFreeContext(poptContext con) {
     if (con->appName) free(con->appName);
     if (con->aliases) free(con->aliases);
     if (con->otherHelp) free(con->otherHelp);
+    if (con->execPath) free(con->execPath);
     free(con);
 }
 
@@ -492,21 +559,21 @@ char * poptBadOption(poptContext con, int flags) {
 const char * poptStrerror(const int error) {
     switch (error) {
       case POPT_ERROR_NOARG:
-	return "missing argument";
+	return POPT_("missing argument");
       case POPT_ERROR_BADOPT:
-	return "unknown option";
+	return POPT_("unknown option");
       case POPT_ERROR_OPTSTOODEEP:
-	return "aliases nested too deeply";
+	return POPT_("aliases nested too deeply");
       case POPT_ERROR_BADQUOTE:
-	return "error in paramter quoting";
+	return POPT_("error in paramter quoting");
       case POPT_ERROR_BADNUMBER:
-	return "invalid numeric value";
+	return POPT_("invalid numeric value");
       case POPT_ERROR_OVERFLOW:
-	return "number too large or too small";
+	return POPT_("number too large or too small");
       case POPT_ERROR_ERRNO:
 	return strerror(errno);
       default:
-	return "unknown error";
+	return POPT_("unknown error");
     }
 }
 
@@ -528,3 +595,7 @@ int poptStuffArgs(poptContext con, char ** argv) {
 
     return 0;
 }
+
+const char * poptGetInvocationName(poptContext con) {
+    return con->os->argv[0];
+}
