@@ -1,13 +1,6 @@
-#include <errno.h>
-#include <fcntl.h>
-#include <stdio.h>
-#include <string.h>
-#include <unistd.h>
-#include <stdlib.h>
+#include "system.h"
 
-#include "header.h"
 #include "rpmlib.h"
-#include "intl.h"
 
 int main(int argc, char ** argv)
 {
@@ -17,7 +10,7 @@ int main(int argc, char ** argv)
     int blockNum = 0;
     rpmdb db;
 
-    rpmReadConfigFiles(NULL, NULL, NULL, 0);
+    rpmReadConfigFiles(NULL, NULL);
 
     if (argc == 2) {
 	dspBlockNum = atoi(argv[1]);
@@ -56,5 +49,3 @@ int main(int argc, char ** argv)
 
     return 0;
 }
-
-  
