@@ -1,13 +1,11 @@
 /* rpmchecksig: verify the signature of an RPM */
 
-#include <stdio.h>
-#include <unistd.h>
-#include <fcntl.h>
+#include "system.h"
 
 #include "rpmlib.h"
+
 #include "rpmlead.h"
 #include "signature.h"
-#include "intl.h"
 
 int main(int argc, char **argv)
 {
