@@ -1,33 +1,35 @@
 /* rpmsignature: spit out the signature portion of a package */
 
-#include <stdio.h>
-#include <unistd.h>
-#include <fcntl.h>
+#include "system.h"
 
 #include "rpmlead.h"
 #include "signature.h"
-#include "intl.h"
 
 int main(int argc, char **argv)
 {
-    int fd;
+    FD_t fdi, fdo;
     struct rpmlead lead;
     Header sig;
     
     if (argc == 1) {
-	fd = 0;
+	fdi = fdDup(STDIN_FILENO);
     } else {
-	fd = open(argv[1], O_RDONLY, 0644);
+	fdi = fdOpen(argv[1], O_RDONLY, 0644);
+	if (fdFileno(fdi) < 0) {
+	    perror(argv[1]);
+	    exit(1);
+	}
     }
 
-    readLead(fd, &lead);
-    rpmReadSignature(fd, &sig, lead.signature_type);
+    readLead(fdi, &lead);
+    rpmReadSignature(fdi, &sig, lead.signature_type);
     switch (lead.signature_type) {
       case RPMSIG_NONE:
 	fprintf(stderr, _("No signature available.\n"));
 	break;
       default:
-	rpmWriteSignature(1, sig);
+	fdo = fdDup(STDOUT_FILENO);
+	rpmWriteSignature(fdo, sig);
     }
     
     return 0;
